/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>
#include <cassert>

#include <Pass.h>
#include <Basics.h>
#include <Degree.h>
#include <Config.h>
#include <Logger.h>
#include <Defines.h>
#include <BasicKick.h>
#include <SayDecision.h>

#include <AdvancedAgent.h>


#define PW_VERY_BAD -75
#define PW_BAD -30
#define PW_NOTHING -15
#define PW_GOOD 60
#define PW_VERY_GOOD 85


using namespace std;
using namespace Basics;
using namespace Degree;

// class CrossCalculate
CrossCalculate::~CrossCalculate()
{
}

void CrossCalculate::init()
{
	setInited(true);

	simulatesNum = CROSS_CALCULATES_NUM;
	simulates = new KickInterceptCalculate[simulatesNum];
}

void CrossCalculate::update(const WorldModel *worldModel)
{
	InterceptCalculate interCalc;
	Command *kickCommand = NULL;

	LOG << "CrossCalculate::update" << endl;

	updateTime = worldModel->getCurTime();

	interCalc.startSession(1, &worldModel->getBody());
//	interCalc.setLogging(true);

	interCalc.setShareCycles(0);

	interCalc.setKickerSeeBallDelay(1);

	interCalc.setTmmPosDevRate(-0.3);
	interCalc.setTmmSlowDashPower(0);
	interCalc.setTmmFastDashPower(100);
	interCalc.setTmmDirectDashPower(0);
	interCalc.setTmmSeeBallDelay(1);

	interCalc.setOppPosDevRate(0.5);
	interCalc.setOppSlowDashPower(10);
	interCalc.setOppFastDashPower(100);
	interCalc.setOppDirectDashPower(60);
	interCalc.setOppGoalieDirectDashPower(100);
	interCalc.setOppSeeBallDelay(1);
	interCalc.setOppGoalieSeeBallDelay(1);

	interCalc.updateByWorldModel(worldModel);

	Vector vector;
	vector.setByPoints(worldModel->getBody().getPos(), CROSS_CENTER_POINT);

	for (unsigned i = 0; i < simulatesNum; i++)
	{
//cout << "Ahmad: Cross pos: " << worldModel->getBody().getPos() + vector << endl;
		kickCommand = FastKickToDir(AT_CROSS_CALCULATE, vector.getDirection() +
									((int)i - CROSS_DIR_NUM / 2.00) * CROSS_DIR_STEP,
									worldModel->getBall(), worldModel->getBody()).getCommand();
//cout << "Ahmad: Cross command: " << kickCommand->toString() << endl;

		simulates[i].init(worldModel,
						  dynamic_cast<KickCommand *>(kickCommand), interCalc);

		delete kickCommand;
	}
}


// class Pass
Pass::Pass(const WorldModel *worldModel, bool dontCreateCrossCalc):
		AdvancedAction(worldModel), dribblePath(worldModel)
{
	logLevel = 0;
	passCommand = NULL;
	loadPassFactors();

	if (!dontCreateCrossCalc)
	{
		crossCalculate = new CrossCalculate();
		crossCalculate->init();
	}
	else
		crossCalculate = NULL;
}

Pass::~Pass()
{
	delete crossCalculate;
	crossCalculate = NULL;
}

float Pass::getValue(const Library &library)
{
	LOG << "Pass::getValue" << endl;

	delete passCommand;
	passCommand = NULL;

	if (worldModel->getPlayMode() == PM_KICK_OFF)
	{
		LOG << "Pass: Our kick off mode." << endl;

		if (worldModel->getCurTime() % 2)
		{
			passCommand = new KickCommand(AT_PASS, 100, -115);
			passPlayer = &worldModel->getFullPlayer(TID_TEAMMATE, 11 - 1);
		}
		else
		{
			passCommand = new KickCommand(AT_PASS, 100, 115);
			passPlayer = &worldModel->getFullPlayer(TID_TEAMMATE, 10 - 1);
		}

		passInterCalc = NULL;
		passModel = PAM_KICK_OFF;
		passValue = AD_NORMAL_VALUE;
		return passValue;
	}

	if (logLevel >= 1)
		loadPassFactors();

	calculateGlobalVars();

	if (crossRate)
		crossCalculate->update(worldModel);

	resetValues();
	fillValues();
	fillMaxValues();
	passValue = getTheBestPass(passInterCalc);

	if (worldModel->getHearedSuggestPassTime() == worldModel->getCurTime())
	{
		LOG << "Suggested pass:" << endl
			<< "\tPlayer: " << worldModel->getHearedSuggestPassNum() << endl
			<< "\tWeight: " << worldModel->getHearedSuggestPassWeight() << endl
			<< "\tSpeed: " << worldModel->getHearedSuggestPassSpeed() << endl
			<< "\tDir: " << worldModel->getHearedSuggestPassDir() << endl;

		if (worldModel->getHearedSuggestPassWeight() +
			SUGGEST_REAL_DELTA_WEIGHT > passValue)
		{
			KickToDirectionVelocity basicKick(AT_PASS,
				worldModel->getHearedSuggestPassDir(),
				worldModel->getHearedSuggestPassSpeed(),
				worldModel->getBall(), worldModel->getBody());

			if (basicKick.isPossible())
			{
				LOG << "Suggested pass accepted." << endl;

				passModel = PAM_SUGGESTED;
				passValue = worldModel->getHearedSuggestPassWeight() +
					SUGGEST_REAL_DELTA_WEIGHT;

				passCommand = dynamic_cast<KickCommand *>
					(basicKick.getCommand());
			}
			else
				LOG << "Suggested pass refused: kick is not possible." << endl;
		}
		else
			LOG << "Suggested pass refused: less weight." << endl;
	}

	if (passModel == PAM_SUGGESTED)
	{
		if (worldModel->getHearedSuggestPassNum() != 0)
			passPlayer = &worldModel->getFullPlayer(TID_TEAMMATE,
													worldModel->getHearedSuggestPassNum() - 1);
		else
			passPlayer = &worldModel->getHalfPlayer(TID_TEAMMATE, 0);
	}
	else
	{
		if (passInterCalc == NULL)
		{
			passPlayer = NULL;
			return AD_MIN_VALUE;
		}

		passCommand = new KickCommand(passInterCalc->getKickCommand());
		passPlayer = &passInterCalc->getFastestTmmInterPlayer().getPlayer();
	}

	return passValue;
}

void Pass::execute(Form &form, const Library &library)
{
	LOG << "Pass::execute" << endl;

	string passModelStr;

	form.sayForm.interceptNum = 0;
	form.sayForm.beforePassSay = true;

	if (passPlayer != NULL)
	{
		switch (passModel)
		{
			case PAM_NONE:
				passModelStr = "None";
				break;
			case PAM_CROSS:
				passModelStr = "Cross";
				break;
			case PAM_OFFENSE:
				passModelStr = "Offense";
				break;
			case PAM_SECURE_BALL:
				passModelStr = "SecureBall";
				break;
			case PAM_BREAK_OFFSIDE:
				passModelStr = "BreakOffside";
				break;
			case PAM_SUGGESTED:
				passModelStr = "Suggested";
				break;
			case PAM_KICK_OFF:
				passModelStr = "KickOff";
				break;

			default:
				assert(0);
		}

		if (passPlayer->getModel() == PLM_FULL)
		{
			form.sayForm.interceptNum = passPlayer->getUniNum();
			LOG << passModelStr << " pass to player: "
				<< passPlayer->getUniNum() << endl;
		}
		else
			LOG << passModelStr << " pass to player: "
				<< "<HalfPlayer>" << endl;
	}
	else
		LOG << "*** There is nobody to pass and Pass is executed ***" << endl;

	if (passCommand)
	{
		command = new KickCommand(*passCommand);
		command->setCreator(AT_PASS);
	}
	else
		command = new EmptyCommand(AT_PASS);

	if (logLevel >= 4)
	{
			// Debuging GoToBallCalculate before simulating ball for next cycle.
		InterceptCalculate interCalc;
		Ball ball = worldModel->getBall();
		ball.simulateByAction(worldModel->getBody(), command);
//		ball.simulateByDynamics(worldModel->getBody());

		interCalc.startSession(1, &worldModel->getBody());
		interCalc.setLogging(true);
		interCalc.updateByWorldModel(worldModel);
		interCalc.setVirtualBall(ball);
		interCalc.run();
	}

}

void Pass::loadPassFactors()
{
	config.add("Pass");
	config["Pass"].setByFile("./Configs/PassFactors.conf");

	// crossFactors
	crossFactors.posDev.minValue = config["Pass"]["PosDeviation"]["MinValue"][0].asFloat();
	crossFactors.posDev.maxValue = config["Pass"]["PosDeviation"]["MaxValue"][0].asFloat();
	crossFactors.posDev.minWeight = config["Pass"]["PosDeviation"]["MinWeight"][0].asFloat();
	crossFactors.posDev.maxWeight = config["Pass"]["PosDeviation"]["MaxWeight"][0].asFloat();

	crossFactors.offense.minValue = config["Pass"]["Offense"]["MinValue"][0].asFloat();
	crossFactors.offense.maxValue = config["Pass"]["Offense"]["MaxValue"][0].asFloat();
	crossFactors.offense.minWeight = config["Pass"]["Offense"]["MinWeight"][0].asFloat();
	crossFactors.offense.maxWeight = config["Pass"]["Offense"]["MaxWeight"][0].asFloat();

	crossFactors.seeInfo.minValue = config["Pass"]["SeeInformation"]["MinValue"][0].asFloat();
	crossFactors.seeInfo.maxValue = config["Pass"]["SeeInformation"]["MaxValue"][0].asFloat();
	crossFactors.seeInfo.minWeight = config["Pass"]["SeeInformation"]["MinWeight"][0].asFloat();
	crossFactors.seeInfo.maxWeight = config["Pass"]["SeeInformation"]["MaxWeight"][0].asFloat();

	crossFactors.ballDist.minValue = config["Pass"]["BallDistance"]["MinValue"][0].asFloat();
	crossFactors.ballDist.maxValue = config["Pass"]["BallDistance"]["MaxValue"][0].asFloat();
	crossFactors.ballDist.minWeight = config["Pass"]["BallDistance"]["MinWeight"][0].asFloat();
	crossFactors.ballDist.maxWeight = config["Pass"]["BallDistance"]["MaxWeight"][0].asFloat();

	crossFactors.dashDist.minValue = config["Pass"]["DashDistance"]["MinValue"][0].asFloat();
	crossFactors.dashDist.maxValue = config["Pass"]["DashDistance"]["MaxValue"][0].asFloat();
	crossFactors.dashDist.minWeight = config["Pass"]["DashDistance"]["MinWeight"][0].asFloat();
	crossFactors.dashDist.maxWeight = config["Pass"]["DashDistance"]["MaxWeight"][0].asFloat();

	crossFactors.breakOffside.minValue = config["Pass"]["BreakOffsideLine"]["MinValue"][0].asFloat();
	crossFactors.breakOffside.maxValue = config["Pass"]["BreakOffsideLine"]["MaxValue"][0].asFloat();
	crossFactors.breakOffside.minWeight = config["Pass"]["BreakOffsideLine"]["MinWeight"][0].asFloat();
	crossFactors.breakOffside.maxWeight = config["Pass"]["BreakOffsideLine"]["MaxWeight"][0].asFloat();

	crossFactors.passedOppNum.minValue = config["Pass"]["PassedOppNum"]["MinValue"][0].asFloat();
	crossFactors.passedOppNum.maxValue = config["Pass"]["PassedOppNum"]["MaxValue"][0].asFloat();
	crossFactors.passedOppNum.minWeight = config["Pass"]["PassedOppNum"]["MinWeight"][0].asFloat();
	crossFactors.passedOppNum.maxWeight = config["Pass"]["PassedOppNum"]["MaxWeight"][0].asFloat();

	crossFactors.statusOppNum.minValue = config["Pass"]["StatusOppNum"]["MinValue"][0].asFloat();
	crossFactors.statusOppNum.maxValue = config["Pass"]["StatusOppNum"]["MaxValue"][0].asFloat();
	crossFactors.statusOppNum.minWeight = config["Pass"]["StatusOppNum"]["MinWeight"][0].asFloat();
	crossFactors.statusOppNum.maxWeight = config["Pass"]["StatusOppNum"]["MaxWeight"][0].asFloat();

	crossFactors.securityDeltaCycle.minValue = config["Pass"]["SecurityDeltaCycle"]["MinValue"][0].asFloat();
	crossFactors.securityDeltaCycle.maxValue = config["Pass"]["SecurityDeltaCycle"]["MaxValue"][0].asFloat();
	crossFactors.securityDeltaCycle.minWeight = config["Pass"]["SecurityDeltaCycle"]["MinWeight"][0].asFloat();
	crossFactors.securityDeltaCycle.maxWeight = config["Pass"]["SecurityDeltaCycle"]["MaxWeight"][0].asFloat();

	crossFactors.statusOppDist.minValue = config["Pass"]["StatusOppDistance"]["MinValue"][0].asFloat();
	crossFactors.statusOppDist.maxValue = config["Pass"]["StatusOppDistance"]["MaxValue"][0].asFloat();
	crossFactors.statusOppDist.minWeight = config["Pass"]["StatusOppDistance"]["MinWeight"][0].asFloat();
	crossFactors.statusOppDist.maxWeight = config["Pass"]["StatusOppDistance"]["MaxWeight"][0].asFloat();

	crossFactors.statusOppDir.minValue = config["Pass"]["StatusOppDirection"]["MinValue"][0].asFloat();
	crossFactors.statusOppDir.maxValue = config["Pass"]["StatusOppDirection"]["MaxValue"][0].asFloat();
	crossFactors.statusOppDir.minWeight = config["Pass"]["StatusOppDirection"]["MinWeight"][0].asFloat();
	crossFactors.statusOppDir.maxWeight = config["Pass"]["StatusOppDirection"]["MaxWeight"][0].asFloat();

	crossFactors.statusShoot.minValue = config["Pass"]["StatusShoot"]["MinValue"][0].asFloat();
	crossFactors.statusShoot.maxValue = config["Pass"]["StatusShoot"]["MaxValue"][0].asFloat();
	crossFactors.statusShoot.minWeight = config["Pass"]["StatusShoot"]["MinWeight"][0].asFloat();
	crossFactors.statusShoot.maxWeight = config["Pass"]["StatusShoot"]["MaxWeight"][0].asFloat();

	crossFactors.dribblePath.minValue = config["Pass"]["DribblePath"]["MinValue"][0].asFloat();
	crossFactors.dribblePath.maxValue = config["Pass"]["DribblePath"]["MaxValue"][0].asFloat();
	crossFactors.dribblePath.minWeight = config["Pass"]["DribblePath"]["MinWeight"][0].asFloat();
	crossFactors.dribblePath.maxWeight = config["Pass"]["DribblePath"]["MaxWeight"][0].asFloat();

	crossFactors.defenseLine.minValue = config["Pass"]["DefenseLine"]["MinValue"][0].asFloat();
	crossFactors.defenseLine.maxValue = config["Pass"]["DefenseLine"]["MaxValue"][0].asFloat();
	crossFactors.defenseLine.minWeight = config["Pass"]["DefenseLine"]["MinWeight"][0].asFloat();
	crossFactors.defenseLine.maxWeight = config["Pass"]["DefenseLine"]["MaxWeight"][0].asFloat();

	crossFactors.extraCatch.minValue = config["Pass"]["ExtraCatch"]["MinValue"][0].asFloat();
	crossFactors.extraCatch.maxValue = config["Pass"]["ExtraCatch"]["MaxValue"][0].asFloat();
	crossFactors.extraCatch.minWeight = config["Pass"]["ExtraCatch"]["MinWeight"][0].asFloat();
	crossFactors.extraCatch.maxWeight = config["Pass"]["ExtraCatch"]["MaxWeight"][0].asFloat();

	// offenseFactors
	offenseFactors.posDev.minValue = config["Pass"]["PosDeviation"]["MinValue"][1].asFloat();
	offenseFactors.posDev.maxValue = config["Pass"]["PosDeviation"]["MaxValue"][1].asFloat();
	offenseFactors.posDev.minWeight = config["Pass"]["PosDeviation"]["MinWeight"][1].asFloat();
	offenseFactors.posDev.maxWeight = config["Pass"]["PosDeviation"]["MaxWeight"][1].asFloat();

	offenseFactors.offense.minValue = config["Pass"]["Offense"]["MinValue"][1].asFloat();
	offenseFactors.offense.maxValue = config["Pass"]["Offense"]["MaxValue"][1].asFloat();
	offenseFactors.offense.minWeight = config["Pass"]["Offense"]["MinWeight"][1].asFloat();
	offenseFactors.offense.maxWeight = config["Pass"]["Offense"]["MaxWeight"][1].asFloat();

	offenseFactors.seeInfo.minValue = config["Pass"]["SeeInformation"]["MinValue"][1].asFloat();
	offenseFactors.seeInfo.maxValue = config["Pass"]["SeeInformation"]["MaxValue"][1].asFloat();
	offenseFactors.seeInfo.minWeight = config["Pass"]["SeeInformation"]["MinWeight"][1].asFloat();
	offenseFactors.seeInfo.maxWeight = config["Pass"]["SeeInformation"]["MaxWeight"][1].asFloat();

	offenseFactors.ballDist.minValue = config["Pass"]["BallDistance"]["MinValue"][1].asFloat();
	offenseFactors.ballDist.maxValue = config["Pass"]["BallDistance"]["MaxValue"][1].asFloat();
	offenseFactors.ballDist.minWeight = config["Pass"]["BallDistance"]["MinWeight"][1].asFloat();
	offenseFactors.ballDist.maxWeight = config["Pass"]["BallDistance"]["MaxWeight"][1].asFloat();

	offenseFactors.dashDist.minValue = config["Pass"]["DashDistance"]["MinValue"][1].asFloat();
	offenseFactors.dashDist.maxValue = config["Pass"]["DashDistance"]["MaxValue"][1].asFloat();
	offenseFactors.dashDist.minWeight = config["Pass"]["DashDistance"]["MinWeight"][1].asFloat();
	offenseFactors.dashDist.maxWeight = config["Pass"]["DashDistance"]["MaxWeight"][1].asFloat();

	offenseFactors.breakOffside.minValue = config["Pass"]["BreakOffsideLine"]["MinValue"][1].asFloat();
	offenseFactors.breakOffside.maxValue = config["Pass"]["BreakOffsideLine"]["MaxValue"][1].asFloat();
	offenseFactors.breakOffside.minWeight = config["Pass"]["BreakOffsideLine"]["MinWeight"][1].asFloat();
	offenseFactors.breakOffside.maxWeight = config["Pass"]["BreakOffsideLine"]["MaxWeight"][1].asFloat();

	offenseFactors.passedOppNum.minValue = config["Pass"]["PassedOppNum"]["MinValue"][1].asFloat();
	offenseFactors.passedOppNum.maxValue = config["Pass"]["PassedOppNum"]["MaxValue"][1].asFloat();
	offenseFactors.passedOppNum.minWeight = config["Pass"]["PassedOppNum"]["MinWeight"][1].asFloat();
	offenseFactors.passedOppNum.maxWeight = config["Pass"]["PassedOppNum"]["MaxWeight"][1].asFloat();

	offenseFactors.statusOppNum.minValue = config["Pass"]["StatusOppNum"]["MinValue"][1].asFloat();
	offenseFactors.statusOppNum.maxValue = config["Pass"]["StatusOppNum"]["MaxValue"][1].asFloat();
	offenseFactors.statusOppNum.minWeight = config["Pass"]["StatusOppNum"]["MinWeight"][1].asFloat();
	offenseFactors.statusOppNum.maxWeight = config["Pass"]["StatusOppNum"]["MaxWeight"][1].asFloat();

	offenseFactors.securityDeltaCycle.minValue = config["Pass"]["SecurityDeltaCycle"]["MinValue"][1].asFloat();
	offenseFactors.securityDeltaCycle.maxValue = config["Pass"]["SecurityDeltaCycle"]["MaxValue"][1].asFloat();
	offenseFactors.securityDeltaCycle.minWeight = config["Pass"]["SecurityDeltaCycle"]["MinWeight"][1].asFloat();
	offenseFactors.securityDeltaCycle.maxWeight = config["Pass"]["SecurityDeltaCycle"]["MaxWeight"][1].asFloat();

	offenseFactors.statusOppDist.minValue = config["Pass"]["StatusOppDistance"]["MinValue"][1].asFloat();
	offenseFactors.statusOppDist.maxValue = config["Pass"]["StatusOppDistance"]["MaxValue"][1].asFloat();
	offenseFactors.statusOppDist.minWeight = config["Pass"]["StatusOppDistance"]["MinWeight"][1].asFloat();
	offenseFactors.statusOppDist.maxWeight = config["Pass"]["StatusOppDistance"]["MaxWeight"][1].asFloat();

	offenseFactors.statusOppDir.minValue = config["Pass"]["StatusOppDirection"]["MinValue"][1].asFloat();
	offenseFactors.statusOppDir.maxValue = config["Pass"]["StatusOppDirection"]["MaxValue"][1].asFloat();
	offenseFactors.statusOppDir.minWeight = config["Pass"]["StatusOppDirection"]["MinWeight"][1].asFloat();
	offenseFactors.statusOppDir.maxWeight = config["Pass"]["StatusOppDirection"]["MaxWeight"][1].asFloat();

	offenseFactors.statusShoot.minValue = config["Pass"]["StatusShoot"]["MinValue"][1].asFloat();
	offenseFactors.statusShoot.maxValue = config["Pass"]["StatusShoot"]["MaxValue"][1].asFloat();
	offenseFactors.statusShoot.minWeight = config["Pass"]["StatusShoot"]["MinWeight"][1].asFloat();
	offenseFactors.statusShoot.maxWeight = config["Pass"]["StatusShoot"]["MaxWeight"][1].asFloat();

	offenseFactors.dribblePath.minValue = config["Pass"]["DribblePath"]["MinValue"][1].asFloat();
	offenseFactors.dribblePath.maxValue = config["Pass"]["DribblePath"]["MaxValue"][1].asFloat();
	offenseFactors.dribblePath.minWeight = config["Pass"]["DribblePath"]["MinWeight"][1].asFloat();
	offenseFactors.dribblePath.maxWeight = config["Pass"]["DribblePath"]["MaxWeight"][1].asFloat();

	offenseFactors.defenseLine.minValue = config["Pass"]["DefenseLine"]["MinValue"][1].asFloat();
	offenseFactors.defenseLine.maxValue = config["Pass"]["DefenseLine"]["MaxValue"][1].asFloat();
	offenseFactors.defenseLine.minWeight = config["Pass"]["DefenseLine"]["MinWeight"][1].asFloat();
	offenseFactors.defenseLine.maxWeight = config["Pass"]["DefenseLine"]["MaxWeight"][1].asFloat();

	offenseFactors.extraCatch.minValue = config["Pass"]["ExtraCatch"]["MinValue"][1].asFloat();
	offenseFactors.extraCatch.maxValue = config["Pass"]["ExtraCatch"]["MaxValue"][1].asFloat();
	offenseFactors.extraCatch.minWeight = config["Pass"]["ExtraCatch"]["MinWeight"][1].asFloat();
	offenseFactors.extraCatch.maxWeight = config["Pass"]["ExtraCatch"]["MaxWeight"][1].asFloat();

	// secureBallFactors
	secureBallFactors.posDev.minValue = config["Pass"]["PosDeviation"]["MinValue"][2].asFloat();
	secureBallFactors.posDev.maxValue = config["Pass"]["PosDeviation"]["MaxValue"][2].asFloat();
	secureBallFactors.posDev.minWeight = config["Pass"]["PosDeviation"]["MinWeight"][2].asFloat();
	secureBallFactors.posDev.maxWeight = config["Pass"]["PosDeviation"]["MaxWeight"][2].asFloat();

	secureBallFactors.offense.minValue = config["Pass"]["Offense"]["MinValue"][2].asFloat();
	secureBallFactors.offense.maxValue = config["Pass"]["Offense"]["MaxValue"][2].asFloat();
	secureBallFactors.offense.minWeight = config["Pass"]["Offense"]["MinWeight"][2].asFloat();
	secureBallFactors.offense.maxWeight = config["Pass"]["Offense"]["MaxWeight"][2].asFloat();

	secureBallFactors.seeInfo.minValue = config["Pass"]["SeeInformation"]["MinValue"][2].asFloat();
	secureBallFactors.seeInfo.maxValue = config["Pass"]["SeeInformation"]["MaxValue"][2].asFloat();
	secureBallFactors.seeInfo.minWeight = config["Pass"]["SeeInformation"]["MinWeight"][2].asFloat();
	secureBallFactors.seeInfo.maxWeight = config["Pass"]["SeeInformation"]["MaxWeight"][2].asFloat();

	secureBallFactors.ballDist.minValue = config["Pass"]["BallDistance"]["MinValue"][2].asFloat();
	secureBallFactors.ballDist.maxValue = config["Pass"]["BallDistance"]["MaxValue"][2].asFloat();
	secureBallFactors.ballDist.minWeight = config["Pass"]["BallDistance"]["MinWeight"][2].asFloat();
	secureBallFactors.ballDist.maxWeight = config["Pass"]["BallDistance"]["MaxWeight"][2].asFloat();

	secureBallFactors.dashDist.minValue = config["Pass"]["DashDistance"]["MinValue"][2].asFloat();
	secureBallFactors.dashDist.maxValue = config["Pass"]["DashDistance"]["MaxValue"][2].asFloat();
	secureBallFactors.dashDist.minWeight = config["Pass"]["DashDistance"]["MinWeight"][2].asFloat();
	secureBallFactors.dashDist.maxWeight = config["Pass"]["DashDistance"]["MaxWeight"][2].asFloat();

	secureBallFactors.breakOffside.minValue = config["Pass"]["BreakOffsideLine"]["MinValue"][2].asFloat();
	secureBallFactors.breakOffside.maxValue = config["Pass"]["BreakOffsideLine"]["MaxValue"][2].asFloat();
	secureBallFactors.breakOffside.minWeight = config["Pass"]["BreakOffsideLine"]["MinWeight"][2].asFloat();
	secureBallFactors.breakOffside.maxWeight = config["Pass"]["BreakOffsideLine"]["MaxWeight"][2].asFloat();

	secureBallFactors.passedOppNum.minValue = config["Pass"]["PassedOppNum"]["MinValue"][2].asFloat();
	secureBallFactors.passedOppNum.maxValue = config["Pass"]["PassedOppNum"]["MaxValue"][2].asFloat();
	secureBallFactors.passedOppNum.minWeight = config["Pass"]["PassedOppNum"]["MinWeight"][2].asFloat();
	secureBallFactors.passedOppNum.maxWeight = config["Pass"]["PassedOppNum"]["MaxWeight"][2].asFloat();

	secureBallFactors.statusOppNum.minValue = config["Pass"]["StatusOppNum"]["MinValue"][2].asFloat();
	secureBallFactors.statusOppNum.maxValue = config["Pass"]["StatusOppNum"]["MaxValue"][2].asFloat();
	secureBallFactors.statusOppNum.minWeight = config["Pass"]["StatusOppNum"]["MinWeight"][2].asFloat();
	secureBallFactors.statusOppNum.maxWeight = config["Pass"]["StatusOppNum"]["MaxWeight"][2].asFloat();

	secureBallFactors.securityDeltaCycle.minValue = config["Pass"]["SecurityDeltaCycle"]["MinValue"][2].asFloat();
	secureBallFactors.securityDeltaCycle.maxValue = config["Pass"]["SecurityDeltaCycle"]["MaxValue"][2].asFloat();
	secureBallFactors.securityDeltaCycle.minWeight = config["Pass"]["SecurityDeltaCycle"]["MinWeight"][2].asFloat();
	secureBallFactors.securityDeltaCycle.maxWeight = config["Pass"]["SecurityDeltaCycle"]["MaxWeight"][2].asFloat();

	secureBallFactors.statusOppDist.minValue = config["Pass"]["StatusOppDistance"]["MinValue"][2].asFloat();
	secureBallFactors.statusOppDist.maxValue = config["Pass"]["StatusOppDistance"]["MaxValue"][2].asFloat();
	secureBallFactors.statusOppDist.minWeight = config["Pass"]["StatusOppDistance"]["MinWeight"][2].asFloat();
	secureBallFactors.statusOppDist.maxWeight = config["Pass"]["StatusOppDistance"]["MaxWeight"][2].asFloat();

	secureBallFactors.statusOppDir.minValue = config["Pass"]["StatusOppDirection"]["MinValue"][2].asFloat();
	secureBallFactors.statusOppDir.maxValue = config["Pass"]["StatusOppDirection"]["MaxValue"][2].asFloat();
	secureBallFactors.statusOppDir.minWeight = config["Pass"]["StatusOppDirection"]["MinWeight"][2].asFloat();
	secureBallFactors.statusOppDir.maxWeight = config["Pass"]["StatusOppDirection"]["MaxWeight"][2].asFloat();

	secureBallFactors.statusShoot.minValue = config["Pass"]["StatusShoot"]["MinValue"][2].asFloat();
	secureBallFactors.statusShoot.maxValue = config["Pass"]["StatusShoot"]["MaxValue"][2].asFloat();
	secureBallFactors.statusShoot.minWeight = config["Pass"]["StatusShoot"]["MinWeight"][2].asFloat();
	secureBallFactors.statusShoot.maxWeight = config["Pass"]["StatusShoot"]["MaxWeight"][2].asFloat();

	secureBallFactors.dribblePath.minValue = config["Pass"]["DribblePath"]["MinValue"][2].asFloat();
	secureBallFactors.dribblePath.maxValue = config["Pass"]["DribblePath"]["MaxValue"][2].asFloat();
	secureBallFactors.dribblePath.minWeight = config["Pass"]["DribblePath"]["MinWeight"][2].asFloat();
	secureBallFactors.dribblePath.maxWeight = config["Pass"]["DribblePath"]["MaxWeight"][2].asFloat();

	secureBallFactors.defenseLine.minValue = config["Pass"]["DefenseLine"]["MinValue"][2].asFloat();
	secureBallFactors.defenseLine.maxValue = config["Pass"]["DefenseLine"]["MaxValue"][2].asFloat();
	secureBallFactors.defenseLine.minWeight = config["Pass"]["DefenseLine"]["MinWeight"][2].asFloat();
	secureBallFactors.defenseLine.maxWeight = config["Pass"]["DefenseLine"]["MaxWeight"][2].asFloat();

	secureBallFactors.extraCatch.minValue = config["Pass"]["ExtraCatch"]["MinValue"][2].asFloat();
	secureBallFactors.extraCatch.maxValue = config["Pass"]["ExtraCatch"]["MaxValue"][2].asFloat();
	secureBallFactors.extraCatch.minWeight = config["Pass"]["ExtraCatch"]["MinWeight"][2].asFloat();
	secureBallFactors.extraCatch.maxWeight = config["Pass"]["ExtraCatch"]["MaxWeight"][2].asFloat();

	// breakOffsideFactors
	breakOffsideFactors.posDev.minValue = config["Pass"]["PosDeviation"]["MinValue"][3].asFloat();
	breakOffsideFactors.posDev.maxValue = config["Pass"]["PosDeviation"]["MaxValue"][3].asFloat();
	breakOffsideFactors.posDev.minWeight = config["Pass"]["PosDeviation"]["MinWeight"][3].asFloat();
	breakOffsideFactors.posDev.maxWeight = config["Pass"]["PosDeviation"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.offense.minValue = config["Pass"]["Offense"]["MinValue"][3].asFloat();
	breakOffsideFactors.offense.maxValue = config["Pass"]["Offense"]["MaxValue"][3].asFloat();
	breakOffsideFactors.offense.minWeight = config["Pass"]["Offense"]["MinWeight"][3].asFloat();
	breakOffsideFactors.offense.maxWeight = config["Pass"]["Offense"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.seeInfo.minValue = config["Pass"]["SeeInformation"]["MinValue"][3].asFloat();
	breakOffsideFactors.seeInfo.maxValue = config["Pass"]["SeeInformation"]["MaxValue"][3].asFloat();
	breakOffsideFactors.seeInfo.minWeight = config["Pass"]["SeeInformation"]["MinWeight"][3].asFloat();
	breakOffsideFactors.seeInfo.maxWeight = config["Pass"]["SeeInformation"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.ballDist.minValue = config["Pass"]["BallDistance"]["MinValue"][3].asFloat();
	breakOffsideFactors.ballDist.maxValue = config["Pass"]["BallDistance"]["MaxValue"][3].asFloat();
	breakOffsideFactors.ballDist.minWeight = config["Pass"]["BallDistance"]["MinWeight"][3].asFloat();
	breakOffsideFactors.ballDist.maxWeight = config["Pass"]["BallDistance"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.dashDist.minValue = config["Pass"]["DashDistance"]["MinValue"][3].asFloat();
	breakOffsideFactors.dashDist.maxValue = config["Pass"]["DashDistance"]["MaxValue"][3].asFloat();
	breakOffsideFactors.dashDist.minWeight = config["Pass"]["DashDistance"]["MinWeight"][3].asFloat();
	breakOffsideFactors.dashDist.maxWeight = config["Pass"]["DashDistance"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.breakOffside.minValue = config["Pass"]["BreakOffsideLine"]["MinValue"][3].asFloat();
	breakOffsideFactors.breakOffside.maxValue = config["Pass"]["BreakOffsideLine"]["MaxValue"][3].asFloat();
	breakOffsideFactors.breakOffside.minWeight = config["Pass"]["BreakOffsideLine"]["MinWeight"][3].asFloat();
	breakOffsideFactors.breakOffside.maxWeight = config["Pass"]["BreakOffsideLine"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.passedOppNum.minValue = config["Pass"]["PassedOppNum"]["MinValue"][3].asFloat();
	breakOffsideFactors.passedOppNum.maxValue = config["Pass"]["PassedOppNum"]["MaxValue"][3].asFloat();
	breakOffsideFactors.passedOppNum.minWeight = config["Pass"]["PassedOppNum"]["MinWeight"][3].asFloat();
	breakOffsideFactors.passedOppNum.maxWeight = config["Pass"]["PassedOppNum"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.statusOppNum.minValue = config["Pass"]["StatusOppNum"]["MinValue"][3].asFloat();
	breakOffsideFactors.statusOppNum.maxValue = config["Pass"]["StatusOppNum"]["MaxValue"][3].asFloat();
	breakOffsideFactors.statusOppNum.minWeight = config["Pass"]["StatusOppNum"]["MinWeight"][3].asFloat();
	breakOffsideFactors.statusOppNum.maxWeight = config["Pass"]["StatusOppNum"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.securityDeltaCycle.minValue = config["Pass"]["SecurityDeltaCycle"]["MinValue"][3].asFloat();
	breakOffsideFactors.securityDeltaCycle.maxValue = config["Pass"]["SecurityDeltaCycle"]["MaxValue"][3].asFloat();
	breakOffsideFactors.securityDeltaCycle.minWeight = config["Pass"]["SecurityDeltaCycle"]["MinWeight"][3].asFloat();
	breakOffsideFactors.securityDeltaCycle.maxWeight = config["Pass"]["SecurityDeltaCycle"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.statusOppDist.minValue = config["Pass"]["StatusOppDistance"]["MinValue"][3].asFloat();
	breakOffsideFactors.statusOppDist.maxValue = config["Pass"]["StatusOppDistance"]["MaxValue"][3].asFloat();
	breakOffsideFactors.statusOppDist.minWeight = config["Pass"]["StatusOppDistance"]["MinWeight"][3].asFloat();
	breakOffsideFactors.statusOppDist.maxWeight = config["Pass"]["StatusOppDistance"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.statusOppDir.minValue = config["Pass"]["StatusOppDirection"]["MinValue"][3].asFloat();
	breakOffsideFactors.statusOppDir.maxValue = config["Pass"]["StatusOppDirection"]["MaxValue"][3].asFloat();
	breakOffsideFactors.statusOppDir.minWeight = config["Pass"]["StatusOppDirection"]["MinWeight"][3].asFloat();
	breakOffsideFactors.statusOppDir.maxWeight = config["Pass"]["StatusOppDirection"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.statusShoot.minValue = config["Pass"]["StatusShoot"]["MinValue"][3].asFloat();
	breakOffsideFactors.statusShoot.maxValue = config["Pass"]["StatusShoot"]["MaxValue"][3].asFloat();
	breakOffsideFactors.statusShoot.minWeight = config["Pass"]["StatusShoot"]["MinWeight"][3].asFloat();
	breakOffsideFactors.statusShoot.maxWeight = config["Pass"]["StatusShoot"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.dribblePath.minValue = config["Pass"]["DribblePath"]["MinValue"][3].asFloat();
	breakOffsideFactors.dribblePath.maxValue = config["Pass"]["DribblePath"]["MaxValue"][3].asFloat();
	breakOffsideFactors.dribblePath.minWeight = config["Pass"]["DribblePath"]["MinWeight"][3].asFloat();
	breakOffsideFactors.dribblePath.maxWeight = config["Pass"]["DribblePath"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.defenseLine.minValue = config["Pass"]["DefenseLine"]["MinValue"][3].asFloat();
	breakOffsideFactors.defenseLine.maxValue = config["Pass"]["DefenseLine"]["MaxValue"][3].asFloat();
	breakOffsideFactors.defenseLine.minWeight = config["Pass"]["DefenseLine"]["MinWeight"][3].asFloat();
	breakOffsideFactors.defenseLine.maxWeight = config["Pass"]["DefenseLine"]["MaxWeight"][3].asFloat();

	breakOffsideFactors.extraCatch.minValue = config["Pass"]["ExtraCatch"]["MinValue"][3].asFloat();
	breakOffsideFactors.extraCatch.maxValue = config["Pass"]["ExtraCatch"]["MaxValue"][3].asFloat();
	breakOffsideFactors.extraCatch.minWeight = config["Pass"]["ExtraCatch"]["MinWeight"][3].asFloat();
	breakOffsideFactors.extraCatch.maxWeight = config["Pass"]["ExtraCatch"]["MaxWeight"][3].asFloat();

}

void Pass::fillOnePassValuesByFactors(const KickInterceptCalculate &interCalc,
									  const PassFactors **factors, float *valuesList[], unsigned num)
{
	float offenseWeight[MAX_FACTORS_NUM];
	float statusOppNumWeight[MAX_FACTORS_NUM];
	float statusOppDistWeight[MAX_FACTORS_NUM];
	float statusOppDirWeight[MAX_FACTORS_NUM];
	float breakOffsideWeight[MAX_FACTORS_NUM];
	float passedOppNumWeight[MAX_FACTORS_NUM];
	float statusShootWeight[MAX_FACTORS_NUM];
	float dribblePathWeight[MAX_FACTORS_NUM];
	float ballDistWeight[MAX_FACTORS_NUM];
	float dashDistWeight[MAX_FACTORS_NUM];
	float posDevWeight[MAX_FACTORS_NUM];
	float seeInfoWeight[MAX_FACTORS_NUM];
	float securityDeltaCycleWeight[MAX_FACTORS_NUM];
	float defenseLineWeight[MAX_FACTORS_NUM];
	float extraCatchWeight[MAX_FACTORS_NUM];
	float planWeight[MAX_FACTORS_NUM];

	const KickInterceptPlayer &checkInterPlayer =
		interCalc.getFastestTmmInterPlayer();

	if (logLevel >= 2)
		LOG << "New pass Player: " << checkInterPlayer.getPlayer().getUniNum()
			<< " - interPoint: " << checkInterPlayer.getInterPoint()
			<< " - ballVel: " << Point(interCalc.getFirstBall().getVel().getMagnitude(),
									   interCalc.getFirstBall().getVel().getDirection())
			<< " ----------------" << endl
			<< "command: " << interCalc.getKickCommand().toString() << endl;

	// calculating quantomic weight
	bool quantomicWeight = true;

	float maxBallVel = reRate(checkInterPlayer.getInterTime(),
							  1, 4, 1, 2.7, RRM_DIRECT);
	if (interCalc.getFirstBall().getVel().getMagnitude() > maxBallVel)
		quantomicWeight = false;

	else if (WorldModel::getOutDistance(
				 checkInterPlayer.getInterPoint()) > 0)
		quantomicWeight = false;

	else if (interCalc.getFirstBall().getDistance(
				 checkInterPlayer.getInterPoint()) < MIN_PASS_DIST)
		quantomicWeight = false;

	else if (checkInterPlayer.getPlayer().getStamina() < 2000)
		quantomicWeight = false;

/*	else if (checkInterPlayer.getPlayer().getPos().getX()
				 > worldModel->getOppOffsideLine() &&
								 checkInterPlayer.getPlayer().getPos().getX() > 0 &&
								 checkInterPlayer.getPlayer().getPos().getX() >
								 worldModel->getBall().getPos().getX())
		quantomicWeight = false;*/

	if (logLevel >= 2)
		LOG << "quantomicWeight: " << quantomicWeight << endl;

	if (!quantomicWeight)
	{
		if (logLevel >= 1)
			LOG << "P" << checkInterPlayer.getPlayer().getUniNum()
				<< "\tPoint: " << checkInterPlayer.getInterPoint()
				<< "\t\tweights:";

		for (unsigned i = 0; i < num; i++)
		{
			*valuesList[i] = -0xFFFF;

			if (logLevel >= 1)
				LOG << " " << *valuesList[i];
		}
		if (logLevel >= 1)
			LOG << endl;

		return;
	}

	// calculating offense weight
	float offenseValue = checkInterPlayer.getInterPoint().x -
		interCalc.getFirstBall().getPos().getX();

	if (logLevel >= 2)
		LOG << "offenseValue: " << offenseValue << endl
			<< "offenseWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		offenseWeight[i] = reRate(offenseValue,
								  factors[i]->offense.minValue, factors[i]->offense.maxValue,
								  factors[i]->offense.minWeight, factors[i]->offense.maxWeight,
								  RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << offenseWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating statusOppNum weight
	float statusOppNumValue = interCalc.getOppCheckedPlayersNum();

	if (logLevel >= 2)
		LOG << "statusOppNumValue: " << statusOppNumValue << endl
			<< "statusOppNumWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		statusOppNumWeight[i] = reRate(statusOppNumValue,
									   factors[i]->statusOppNum.minValue,
									   factors[i]->statusOppNum.maxValue,
									   factors[i]->statusOppNum.minWeight,
									   factors[i]->statusOppNum.maxWeight,
									   RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << statusOppNumWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating statusOppDist weight and statusOppDir weight
	float statusOppDistValue = interCalc.getNearestOppToFastestDist();
	float statusOppDirValue = interCalc.getNearestOppToFastestDir();

	statusOppDistValue += 2.1;

	if (logLevel >= 2)
		LOG << "statusOppDistValue: " << statusOppDistValue << endl
			<< "statusOppDirValue: " << statusOppDirValue << endl
			<< "statusOpp(Dist,Dir)Weight:";

	for (unsigned i = 0; i < num; i++)
	{
		statusOppDistWeight[i] = reRate(statusOppDistValue,
										factors[i]->statusOppDist.minValue,
										factors[i]->statusOppDist.maxValue,
										factors[i]->statusOppDist.minWeight,
										factors[i]->statusOppDist.maxWeight,
										RRM_DIRECT);

		statusOppDirWeight[i] = reRate(statusOppDirValue,
									   factors[i]->statusOppDir.minValue,
									   factors[i]->statusOppDir.maxValue,
									   factors[i]->statusOppDir.minWeight,
									   factors[i]->statusOppDir.maxWeight,
									   RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << Point(statusOppDistWeight[i], statusOppDirWeight[i]);
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating breakOffside weight
	float breakOffsideValue = checkInterPlayer.getInterPoint().x -
		worldModel->getOppOffsideLine();

	if (logLevel >= 2)
		LOG << "breakOffsideValue: " << breakOffsideValue << endl
			<< "breakOffsideWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		breakOffsideWeight[i] = reRate(breakOffsideValue,
									   factors[i]->breakOffside.minValue,
									   factors[i]->breakOffside.maxValue,
									   factors[i]->breakOffside.minWeight,
									   factors[i]->breakOffside.maxWeight,
									   RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << breakOffsideWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating passedOppNum weight
	float passedOppNumValue = interCalc.getPassedOppPlayers();

	if (logLevel >= 2)
		LOG << "passedOppNumValue: " << passedOppNumValue << endl
			<< "passedOppNumWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		passedOppNumWeight[i] = reRate(passedOppNumValue,
									   factors[i]->passedOppNum.minValue,
									   factors[i]->passedOppNum.maxValue,
									   factors[i]->passedOppNum.minWeight,
									   factors[i]->passedOppNum.maxWeight,
									   RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << passedOppNumWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating statusShoot weight
	float statusShootDir = getShootDirValue(
		checkInterPlayer.getInterPoint(), worldModel);
	float statusShootDist = hypot(checkInterPlayer.getInterPoint().x - 52.5,
								  fmin(checkInterPlayer.getInterPoint().y - 4,
									   checkInterPlayer.getInterPoint().y + 4));

	float statusShootValue = statusShootDir -
		fmax(statusShootDist - 10, 0) * 1.15;

	if (logLevel >= 2)
		LOG << "statusShootDir: " << statusShootDir << endl
			<< "statusShootDist: " << statusShootDist << endl
			<< "statusShootValue: " << statusShootValue << endl
			<< "statusShootWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		statusShootWeight[i] = reRate(statusShootValue,
									  factors[i]->statusShoot.minValue,
									  factors[i]->statusShoot.maxValue,
									  factors[i]->statusShoot.minWeight,
									  factors[i]->statusShoot.maxWeight,
									  RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << statusShootWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating dribblePath weight
	dribblePath.setBeginPoint(checkInterPlayer.getInterPoint());
	dribblePath.setDefaultParams();

	float dribblePathValue = dribblePath.getBestTargetWeight();

	if (logLevel >= 2)
		LOG << "dribblePathValue: " << dribblePathValue << endl
			<< "dribblePathWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		dribblePathWeight[i] = 0;

		if (logLevel >= 2)
			LOG << " " << dribblePathWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating posDev weight
	float posDevValue = checkInterPlayer.getPlayer().getPosDeviation() *
		checkInterPlayer.getPlayer().getPosDeviation();

	if (logLevel >= 2)
		LOG << "posDevValue: " << posDevValue << endl
			<< "posDevWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		posDevWeight[i] = reRate(posDevValue,
								 factors[i]->posDev.minValue, factors[i]->posDev.maxValue,
								 factors[i]->posDev.minWeight, factors[i]->posDev.maxWeight,
								 RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << posDevWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating ballDist weight
	float ballDistValue = interCalc.getFirstBall().
		getDistance(checkInterPlayer.getInterPoint());

	if (logLevel >= 2)
		LOG << "ballDistValue: " << ballDistValue << endl
			<< "ballDistWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		ballDistWeight[i] = reRate(ballDistValue,
								   factors[i]->ballDist.minValue, factors[i]->ballDist.maxValue,
								   factors[i]->ballDist.minWeight, factors[i]->ballDist.maxWeight,
								   RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << ballDistWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating dashDist weight
	float dashDistValue = checkInterPlayer.getDashDist();

	if (logLevel >= 2)
		LOG << "dashDistValue: " << dashDistValue << endl
			<< "dashDistWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		dashDistWeight[i] = reRate(dashDistValue, 
								   factors[i]->dashDist.minValue, factors[i]->dashDist.maxValue, 
								   factors[i]->dashDist.minWeight, factors[i]->dashDist.maxWeight, 
								   RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << dashDistWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating securityDeltaCycle weight
	unsigned securityDeltaCycleValue = checkInterPlayer.getInterTime() -
		interCalc.getFastestInterPlayer().getInterTime();

	if (logLevel >= 2)
		LOG << "TarPlyGTBTime: " << checkInterPlayer.getInterTime() << endl
			<< "FastestGTBTime: "
			<< interCalc.getFastestInterPlayer().getInterTime() << endl
			<< "securityDeltaCycleValue: " << securityDeltaCycleValue << endl
			<< "securityDeltaCycleWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		securityDeltaCycleWeight[i] = reRate(securityDeltaCycleValue,
											 factors[i]->securityDeltaCycle.minValue,
											 factors[i]->securityDeltaCycle.maxValue,
											 factors[i]->securityDeltaCycle.minWeight,
											 factors[i]->securityDeltaCycle.maxWeight, RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << securityDeltaCycleWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating seeInfo weight
	float passAngle = interCalc.getFirstBall().getVel().getDirection();
	float avgSeenBefore = worldModel->getSeeDeltaCycle(passAngle, 0.3);

	float seeInfoValue = 0.333 * (avgSeenBefore * avgSeenBefore) +
		avgSeenBefore;

	if (logLevel >= 2)
		LOG << "AvgSeenBefore: " << avgSeenBefore << endl
			<< "SeeInfoValue: " << seeInfoValue << endl
			<< "SeeInfoWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		seeInfoWeight[i] = reRate(seeInfoValue,
								  factors[i]->seeInfo.minValue, factors[i]->seeInfo.maxValue,
								  factors[i]->seeInfo.minWeight, factors[i]->seeInfo.maxWeight,
								  RRM_REVERSE);

		if (logLevel >= 2)
			LOG << " " << seeInfoWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating defenseLine weight
	float defenseLineValue = checkInterPlayer.getInterPoint().x -
		worldModel->getTmmDefenseLine();

	if (logLevel >= 2)
		LOG << "defenseLineValue: " << defenseLineValue << endl
			<< "defenseLineWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		defenseLineWeight[i] = reRate(defenseLineValue,
									  factors[i]->defenseLine.minValue,
									  factors[i]->defenseLine.maxValue,
									  factors[i]->defenseLine.minWeight,
									  factors[i]->defenseLine.maxWeight,
									  RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << defenseLineWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating extraCatch weight
	float extraCatchValue = 0;

	if (logLevel >= 2)
		LOG << "extraCatchValue: " << extraCatchValue << endl
			<< "extraCatchWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		extraCatchWeight[i] = reRate(extraCatchValue,
									 factors[i]->extraCatch.minValue,
									 factors[i]->extraCatch.maxValue,
									 factors[i]->extraCatch.minWeight,
									 factors[i]->extraCatch.maxWeight,
									 RRM_DIRECT);

		if (logLevel >= 2)
			LOG << " " << extraCatchWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating plan weight
	if (logLevel >= 2)
		LOG << "planInfo: " << interCalc.getKickPlayer()->getUniNum()
			<< " --> " << checkInterPlayer.getPlayer().getUniNum() << endl
			<< "planWeight:";

	for (unsigned i = 0; i < num; i++)
	{
		planWeight[i] = getPlanWeight(interCalc.getKickPlayer()->getUniNum(),
									  checkInterPlayer.getPlayer().getUniNum(),
									  interCalc.getFirstBall().getPos());

		if (logLevel >= 2)
			LOG << " " << planWeight[i];
	}

	if (logLevel >= 2)
		LOG << endl;

	// calculating Current weight
	if (logLevel >= 1)
		LOG << "P" << checkInterPlayer.getPlayer().getUniNum()
			<< "\tPoint: " << checkInterPlayer.getInterPoint()
			<< "\t\tweights:";

	for (unsigned i = 0; i < num; i++)
	{
		*valuesList[i] =
			offenseWeight[i] +
			statusShootWeight[i] +
			statusOppNumWeight[i] +
			statusOppDistWeight[i] +
			statusOppDirWeight[i] +
			breakOffsideWeight[i] +
			passedOppNumWeight[i] +
			statusShootWeight[i] +
			dribblePathWeight[i] +
			ballDistWeight[i] +
			dashDistWeight[i] +
			posDevWeight[i] +
			seeInfoWeight[i] +
			securityDeltaCycleWeight[i] +
			defenseLineWeight[i] +
			extraCatchWeight[i] +
			planWeight[i];

		if (logLevel >= 1)
			LOG << "\t" << *valuesList[i];
	}

	if (logLevel >= 1)
		LOG << endl;
}

void
Pass::fillPassValuesByFactors(const KickCalculate &kickCalc, const PassFactors **factors,
							  float *valuesList[], unsigned num)
{
	float *values[MAX_FACTORS_NUM];

	for (unsigned i = 0; i < kickCalc.getSimulatesNum(); i++)
	{
		if (logLevel >= 3)
			LOG << "New kick: ballVel: "
				<< Point(kickCalc.getSimulate(i).getFirstBall().getVel().getMagnitude(),
						 kickCalc.getSimulate(i).getFirstBall().getVel().getDirection()) << endl;

		if (!kickCalc.getSimulate(i).isValid())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: is not valid." << endl;
			continue;
		}

		if (!kickCalc.getSimulate(i).existFastestPlayer())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: no fastest." << endl;
			continue;
		}

		if (!kickCalc.getSimulate(i).existFastestTmmPlayer())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: opp fastest: "
					<< kickCalc.getSimulate(i).getFastestPlayer().getTeamId() << " "
					<< kickCalc.getSimulate(i).getFastestPlayer().getUniNum() << endl;
			continue;
		}

		const KickInterceptPlayer &checkInterPlayer =
			kickCalc.getSimulate(i).getFastestTmmInterPlayer();

		if (logLevel >= 2)
			LOG << "CheckPlayer: " << checkInterPlayer.getPlayer().getTeamId()
				<< " " << checkInterPlayer.getPlayer().getUniNum() << endl;

		if (&checkInterPlayer.getPlayer() == 
			kickCalc.getSimulate(i).getKickPlayer())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: is kickPlayer." << endl;
			continue;
		}

		if (checkInterPlayer.getPlayer().isGoalie())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: is goalie." << endl;
			continue;
		}

		if (!checkInterPlayer.isExtraCatched())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: not extra catched." << endl;
			continue;
		}

/*		if (checkInterPlayer.getPlayer().getPos().getX() >
				worldModel->getOppOffsideLine() + PASS_OFFSIDE_DELTA_DIST)
		{
				if (logLevel >= 3)
				LOG << "\tBREAK: offside." << endl;
				continue;
		}*/

		for (unsigned j = 0; j < num; j++)
			values[j] = &valuesList[j][i];

		fillOnePassValuesByFactors(kickCalc.getSimulate(i),
								   factors, values, num);
	}
}

void Pass::fillOtherCrossValuesByFactors(const CrossCalculate &crossCalc,
										 const PassFactors &factors, float values[])
{
	float *valuesList[1];
	const PassFactors *factorsList[1] = {&factors};

	if (logLevel >= 1)
		LOG << "Cross Special Values:" << endl;

	for (unsigned i = 0; i < crossCalc.getSimulatesNum(); i++)
	{
		if (logLevel >= 3)
			LOG << "New cross kick: ballVel: "
				<< Point(crossCalc.getSimulate(i).getFirstBall().getVel().getMagnitude(),
						 crossCalc.getSimulate(i).getFirstBall().getVel().getDirection()) << endl;
		if (!crossCalc.getSimulate(i).isValid())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: is not valid." << endl;
			continue;
		}
		if (!crossCalc.getSimulate(i).existFastestTmmPlayer())
		{
			if (logLevel >= 3)
				LOG << "\tBREAK: no fastest." << endl;
			continue;
		}

		const KickInterceptPlayer &checkInterPlayer =
			crossCalc.getSimulate(i).getFastestTmmInterPlayer();

		if (logLevel >= 2)
			LOG << "CheckPlayer: " << checkInterPlayer.getPlayer().getTeamId()
				<< " " << checkInterPlayer.getPlayer().getUniNum() << endl;

		if (checkInterPlayer.getPlayer().isBody())
			continue;
		if (checkInterPlayer.getPlayer().isGoalie())
			continue;
		if (checkInterPlayer.getPlayer().getPos().getX() >
			worldModel->getOppOffsideLine() + PASS_OFFSIDE_DELTA_DIST)
			continue;

		valuesList[0] = &values[i + KICK_CALCULATES_NUM];
		fillOnePassValuesByFactors(crossCalc.getSimulate(i),
								   factorsList, valuesList);
	}
}

void Pass::calculateGlobalVars()
{
	logLevel = config["Pass"]["GlobalVars"]["LogLevel"].asInt();

	crossRate = config["Pass"]["GlobalVars"]["CrossRate"].asFloat();
	crossAlpha = config["Pass"]["GlobalVars"]["CrossAlpha"].asFloat();
	offenseRate = config["Pass"]["GlobalVars"]["OffenseRate"].asFloat();
	offenseAlpha = config["Pass"]["GlobalVars"]["OffenseAlpha"].asFloat();
	secureBallRate = config["Pass"]["GlobalVars"]["SecureBallRate"].asFloat();
	secureBallAlpha = config["Pass"]["GlobalVars"]["SecureBallAlpha"].asFloat();
	breakOffsideRate = config["Pass"]["GlobalVars"]["BreakOffsideRate"].asFloat();
	breakOffsideAlpha = config["Pass"]["GlobalVars"]["BreakOffsideAlpha"].asFloat();

	if (worldModel->getBall().getPos().getX() < CROSS_PERMITTED_X)
		crossRate = 0;
}

void Pass::fillMaxValues()
{
	offenseMaxer = NOVALUE;
	secureBallMaxer = NOVALUE;
	breakOffsideMaxer = NOVALUE;
	crossMaxer = NOVALUE;

	if (offenseRate)
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (offenseMaxer == NOVALUE ||
				offenseValues[i] > offenseValues[offenseMaxer])
				offenseMaxer = i;

	if (secureBallRate)
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (secureBallMaxer == NOVALUE ||
				secureBallValues[i] > secureBallValues[secureBallMaxer])
				secureBallMaxer = i;

	if (breakOffsideRate)
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (breakOffsideMaxer == NOVALUE ||
				breakOffsideValues[i] > breakOffsideValues[breakOffsideMaxer])
				breakOffsideMaxer = i;

	if (crossRate)
		for (unsigned i = 0; i < KICK_CALCULATES_NUM + CROSS_CALCULATES_NUM; i++)
			if (crossMaxer == NOVALUE ||
				crossValues[i] > crossValues[crossMaxer])
				crossMaxer = i;

	LOG << "\t\t\t\tOffense SecureBall BreakOffside Cross" << endl;
	LOG << "Max Values:\t"
		<< "\t" << (offenseMaxer != NOVALUE ? offenseValues[offenseMaxer] : -0xFFFF)
		<< "\t" << (secureBallMaxer != NOVALUE ? secureBallValues[secureBallMaxer] : -0xFFFF)
		<< "\t" << (breakOffsideMaxer != NOVALUE ? breakOffsideValues[breakOffsideMaxer] : -0xFFFF)
		<< "\t" << (crossMaxer != NOVALUE ? crossValues[crossMaxer]  : -0xFFFF) << endl;
}

void Pass::fillValues()
{
	float *values[MAX_FACTORS_NUM];
	const PassFactors *factors[MAX_FACTORS_NUM];
	unsigned num = 0;

	if (logLevel >= 1)
		LOG << "Pass weights sequence:";

	if (crossRate)
	{
		values[num] = crossValues;
		factors[num] = &crossFactors;
		num++;

		if (logLevel >= 1)
			LOG << " Cross";
	}

	if (offenseRate)
	{
		values[num] = offenseValues;
		factors[num] = &offenseFactors;
		num++;

		if (logLevel >= 1)
			LOG << " Offense";
	}

	if (secureBallRate)
	{
		values[num] = secureBallValues;
		factors[num] = &secureBallFactors;
		num++;

		if (logLevel >= 1)
			LOG << " SecureBall";
	}

	if (breakOffsideRate)
	{
		values[num] = breakOffsideValues;
		factors[num] = &breakOffsideFactors;
		num++;

		if (logLevel >= 1)
			LOG << " BreakOffside";
	}

	if (logLevel >= 1)
		LOG << endl;

	fillPassValuesByFactors(*curKickCalculate,
							factors, values, num);

	if (crossRate)
		fillOtherCrossValuesByFactors(*crossCalculate,
									  crossFactors, crossValues);
}

void Pass::resetValues()
{
	for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
	{
		crossValues[i] = -0xFFFF;
		offenseValues[i] = -0xFFFF;
		secureBallValues[i] = -0xFFFF;
		breakOffsideValues[i] = -0xFFFF;
	}

	for (unsigned i = 0; i < CROSS_CALCULATES_NUM; i++)
		crossValues[i + KICK_CALCULATES_NUM] = -0xFFFF;
}

void Pass::applyNoiseValues()
{
/*	unsigned neighbour1, neighbour2;
	float tempCrossValues[KICK_CALCULATES_NUM];
	float tempOffenseValues[KICK_CALCULATES_NUM];
	float tempSecureBallValues[KICK_CALCULATES_NUM];
	float tempBreakOffsideValues[KICK_CALCULATES_NUM];

	if (logLevel >= 2)
		LOG << "Pass::applyNoiseValues" << endl
		<< "   "
		<< "\t\t" << "cross"
		<< "\t" << "offense"
		<< "\t" << "secureBall"
		<< "\t" << "breakOffside" << endl;

	for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
	{
	/u/u	worldModel->getCurKickCalculate().
		getNeighboursByDir(i, neighbour1, neighbour2);

		tempCrossValues[i] = ((crossValues[i] * 3) +
		(crossValues[neighbour1] * 1) +
		(crossValues[neighbour2] * 1)) / 5;

		tempOffenseValues[i] = ((offenseValues[i] * 3) +
		(offenseValues[neighbour1] * 1) +
		(offenseValues[neighbour2] * 1)) / 5;

		tempSecureBallValues[i] = ((secureBallValues[i] * 3) +
		(secureBallValues[neighbour1] * 1) +
		(secureBallValues[neighbour2] * 1)) / 5;

		tempBreakOffsideValues[i] = ((breakOffsideValues[i] * 3) +
		(breakOffsideValues[neighbour1] * 1) +
		(breakOffsideValues[neighbour2] * 1)) / 5;

	}

	for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
	{
		crossValues[i] = tempCrossValues[i];
		offenseValues[i] = tempOffenseValues[i];
		secureBallValues[i] = tempSecureBallValues[i];
		breakOffsideValues[i] = tempBreakOffsideValues[i];

		if (logLevel >= 2)
		LOG << "i >"
		<< "\t\t" << crossValues[i]
		<< "\t" << offenseValues[i]
		<< "\t" << secureBallValues[i]
		<< "\t" << breakOffsideValues[i] << endl;
	}*/
}

float Pass::getTheBestPass(const KickInterceptCalculate *&interCalc)
{
	float offenseWeight = -0xFFFF;
	if (offenseRate && offenseValues[offenseMaxer] != -0xFFFF)
		offenseWeight = offenseValues[offenseMaxer] *
			offenseRate + offenseAlpha;

	float secureBallWeight = -0xFFFF;
	if (secureBallRate && secureBallValues[secureBallMaxer] != -0xFFFF)
		secureBallWeight = secureBallValues[secureBallMaxer] *
			secureBallRate + secureBallAlpha;

	float breakOffsideWeight = -0xFFFF;
	if (breakOffsideRate && breakOffsideValues[breakOffsideMaxer] != -0xFFFF)
		breakOffsideWeight = breakOffsideValues[breakOffsideMaxer] *
			breakOffsideRate + breakOffsideAlpha;

	float crossWeight = -0xFFFF;
	if (crossRate && crossValues[crossMaxer] != -0xFFFF)
		crossWeight = crossValues[crossMaxer] *
			crossRate + crossAlpha;

	LOG << "Max Weights:"
		<< "\t" << offenseWeight
		<< "\t" << secureBallWeight
		<< "\t" << breakOffsideWeight
		<< "\t" << crossWeight << endl;

	float max = offenseWeight;
	unsigned maxer = 1;

	if (secureBallWeight > max)
	{
		maxer = 2;
		max = secureBallWeight;
	}

	if (breakOffsideWeight > max)
	{
		maxer = 3;
		max = breakOffsideWeight;
	}

	if (crossWeight > max)
	{
		maxer = 4;
		max = crossWeight;
	}

	if (max == -0xFFFF)
	{
		interCalc = NULL;
		return -0xFFFF;
	}

	switch (maxer)
	{
	case 1:
			interCalc = &curKickCalculate->getSimulate(offenseMaxer);
			passModel = PAM_OFFENSE;
			break;

	case 2:
			interCalc = &curKickCalculate->getSimulate(secureBallMaxer);
			passModel = PAM_SECURE_BALL;
			break;

	case 3:
			interCalc = &curKickCalculate->getSimulate(breakOffsideMaxer);
			passModel = PAM_BREAK_OFFSIDE;
			break;

	case 4:
			if (crossMaxer < KICK_CALCULATES_NUM)
			{
				interCalc = &curKickCalculate->getSimulate(crossMaxer);
				LOG << "Cross pass executed." << endl;
			}
			else
			{
				interCalc = &crossCalculate->
					getSimulate(crossMaxer - KICK_CALCULATES_NUM);
				LOG << "Cross special pass executed." << endl;
			}
			passModel = PAM_CROSS;
			break;

	default:
			assert(0);
	}

	return max;
}

float Pass::getShootDirValue(const Point &point, const WorldModel *worldModel)
{
	Vector goalUpVector, goalDownVector, goalieVector;
	const Player *oppGoalie = worldModel->getOppGoalie();

	goalUpVector.setByPoints(point, Point(52.5, 7));
	goalDownVector.setByPoints(point, Point(52.5, -7));

	if (oppGoalie == NULL)
		return fabs(Degree::getDeltaAngle(goalDownVector.getDirection(),
										  goalUpVector.getDirection()));

	goalieVector.setByPoints(point, oppGoalie->getPos());

	float upDeltaAngle = Degree::getDeltaAngle(
		goalieVector.getDirection(), goalUpVector.getDirection());
	float downDeltaAngle = Degree::getDeltaAngle(
		goalieVector.getDirection(), goalDownVector.getDirection());

	if ((upDeltaAngle > 0 && downDeltaAngle > 0) ||
		(upDeltaAngle < 0 && downDeltaAngle < 0))
		return fabs(Degree::getDeltaAngle(goalDownVector.getDirection(),
										  goalUpVector.getDirection()));

	return fmax(fabs(upDeltaAngle), fabs(downDeltaAngle));
}

float Pass::getPlanWeight(unsigned kickerNum, unsigned passNum,
						  const Point &kickPoint)
{
	if (kickerNum == 5)
	{
		if (6 <= passNum && passNum <= 9)
			return PW_VERY_BAD;
		if (passNum == 10 || passNum == 11)
			return PW_VERY_GOOD;
		if (passNum <= 4)
			return PW_GOOD;
	}

	if (6 <= kickerNum && kickerNum <= 9)
	{
		if (passNum == 5)
			return PW_NOTHING;
		if (6 <= passNum && passNum <= 9)
			return PW_VERY_BAD;
		if (passNum == 10 || passNum == 11)
			return PW_VERY_GOOD;
		if (passNum <= 4)
			return PW_GOOD;
	}

/*	if (kickerNum == 2)
	{
		if (passNum == 10)
		{
			if (kickPoint.x > 35)
				return PW_NOTHING;
			else
				return PW_VERY_GOOD;
		}
		if (passNum == 3 || passNum == 4)
			return PW_GOOD;
	}

	if (kickerNum == 4)
	{
		if (passNum == 11)
		{
			if (kickPoint.x > 35)
				return PW_NOTHING;
			else
				return PW_VERY_GOOD;
		}
		if (passNum == 3 || passNum == 2)
			return PW_GOOD;
	}

	if (kickerNum == 3)
	{
		if (passNum == 2 || passNum == 4)
			return PW_GOOD;
		if (passNum == 10 || passNum == 11)
		{
			if (kickPoint.x > 35)
				return PW_NOTHING;
			else
				return PW_GOOD;
		}
	}

	if (kickerNum == 10 || kickerNum == 11)
	{
		if (kickPoint.x > 37.5)
			return PW_GOOD;
		else
			return PW_BAD;
	}*/

	if (isInPlayers(worldModel->getBody().getUniNum(), "234AB"))
		return PW_GOOD;

	return 0;
}

// getting functions
float Pass::getPassValue() const
{
	return passValue;
}

PassModel Pass::getPassModel() const
{
	return passModel;
}

const Player *Pass::getPassPlayer() const
{
	return passPlayer;
}

const KickInterceptCalculate *Pass::getPassInterCalc() const
{
	return passInterCalc;
}

// setting functions
void
Pass::setCurKickCalculate(const KickCalculate *curKickCalculateArg)
{
	curKickCalculate = curKickCalculateArg;
}


// class PassLibrary
PassLibrary::PassLibrary(const WorldModel *worldModel): Pass(worldModel)
{
}

void PassLibrary::calculate()
{
	LOG << "PassLibrary::calculate" << endl;

	delete passCommand;
	passCommand = NULL;

	if (logLevel >= 1)
		loadPassFactors();

	calculateGlobalVars();

	if (crossRate)
		crossCalculate->update(worldModel);

	resetValues();
	fillValues();
}

ExactPassFlag PassLibrary::getTheBestPass(unsigned models, unsigned number,
		float passAngleFrom, float passAngleTo,
		float playerAngleFrom, float playerAngleTo) const
{
	LOG << "PassLibrary::getTheBestPass" << endl;

	ExactPassFlag flag;

	unsigned offenseMaxer = NOVALUE;
	unsigned secureBallMaxer = NOVALUE;
	unsigned breakOffsideMaxer = NOVALUE;
	unsigned crossMaxer = NOVALUE;

	passAngleFrom = Degree::normalizeAngle(passAngleFrom);
	passAngleTo = Degree::normalizeAngle(passAngleTo);

	if (offenseRate && (models & BPM_OFFENSE))
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (offenseValues[i] != -0xFFFF &&
				(number == BPN_ALL ||
				 (number > 11 && // Half players are also available.
				  curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getModel() == PLM_HALF) ||
				 (curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getUniNum() == Basics::baghimande(number - 1, 11) + 1)) &&
				(curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() >= passAngleFrom &&
				 curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() <= passAngleTo) &&
				(curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() >= playerAngleFrom &&
				 curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() <= playerAngleTo) &&
				(offenseMaxer == NOVALUE ||
				 offenseValues[i] > offenseValues[offenseMaxer]))
				offenseMaxer = i;

	if (secureBallRate && (models & BPM_SECUREBALL))
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (secureBallValues[i] != -0xFFFF &&
				(number == BPN_ALL ||
				 (number > 11 && // Half players are also available.
				  curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getModel() == PLM_HALF) ||
				 (curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getUniNum() == Basics::baghimande(number - 1, 11) + 1)) &&
				(curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() >= passAngleFrom &&
				 curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() <= passAngleTo) &&
				(curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() >= playerAngleFrom &&
				 curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() <= playerAngleTo) &&
				(secureBallMaxer == NOVALUE ||
				 secureBallValues[i] > secureBallValues[secureBallMaxer]))
				secureBallMaxer = i;

	if (breakOffsideRate && (models & BPM_BREAKOFFSIDE))
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (breakOffsideValues[i] != -0xFFFF &&
				(number == BPN_ALL ||
				 (number > 11 && // Half players are also available.
				  curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getModel() == PLM_HALF) ||
				  (curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getUniNum() == Basics::baghimande(number - 1, 11) + 1)) &&
				(curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() >= passAngleFrom &&
				 curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() <= passAngleTo) &&
				(curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() >= playerAngleFrom &&
				 curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() <= playerAngleTo) &&
				(breakOffsideMaxer == NOVALUE ||
				 breakOffsideValues[i] > breakOffsideValues[breakOffsideMaxer]))
				breakOffsideMaxer = i;

	if (crossRate && (models & BPM_CROSS))
	{
		for (unsigned i = 0; i < KICK_CALCULATES_NUM; i++)
			if (crossValues[i] != -0xFFFF &&
				(number == BPN_ALL ||
				 (number > 11 && // Half players are also available.
				  curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getModel() == PLM_HALF) ||
				 (curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
				 		getPlayer().getUniNum() == Basics::baghimande(number - 1, 11) + 1)) &&
				(curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() >= passAngleFrom &&
				 curKickCalculate->getSimulate(i).getFirstBall().
						getVel().getDirection() <= passAngleTo) &&
				(curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() >= playerAngleFrom &&
				 curKickCalculate->getSimulate(i).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() <= playerAngleTo) &&
				(crossMaxer == NOVALUE ||
				 crossValues[i] > crossValues[crossMaxer]))
				crossMaxer = i;

		for (unsigned i = KICK_CALCULATES_NUM;
				i < KICK_CALCULATES_NUM + CROSS_CALCULATES_NUM; i++)
			if (crossValues[i] != -0xFFFF &&
				(number == BPN_ALL ||
				 (number > 11 && // Half players are also available.
				  crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).getFastestTmmInterPlayer().
				 		getPlayer().getModel() == PLM_HALF) ||
				 (crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).getFastestTmmInterPlayer().
				 		getPlayer().getUniNum() == Basics::baghimande(number - 1, 11) + 1)) &&
				(crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).
						getFirstBall().getVel().getDirection() >= passAngleFrom &&
				 crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).
				 		getFirstBall().getVel().getDirection() <= passAngleTo) &&
				(crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() >= playerAngleFrom &&
				 crossCalculate->getSimulate(i - KICK_CALCULATES_NUM).getFastestTmmInterPlayer().
						getPlayer().getAbsVec().getDirection() <= playerAngleTo) &&
				(crossMaxer == NOVALUE ||
				 crossValues[i] > crossValues[crossMaxer]))
				crossMaxer = i;
	}

	bool haveSuggested = false;
	if (worldModel->getHearedSuggestPassTime() == worldModel->getCurTime() &&
		(models & BPM_SUGGESTED))
	{
		LOG << "Suggested pass:" << endl
			<< "\tPlayer: " << worldModel->getHearedSuggestPassNum() << endl
			<< "\tWeight: " << worldModel->getHearedSuggestPassWeight() << endl
			<< "\tSpeed: " << worldModel->getHearedSuggestPassSpeed() << endl
			<< "\tDir: " << worldModel->getHearedSuggestPassDir() << endl;

		KickToDirectionVelocity basicKick(AT_PASS,
				worldModel->getHearedSuggestPassDir(),
				worldModel->getHearedSuggestPassSpeed(),
				worldModel->getBall(), worldModel->getBody());

		if (!basicKick.isPossible())
		{
			LOG << "Suggested pass refused: kick is not possible." << endl;
			haveSuggested = false;
		}
		else
		{
			LOG << "Suggested pass accepted." << endl;
			haveSuggested = true;
		}
	}

	LOG << "\t\t\t\tOffense SecureBall BreakOffside Cross Suggested" << endl;
	LOG << "Max Values:\t"
		<< "\t" << (offenseMaxer != NOVALUE ? offenseValues[offenseMaxer] : -0xFFFF)
		<< "\t" << (secureBallMaxer != NOVALUE ? secureBallValues[secureBallMaxer] : -0xFFFF)
		<< "\t" << (breakOffsideMaxer != NOVALUE ? breakOffsideValues[breakOffsideMaxer] : -0xFFFF)
		<< "\t" << (crossMaxer != NOVALUE ? crossValues[crossMaxer] : -0xFFFF)
		<< "\t" << (haveSuggested ? "YES" : "NO") << endl;

	float offenseWeight = -0xFFFF;
	if (offenseRate && offenseMaxer != NOVALUE &&
			offenseValues[offenseMaxer] != -0xFFFF)
		offenseWeight = offenseValues[offenseMaxer] *
			offenseRate + offenseAlpha;

	float secureBallWeight = -0xFFFF;
	if (secureBallRate && secureBallMaxer != NOVALUE &&
			secureBallValues[secureBallMaxer] != -0xFFFF)
		secureBallWeight = secureBallValues[secureBallMaxer] *
			secureBallRate + secureBallAlpha;

	float breakOffsideWeight = -0xFFFF;
	if (breakOffsideRate && breakOffsideMaxer != NOVALUE &&
			breakOffsideValues[breakOffsideMaxer] != -0xFFFF)
		breakOffsideWeight = breakOffsideValues[breakOffsideMaxer] *
			breakOffsideRate + breakOffsideAlpha;

	float crossWeight = -0xFFFF;
	if (crossRate && crossMaxer != NOVALUE &&
			crossValues[crossMaxer] != -0xFFFF)
		crossWeight = crossValues[crossMaxer] *
			crossRate + crossAlpha;

	float suggestedWeight = -0xFFFF;
	if (haveSuggested)
		suggestedWeight = worldModel->getHearedSuggestPassWeight() +
				SUGGEST_REAL_DELTA_WEIGHT;

	LOG << "Max Weights:"
		<< "\t" << offenseWeight
		<< "\t" << secureBallWeight
		<< "\t" << breakOffsideWeight
		<< "\t" << crossWeight
		<< "\t" << suggestedWeight << endl;

	float max = offenseWeight;
	unsigned maxer = 1;

	if (secureBallWeight > max)
	{
		maxer = 2;
		max = secureBallWeight;
	}

	if (breakOffsideWeight > max)
	{
		maxer = 3;
		max = breakOffsideWeight;
	}

	if (crossWeight > max)
	{
		maxer = 4;
		max = crossWeight;
	}

	if (suggestedWeight > max)
	{
		maxer = 5;
		max = suggestedWeight;
	}

	if (max == -0xFFFF)
	{
		flag.weight = 0;
		flag.simulateID = NOVALUE;
		flag.model = PAM_NONE;
		flag.player = NULL;
		flag.interCalc = NULL;

		return flag;
	}

	switch (maxer)
	{
	case 1:
		flag.weight = max;
		flag.simulateID = offenseMaxer;
		flag.model = PAM_OFFENSE;
		flag.player = &curKickCalculate->getSimulate(offenseMaxer).
				getFastestTmmInterPlayer().getPlayer();
		flag.interCalc = &curKickCalculate->getSimulate(offenseMaxer);
		break;

	case 2:
		flag.weight = max;
		flag.simulateID = secureBallMaxer;
		flag.model = PAM_SECURE_BALL;
		flag.player = &curKickCalculate->getSimulate(secureBallMaxer).
				getFastestTmmInterPlayer().getPlayer();
		flag.interCalc = &curKickCalculate->getSimulate(secureBallMaxer);
		break;

	case 3:
		flag.weight = max;
		flag.simulateID = breakOffsideMaxer;
		flag.model = PAM_BREAK_OFFSIDE;
		flag.player = &curKickCalculate->getSimulate(breakOffsideMaxer).
				getFastestTmmInterPlayer().getPlayer();
		flag.interCalc = &curKickCalculate->getSimulate(breakOffsideMaxer);
		break;

	case 4:
		flag.weight = max;
		flag.simulateID = crossMaxer;
		flag.model = PAM_CROSS;

		if (crossMaxer < KICK_CALCULATES_NUM)
		{
			flag.player = &curKickCalculate->getSimulate(crossMaxer).
					getFastestTmmInterPlayer().getPlayer();
			flag.interCalc = &curKickCalculate->getSimulate(crossMaxer);
//			LOG << "Cross pass executed." << endl;
		}
		else
		{
			flag.player = &crossCalculate->getSimulate(crossMaxer - KICK_CALCULATES_NUM).
					getFastestTmmInterPlayer().getPlayer();
			flag.interCalc = &crossCalculate->getSimulate(crossMaxer - KICK_CALCULATES_NUM);
//			LOG << "Cross special pass executed." << endl;
		}

		break;

	case 5:
		flag.weight = max;
		flag.simulateID = NOVALUE;
		flag.model = PAM_SUGGESTED;

		if (worldModel->getHearedSuggestPassNum() != 0)
			flag.player = &worldModel->getFullPlayer(TID_TEAMMATE,
								worldModel->getHearedSuggestPassNum() - 1);
		else
			flag.player = &worldModel->getHalfPlayer(TID_TEAMMATE, 0);

		flag.interCalc = NULL;
		break;

	default:
		assert(0);
	}

	return flag;
}

Command* PassLibrary::execute(const ExactPassFlag passLib,
		Form &form, const Library &library) const
{
	LOG << "PassLibrary::execute" << endl;

	string passModelStr;

	form.sayForm.interceptNum = 0;
	form.sayForm.beforePassSay = true;

	if (passLib.player != NULL)
	{
		switch (passLib.model)
		{
			case PAM_NONE:
				passModelStr = "None";
				break;
			case PAM_CROSS:
				passModelStr = "Cross";
				break;
			case PAM_OFFENSE:
				passModelStr = "Offense";
				break;
			case PAM_SECURE_BALL:
				passModelStr = "SecureBall";
				break;
			case PAM_BREAK_OFFSIDE:
				passModelStr = "BreakOffside";
				break;
			case PAM_SUGGESTED:
				passModelStr = "Suggested";
				break;
			case PAM_KICK_OFF:
				passModelStr = "KickOff";
				break;

			default:
				assert(0);
		}

		if (passLib.player->getModel() == PLM_FULL)
		{
			form.sayForm.interceptNum = passLib.player->getUniNum();
			LOG << passModelStr << " pass to player: "
				<< passLib.player->getUniNum() << endl;
		}
		else
			LOG << passModelStr << " pass to player: "
				<< "<HalfPlayer>" << endl;
	}
	else
		LOG << "*** nobody has specified but Pass is executed ***" << endl;

	Command *retCommand;
	if (passLib.model != PAM_SUGGESTED)
		retCommand = new KickCommand(passLib.interCalc->getKickCommand());
	else
	{
		KickToDirectionVelocity basicKick(AT_PASS,
			worldModel->getHearedSuggestPassDir(),
			worldModel->getHearedSuggestPassSpeed(),
			worldModel->getBall(), worldModel->getBody());

		retCommand = dynamic_cast<KickCommand *>(basicKick.getCommand());
	}

	retCommand->setCreator(AT_PASS);

	if (logLevel >= 4)
	{
		// Debuging GoToBallCalculate before simulating ball for next cycle.
		InterceptCalculate interCalc;
		Ball ball = worldModel->getBall();
		ball.simulateByAction(worldModel->getBody(), retCommand);
//		ball.simulateByDynamics(worldModel->getBody());

		interCalc.startSession(1, &worldModel->getBody());
		interCalc.setLogging(true);
		interCalc.updateByWorldModel(worldModel);
		interCalc.setVirtualBall(ball);
		interCalc.run();
	}

	return retCommand;
}

