/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Mohammad Salehe
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cassert>
#include <cmath>
#include <iostream>

#include <OppTypes.h>
#include <Logger.h>
#include <Degree.h>
#include <Basics.h>

using namespace std;

#define OPPTYPES_LOG1
//#define OPPTYPES_LOG3
//#define OPPTYPES_LOG3
#define PLAYER_EXTRA_RAND_TURN 0.02
#define PLAYER_EXTRA_RAND_DASH 0.21
#define PLAYER_EXTRA_RAND_KICK 0.0145
// Class TypeInfo

void TypeInfo::setByPlayerType(const Param &playerParam)
{
	playerSpeedMax = playerParam["player_speed_max"].asFloat();
	staminaIncMax = playerParam["stamina_inc_max"].asFloat();
	playerDecay = playerParam["player_decay"].asFloat();
	inertiaMoment = playerParam["inertia_moment"].asFloat();
	dashPowerRate = playerParam["dash_power_rate"].asFloat();
	playerSize = playerParam["player_size"].asFloat();
	kickableMargin = playerParam["kickable_margin"].asFloat();
	kickRand = playerParam["kick_rand"].asFloat();
	extraStamina = playerParam["extra_stamina"].asFloat();
	effortMax = playerParam["effort_max"].asFloat();
	effortMin = playerParam["effort_min"].asFloat();
}

// playerPos,ballPos,playerBodyDir are for the kick cycle.
int TypeInfo::isAbleToKick(Vector playerPos, Vector ballPos,
	   	float playerBodyDir, Vector accleration, Side ourSide,
		const Param &serverParam)
{
	float rate;
	float needPower;
	float randFactor;
	float actualKickRand;
	Vector bodyVector;

#ifdef OPPTYPES_LOG3
	LOG << " is able to kick acc=" << accleration << " bd=" << playerBodyDir << endl;
#endif	

	bodyVector = ballPos - playerPos;
	
#ifdef OPPTYPES_LOG3
	LOG << "                 bodyVector=" << bodyVector << endl;
#endif	

	bodyVector.rotate(-playerBodyDir);
	rate = (1 - 0.25 * (bodyVector.getMagnitude() - playerSize -
				serverParam["ball_size"].asFloat()) / kickableMargin -
				0.25 * (fabs(bodyVector.getDirection()) / 180));
	if (rate < 0.5)
		return 0;

#ifdef OPPTYPES_LOG3
	LOG << "                 rate=" << rate << endl;
#endif	

    needPower = (accleration.getMagnitude() / rate) / 
		serverParam["kick_power_rate"].asFloat();

#ifdef OPPTYPES_LOG3
	LOG << "                 needPower=" << needPower << endl;
#endif	

	if (needPower <= serverParam["maxpower"].asFloat() &&
		needPower >= serverParam["minpower"].asFloat())
		return 2;

	if (ourSide == SI_LEFT)
		actualKickRand = serverParam["kick_rand_factor_r"].asFloat() * kickRand;
	else
		actualKickRand = serverParam["kick_rand_factor_l"].asFloat() * kickRand;
#ifdef OPPTYPES_LOG3
	LOG << "  actualKickRand=" << actualKickRand << endl;
#endif	

	randFactor = (accleration.getMagnitude() - M_SQRT2 * actualKickRand) /
	   accleration.getMagnitude();
#ifdef OPPTYPES_LOG3
	LOG << "  randFactor=" << randFactor << endl;
#endif	

	if (randFactor < 0) 
		randFactor = 1;
	accleration *= randFactor;
#ifdef OPPTYPES_LOG3
	LOG << "  newacc=" << accleration << endl;
#endif	

    needPower = (accleration.getMagnitude() / rate) / 
		serverParam["kick_power_rate"].asFloat();
#ifdef OPPTYPES_LOG3
	LOG << "              newneedPower=" << needPower << endl;
#endif	
	
	if (needPower <= serverParam["maxpower"].asFloat() *
			(1 + PLAYER_EXTRA_RAND_KICK)  &&
		needPower >= serverParam["minpower"].asFloat() *
			(1 + PLAYER_EXTRA_RAND_KICK))
		return 1;

	return 0;
}
		
// playerBodyDir is for the dash cycle.
int TypeInfo::isAbleToDash(Vector playerVel, Vector playerPrevVel,
		float playerBodyDir, const Param &serverParam)
{
	Vector accleration;
	float needPower;

#ifdef OPPTYPES_LOG3
	LOG << " is able to dash prevV=" << playerPrevVel << " nextV=" << playerVel << " bd=" << playerBodyDir << endl;
#endif	

	if (playerVel.getMagnitude() > playerSpeedMax)
		return 0;
	if (playerPrevVel.getMagnitude() > playerSpeedMax)
		return 0;
	
	accleration = (playerVel * (1 / playerDecay)) - playerPrevVel;
	accleration.rotate(-playerBodyDir);

#ifdef OPPTYPES_LOG3
	LOG << "accleration = " << accleration << endl;
#endif	
	
	if (accleration.getY() > 0.2)
		return 2;
	
	needPower = accleration.getX() / dashPowerRate;
	
#ifdef OPPTYPES_LOG3
	LOG << "power = " << needPower << endl;
#endif	

	if (needPower <= serverParam["maxpower"].asFloat() &&
		needPower >= serverParam["minpower"].asFloat())
		return 2;
	
	if (needPower <= serverParam["maxpower"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH)  &&
		needPower >= serverParam["minpower"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH))
		return 1;
	
	return 0;
}

// playerBodyDir is for the dash cycle.
bool TypeInfo::isMaxDash(Vector playerVel, Vector playerPrevVel,
		float playerBodyDir, const Param &serverParam)
{
	Vector accleration;
	float needPower;
#ifdef OPPTYPES_LOG3
	LOG << " is max dash prevV=" << playerPrevVel << " nextV=" << playerVel << " bd=" << playerBodyDir << endl;
#endif	

	if (playerVel.getMagnitude() > playerSpeedMax)
		return true;
	
	accleration = (playerVel * (1 / playerDecay)) - playerPrevVel;
	accleration.rotate(-playerBodyDir);

#ifdef OPPTYPES_LOG3
	LOG << "accleration = " << accleration << endl;
#endif	
	
	if (accleration.getY() > 0.2)
		return false;
	
	needPower = accleration.getX() / dashPowerRate;

#ifdef OPPTYPES_LOG3
	LOG << "power = " << needPower << endl;
#endif	

	if (needPower <= serverParam["maxpower"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH) &&
		needPower >= serverParam["maxpower"].asFloat() /
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH))
		return true;
	
	if (needPower >= serverParam["minpower"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH) &&
		needPower <= serverParam["minpower"].asFloat() /
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_DASH))
		return true;
	
	return false;
}

// playerVel is for the turn cycle.
int TypeInfo::isAbleToTurn(Vector playerVel, float angle,
		const Param &serverParam)
{
	float needMoment;

	needMoment = angle * (1 + playerVel.getMagnitude() *
			 inertiaMoment);
#ifdef OPPTYPES_LOG3
	LOG << " is able to turn angle=" << angle << " needMoment=" << needMoment << endl;
#endif
	
	if (needMoment <= serverParam["maxmoment"].asFloat() &&
		needMoment >= serverParam["minmoment"].asFloat())
		return 2;
	
	if (needMoment <= serverParam["maxmoment"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN)  &&
		needMoment >= serverParam["minmoment"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN))
		return 1;

	return 0;
}

// playerVel is for the turn cycle.
bool TypeInfo::isMaxTurn(Vector playerVel, float angle,
		const Param &serverParam)
{
	float needMoment;

	needMoment = angle * (1 + playerVel.getMagnitude() *
			 inertiaMoment);
#ifdef OPPTYPES_LOG3
	LOG << " is max turn angle=" << angle << " needMoment=" << needMoment << endl;
#endif	
	
	if (needMoment <= serverParam["maxmoment"].asFloat() *
		   (1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN) &&
		needMoment >= serverParam["maxmoment"].asFloat() /
	   		(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN))
		return true;
	
	if (needMoment >= serverParam["minmoment"].asFloat() *
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN) &&
		needMoment <= serverParam["minmoment"].asFloat() /
			(1 + serverParam["player_rand"].asFloat() + PLAYER_EXTRA_RAND_TURN))
		return true;

	return false;
}

// Class OppTypes

OppTypes::OppTypes(const CoachWorldModel *worldModel,
		const CoachWorldModelHistory &worldModelHistory):
	Task(worldModel, worldModelHistory)
{
	int i;
	typesSet = false;
	for (i = 0; i < 11; i++)
		resetPlayerForFirstTime(i);
}

void OppTypes::resetPlayer(int uniNum)
{
	int i;	

#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: resetPlayer " << uniNum << endl;
#endif
	
	for (i = 0; i < 7; i++)
	{
		sureNot[uniNum][i] = false;
		sureIs[uniNum][i] = 0;
	}
	sureIs[uniNum][0] = 0.01;
	foundSure[uniNum] = false;
	foundType[uniNum] = 0;
}

void OppTypes::resetPlayerForFirstTime(int uniNum)
{
	int i;	

#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: resetPlayerForFirstTime " << uniNum << endl;
#endif
	
	for (i = 0; i < 7; i++)
	{
		sureNot[uniNum][i] = true;
		sureIs[uniNum][i] = 0;
	}

	sureNot[uniNum][0] = false;
	sureIs[uniNum][0] = 0.01;
	foundSure[uniNum] = true;
	foundType[uniNum] = 0;
}

void OppTypes::informOppChangePlayerType(unsigned playerNum)
{
	resetPlayer(playerNum - 1);
}


void OppTypes::checkForTurns()
{
	int uniNum;
#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checkForTurns " << endl;
#endif
	if (worldModelHistory.getSize() <= 2)
		return;

	if (worldModelHistory[-1]->getPlayMode() != worldModel->getPlayMode() ||
		worldModelHistory[-1]->getVirtualPlayMode() != worldModel->getVirtualPlayMode())
		return;

#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checking " << endl;
#endif

	for (uniNum = 0;uniNum < 11; uniNum++)
	{
		if (foundSure[uniNum])
				continue;
		if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
			worldModel->getFullPlayer(TID_OPPONENT, uniNum).isAlive())
		{
//			LOG << "Turn OppAlive " << uniNum + 1 << endl;
//			LOG << "prev ba" << worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() << endl;
//			LOG << "prev time" << worldModelHistory[-1]->getCurTime() << endl;
//			LOG << "curr ba" << worldModel->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() << endl;
//			LOG << "curr time" << worldModel->getCurTime() << endl;
			if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() !=
				worldModel->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir())
				informTurn(uniNum, worldModelHistory[-1]->getCurTime());
		}
	}

	if (worldModelHistory.getSize() <= 3)
		return;
	for (uniNum = 0;uniNum < 11; uniNum++)
	{
		if (foundSure[uniNum])
				continue;
		if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
		    worldModelHistory[-2]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
			worldModel->getFullPlayer(TID_OPPONENT, uniNum).isAlive())
		{
			if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() !=
				worldModel->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() &&
			    worldModelHistory[-2]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() !=
				worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir())
				informMaxTurn(uniNum, worldModelHistory[-2]->getCurTime());
		}
	}
}

void OppTypes::informTurn(int uniNum, unsigned time)
{
	float angle;
	int i;
#ifdef OPPTYPES_LOG2
	char logString[8];
#endif
	angle = worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() -
		worldModelHistory[time + 1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir();
	angle = fabs(Degree::normalizeAngle(angle));


	for (i = 0; i < 7; i++)
	{
		if (sureNot[uniNum][i])
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'X';
#endif
			continue;
		}
		if (!types[i].isAbleToTurn(worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).
					getVel(), angle, worldModel->getServerParam()))
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'N';
#endif
			sureNot[uniNum][i] = true;
		}
		else
		{
#ifdef OPPTYPES_LOG2
			logString[i] = '0';
#endif
		}
	}
#ifdef OPPTYPES_LOG2
	logString[7] = 0;
	LOG << "OppTypes:informTurn p=" << uniNum + 1 << " t=" << time << " " << 
		" angle=" << angle << " info:" << logString << endl;
#endif
}

void OppTypes::informMaxTurn(int uniNum, unsigned time)
{
	float angle;
	int i;
#ifdef OPPTYPES_LOG2
	char logString[8];
#endif
	angle = worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() -
		worldModelHistory[time + 1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir();
	angle = fabs(Degree::normalizeAngle(angle));
	
	for (i = 0; i < 7; i++)
	{
		if (sureNot[uniNum][i])
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'X';
#endif
			continue;
		}
		if (types[i].isMaxTurn(worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).
					getVel(), angle, worldModel->getServerParam()))
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'Y';
#endif
			sureIs[uniNum][i] += (100 - sureIs[uniNum][i]) * 0.20;
		}
		else
		{
#ifdef OPPTYPES_LOG2
			logString[i] = '0';
#endif
		}
	}
#ifdef OPPTYPES_LOG2
	logString[7] = 0;
	LOG << "OppTypes:informMaxTurn p=" << uniNum + 1 << " t=" << time << " " << 
		" angle=" << angle << " info:" << logString << endl;
#endif
}

void OppTypes::checkForDashes()
{
	int uniNum;
	Vector accleration;

#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checkForDashes " << endl;
#endif

	if (worldModelHistory.getSize() <= 3)
		return;
	if (worldModel->getCurTime() - worldModel->getRefereeHearTime() < 3 ||
			worldModel->getPlayMode() == PM_BEFORE_KICK_OFF ||
			worldModel->getPlayMode() == PM_GOAL ||
			worldModel->getPlayMode() == PM_GOAL_OPP)
	{
		LOG << "OppTypes: Ignoring checkForDashes because of playmode" << endl; 
		return;
	}
	
#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checking " << endl;
#endif

	for (uniNum = 0;uniNum < 11; uniNum++)
	{
		if (foundSure[uniNum])
				continue;
		if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
			worldModel->getFullPlayer(TID_OPPONENT, uniNum).isAlive())
		{
			accleration = (worldModel->getFullPlayer(TID_OPPONENT, uniNum).getVel()
				* (1 / maxPlayerDecay)) -
				worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getVel();
			if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() ==
				worldModel->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() &&
				accleration.getMagnitude() > 0.3) //TODO : 0.3 <= (dash_power_rate  - new_dash_power_rate_delta_max)* max_power
				informDash(uniNum, worldModelHistory[-1]->getCurTime());
		}
	}

	if (worldModelHistory.getSize() <= 4)
		return;
	for (uniNum = 0;uniNum < 11; uniNum++)
	{
		if (foundSure[uniNum])
				continue;
		if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
		    worldModelHistory[-2]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
			worldModel->getFullPlayer(TID_OPPONENT, uniNum).isAlive())
		{
			accleration = (worldModel->getFullPlayer(TID_OPPONENT, uniNum).getVel()
				* (1 / maxPlayerDecay)) -
				worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getVel();
			if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() ==
				worldModel->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() &&
				accleration.getMagnitude() > 0.3)
			{
				accleration = (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).
					getVel() * (1 / maxPlayerDecay)) -
					worldModelHistory[-2]->getFullPlayer(TID_OPPONENT, uniNum).getVel();
				if (worldModelHistory[-2]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() ==
					worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir() &&
					accleration.getMagnitude() > 0.3) //TODO : 0.3 <= (dash_power_rate  - new_dash_power_rate_delta_max)* max_power

				informMaxDash(uniNum, worldModelHistory[-2]->getCurTime());
			}
		}
	}
}

void OppTypes::informDash(int uniNum, unsigned time)
{
	int i;
#ifdef OPPTYPES_LOG2
	char logString[8];
#endif
	
	for (i = 0; i < 7; i++)
	{
		if (sureNot[uniNum][i])
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'X';
#endif
			continue;
		}
		if (!types[i].isAbleToDash(worldModelHistory[time + 1]->
					getFullPlayer(TID_OPPONENT, uniNum).getVel(), worldModelHistory[time]->
					getFullPlayer(TID_OPPONENT, uniNum).getVel(), worldModelHistory[time]->
					getFullPlayer(TID_OPPONENT, uniNum).getBodyDir(),
					worldModel->getServerParam()))
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'N';
#endif
			sureNot[uniNum][i] = true;
		}
		else
		{
#ifdef OPPTYPES_LOG2
			logString[i] = '0';
#endif
		}
	}
#ifdef OPPTYPES_LOG2
	logString[7] = 0;
	LOG << "OppTypes:informDash p=" << uniNum + 1 << " t=" << time << " " << 
		" info:" << logString << endl;
#endif
}

void OppTypes::informMaxDash(int uniNum, unsigned time)
{
	int i;
#ifdef OPPTYPES_LOG2
	char logString[8];
#endif
	
	for (i = 0; i < 7; i++)
	{
		if (sureNot[uniNum][i])
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'X';
#endif
			continue;
		}
		if (types[i].isMaxDash(worldModelHistory[time + 1]->
					getFullPlayer(TID_OPPONENT, uniNum).getVel(), worldModelHistory[time]->
					getFullPlayer(TID_OPPONENT, uniNum).getVel(), worldModelHistory[time]->
					getFullPlayer(TID_OPPONENT, uniNum).getBodyDir(),
					worldModel->getServerParam()))
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'Y';
#endif
			sureIs[uniNum][i] += (100 - sureIs[uniNum][i]) * 0.02;
		}
		else
		{
#ifdef OPPTYPES_LOG2
			logString[i] = '0';
#endif
		}
	}
#ifdef OPPTYPES_LOG2
	logString[7] = 0;
	LOG << "OppTypes:informMaxDash p=" << uniNum + 1 << " t=" << time << " " << 
		" info:" << logString << endl;
#endif
}

void OppTypes::checkForKicks()
{
	int numPlayersAbleToKick = 0; // Tackle is a kick for us!
	int lastOppPlayerAbleToKick = -1;
	Vector acclerate;
	int uniNum;

#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checkForKicks " << endl;
#endif

	if (worldModelHistory.getSize() <= 2)
		return;
	if (worldModel->getCurTime() - worldModel->getRefereeHearTime() < 2 ||
			worldModel->getPlayMode() == PM_BEFORE_KICK_OFF ||
			worldModel->getPlayMode() == PM_GOAL ||
			worldModel->getPlayMode() == PM_GOAL_OPP)
	{
#ifdef OPPTYPES_LOG1
		LOG << "OppTypes: Ignoring checkForKicks because of playmode" << endl; 
#endif		
		return;
	}
	
#ifdef OPPTYPES_LOG1
	LOG << "OppTypes: checking " << endl;
#endif

	acclerate = worldModel->getBall().getVel() *
		(1 / (worldModel->getServerParam())["ball_decay"].asFloat()) - 
		worldModelHistory[-1]->getBall().getVel();
	
	if (acclerate.getMagnitude() > 0.2)
	{
#ifdef OPPTYPES_LOG1
		LOG << "OppTypes::Ball is kicked" << endl;
#endif		
	
		// check ability of opponents
		for (uniNum = 0;uniNum < 11; uniNum++)
		{
			if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).isAlive() &&
				worldModel->getFullPlayer(TID_OPPONENT, uniNum).isAlive())
			{
				if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).
						getDistance(worldModelHistory[-1]->getBall()) <= maxKickableArea)
				{
#ifdef OPPTYPES_LOG1
					LOG << "   Opp " << uniNum + 1 << " is able to kick" << endl;
#endif					
					lastOppPlayerAbleToKick = uniNum;
					numPlayersAbleToKick++;
				}

				if (worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).
						getDistance(worldModelHistory[-1]->getBall()) <
						(worldModel->getServerParam())["tackle_dist"].asFloat() &&
						worldModelHistory[-1]->getFullPlayer(TID_OPPONENT, uniNum).
						isTackling() == false &&
						worldModel->getFullPlayer(TID_OPPONENT, uniNum).
						isTackling() == true)
				{
#ifdef OPPTYPES_LOG1
					LOG << "   Opp " << uniNum + 1 << " is able to kick(tackle)" << endl;
#endif					
					lastOppPlayerAbleToKick = -1;
					numPlayersAbleToKick++;
				}
			}
		}

		// check ability of teammates
		for (uniNum = 0;uniNum < 11; uniNum++)
		{
			if (worldModelHistory[-1]->getFullPlayer(TID_TEAMMATE, uniNum).isAlive() &&
				worldModel->getFullPlayer(TID_TEAMMATE, uniNum).isAlive())
			{
//				LOG << "   testTmm " << uniNum + 1 << " ball_dist = ";
//				LOG << worldModelHistory[-1]->getFullPlayer(TID_TEAMMATE, uniNum).
//					getDistance(worldModelHistory[-1]->getBall()) << endl;
				if (worldModelHistory[-1]->getFullPlayer(TID_TEAMMATE, uniNum).
						getDistance(worldModelHistory[-1]->getBall()) <= maxKickableArea)
				{
#ifdef OPPTYPES_LOG1
					LOG << "   Tmm " << uniNum + 1 << " is able to kick" << endl;
#endif					
					numPlayersAbleToKick++;
				}
				if (worldModelHistory[-1]->getFullPlayer(TID_TEAMMATE, uniNum).
						getDistance(worldModelHistory[-1]->getBall()) <
						(worldModel->getServerParam())["tackle_dist"].asFloat() &&
						worldModelHistory[-1]->getFullPlayer(TID_TEAMMATE, uniNum).
						isTackling() == false &&
						worldModel->getFullPlayer(TID_TEAMMATE, uniNum).
						isTackling() == true)
				{
#ifdef OPPTYPES_LOG1
					LOG << "   Tmm " << uniNum + 1 << " is able to kick(tackle)" << endl;
#endif					
					numPlayersAbleToKick++;
				}
			}
		}
		
		if (numPlayersAbleToKick == 0)
		{
#ifdef OPPTYPES_LOG1
			LOG << "OppTypes: WHAT!!!!! ball kicked without kicker" << endl;
#endif			
			return;
		}

		if (numPlayersAbleToKick > 1)
		{
#ifdef OPPTYPES_LOG1
			LOG << "OppTypes: many kickers" << endl;
#endif			
			return;
		}

		if (lastOppPlayerAbleToKick == -1)
		{
#ifdef OPPTYPES_LOG1
			LOG << "only teammate kickers or opponents tacklers" << endl;
#endif			
			return;
		}

		if (foundSure[lastOppPlayerAbleToKick])
		{
#ifdef OPPTYPES_LOG1
			LOG << "foundSure = true" << endl;
#endif			
			return;
		}

		informKick(lastOppPlayerAbleToKick, worldModelHistory[-1]->getCurTime(), acclerate);
	}

}

void OppTypes::informKick(int uniNum, unsigned time, Vector acclerate)
{
	int i;
#ifdef OPPTYPES_LOG2
	char logString[8];
#endif
	
	for (i = 0; i < 7; i++)
	{
		if (sureNot[uniNum][i])
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'X';
#endif
			continue;
		}
		if (!types[i].isAbleToKick(
					worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).getPos(),
					worldModelHistory[time]->getBall().getPos(),
					worldModelHistory[time]->getFullPlayer(TID_OPPONENT, uniNum).getBodyDir(),
					acclerate, worldModel->getFieldSide(), worldModel->getServerParam()))
		{
#ifdef OPPTYPES_LOG2
			logString[i] = 'N';
#endif
			sureNot[uniNum][i] = true;
		}
		else
		{
#ifdef OPPTYPES_LOG2
			logString[i] = '0';
#endif
		}
	}
#ifdef OPPTYPES_LOG2
	logString[7] = 0;
	LOG << "OppTypes:informKick p=" << uniNum + 1 << " t=" << time << " " << 
		" info:" << logString << endl;
#endif

}

bool OppTypes::find()
{
	bool changed;
	int newType[11];
	int uniNum, i;

	float max;
	int maxi, ableNum;
#ifdef OPPTYPES_LOG1
	char logString[8];
#endif
	
	for (uniNum = 0; uniNum < 11; uniNum++)
	{
		ableNum = 0;
		max = -1;
		maxi = 0;
		for (i = 0; i < 7; i++)
			if (!sureNot[uniNum][i])
			{
				ableNum++;
				if (sureIs[uniNum][i] > max)
				{
					max = sureIs[uniNum][i];
					maxi = i;
				}
			}
		if (max < 0)
		{
			maxi = 0;
			resetPlayer(uniNum);
		}
		if (ableNum == 1)
			foundSure[uniNum] = true;
		else
			foundSure[uniNum] = false;

		newType[uniNum] = maxi;
	}
	
	changed = false;
	for (uniNum = 0; uniNum < 11; uniNum++)
	{
		if (newType[uniNum] != foundType[uniNum])
		{
			changed = true;
			foundType[uniNum] = newType[uniNum];
		}
	}

#ifdef OPPTYPES_LOG1
	logString[7] = 0;
	for (uniNum = 0; uniNum < 11; uniNum++)
	{
		for (i = 0; i < 7; i++)
		{
			if (sureNot[uniNum][i])
				logString[i] = 'N';
			else
				logString[i] = 'Y';
		}
		if (foundSure[uniNum])
			LOG << "foundOppType p " << uniNum + 1 << " =  " << foundType[uniNum] << 
				"  <" << sureIs[uniNum][foundType[uniNum]] << "> info:" << logString << endl;
		else
			LOG << "foundOppType p " << uniNum + 1 << " = (" << foundType[uniNum] << 
				") <" << sureIs[uniNum][foundType[uniNum]] << "> info:" << logString << endl;
	}
#endif

	if (changed)
		return true;
	else
		return false;
}

void OppTypes::setFreeForm()
{
	int i;
	freeFormString = "(T "; // Opponent heterogenous types flag
	for (i = 0; i < 11; i++)
		freeFormString += char('0' + foundType[i]);
	freeFormString += ")";
	freeFormReadyFlag = true;
}

OppTypes::~OppTypes()
{
}
	
bool OppTypes::execute()
{
	LOG << "OppTypes execute" << endl;
	if (!typesSet)
		initialize();
	// check goalie
	int i;

	for (i = 0; i < 10; i++)
		if (worldModel->getFullPlayer(TID_OPPONENT, i).isAlive() &&
			worldModel->getFullPlayer(TID_OPPONENT, i).isGoalie())
		resetPlayerForFirstTime(i);
	checkForTurns();
	checkForDashes();
	checkForKicks();
	if (find())
		setFreeForm();
	return true;
}

void OppTypes::initialize()
{
	int i;

	for (i = 0; i < 7; i++)
		types[i].setByPlayerType(worldModel->getPlayerType(i));

	maxPlayerDecay = 0;
	for (i = 0; i < 7; i++)
		if (types[i].playerDecay > maxPlayerDecay)
			maxPlayerDecay = types[i].playerDecay;
	
	maxKickableArea = 0;
	for (i = 0; i < 7; i++)
		if (types[i].kickableMargin + types[i].playerSize +
				worldModel->getServerParam()["ball_size"].asFloat() > maxKickableArea)
			maxKickableArea = types[i].kickableMargin + types[i].playerSize +
				worldModel->getServerParam()["ball_size"].asFloat();
#ifdef OPPTYPES_LOG3	
	LOG << "OppTypes:maxKickableArea = " << maxKickableArea << endl;
#endif	
	typesSet = true;
}	
