/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *  and is modified by: Mohammad Salehe, Mostafa Rokooey,
 *                      Sassan Haradji, Darioush Jalali
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <cmath>
#include <sstream>
#include <cassert>
#include <iostream>

#include <Config.h>
#include <Logger.h>
#include <Basics.h>
#include <Degree.h>
#include <Defines.h>
#include <WorldModel.h>
#include <SayEncoding.h>

using namespace std;
using namespace Basics;
using namespace Degree;

#define OPP_OFFLINE_LOCAL_X 20
#define OPP_OFFLINE_LOCAL_Y 10

#define MARKING_DECAY_FACTOR (0.8)
#define MARKING_THRESHOLD    (2.0)
#define MARKING_AREA         (2.75)

// class WorldModel

WorldModel::WorldModel()
{
	unsigned i;

	if (config["Agent"]["Server"]["OfflinePlayer"].asBool())
		timer = new VirtualTimer();
	else
		timer = new OnlineTimer();

	seeCycle = 0;
	curCycle = 0;
	curTime = 0;
	fieldSide = SI_LEFT;
	playMode = PM_BEFORE_KICK_OFF;
	virtualPlayMode = VPM_NONE;
	playModeRemainCycle = 0;

	setBodyReplaced(false);

	initViewPartDirs();
	resetSeeDeltaCycles();
	freeKickAdvicerSayCycle = 0;

	curKickCalculate.init();

	// Objects on the field.
	flagsNum = 0;
	goalsNum = 0;
	linesNum = 0;

	for (i = 0; i < FLAGS_NUM; i++)
		flags[i] = new FieldFlag();
	for (i = 0; i < GOALS_NUM; i++)
		goals[i] = new FieldGoal();
	for (i = 0; i < LINES_NUM; i++)
		lines[i] = new FieldLine();

	ball = new Ball;
	body = new Body;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		fullPlayers[TID_TEAMMATE][i] = new Player(PLM_FULL);
		fullPlayers[TID_TEAMMATE][i]->setUniNum(i + 1);
	}
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		fullPlayers[TID_OPPONENT][i] = new Player(PLM_FULL);
		fullPlayers[TID_OPPONENT][i]->setUniNum(i + 1);
	}

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_TEAMMATE][i] = new Player(PLM_HALF);
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_OPPONENT][i] = new Player(PLM_HALF);

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		quarterPlayers[i] = new Player(PLM_QUARTER);
}

WorldModel::WorldModel(const WorldModel &worldModelCopy)
{
	copyFrom(worldModelCopy, false);

	LOG << "WorldModel copied." << endl;
}

WorldModel::~WorldModel()
{
	unsigned i;

	delete timer;

	for (i = 0; i < FLAGS_NUM; i++)
		delete flags[i];
	for (i = 0; i < GOALS_NUM; i++)
		delete goals[i];
	for (i = 0; i < LINES_NUM; i++)
		delete lines[i];

	delete ball;

	if (!isBodyReplaced())
		delete body; // else it will be deleted with other players.

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		delete fullPlayers[TID_TEAMMATE][i];
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		delete fullPlayers[TID_OPPONENT][i];

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		delete halfPlayers[TID_TEAMMATE][i];
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		delete halfPlayers[TID_OPPONENT][i];

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		delete quarterPlayers[i];
}

void WorldModel::copyFrom(const WorldModel &worldModelCopy, bool constructed)
{
	unsigned i;
	if (constructed)
	{
		LOG << "WorldModel copying 1" << endl;
	
		delete timer;
	
		for (i = 0; i < FLAGS_NUM; i++)
			delete flags[i];
		for (i = 0; i < GOALS_NUM; i++)
			delete goals[i];
		for (i = 0; i < LINES_NUM; i++)
			delete lines[i];
	
		delete ball;
	
		if (!isBodyReplaced())
			delete body; // else it will be deleted with other players.
	
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			delete fullPlayers[TID_TEAMMATE][i];
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			delete fullPlayers[TID_OPPONENT][i];
	
		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			delete halfPlayers[TID_TEAMMATE][i];
		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			delete halfPlayers[TID_OPPONENT][i];
	
		for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
			delete quarterPlayers[i];
	}
	LOG << "WorldModel copying 2" << endl;

	*this = worldModelCopy;

	if (dynamic_cast<OnlineTimer *>(worldModelCopy.timer))
		timer = new OnlineTimer((const OnlineTimer&)*(worldModelCopy.timer));
	else
		timer = new VirtualTimer((const VirtualTimer&)*(worldModelCopy.timer));

	for (i = 0; i < FLAGS_NUM; i++)
		flags[i] = new FieldFlag(*(worldModelCopy.flags[i]));
	for (i = 0; i < GOALS_NUM; i++)
		goals[i] = new FieldGoal(*(worldModelCopy.goals[i]));
	for (i = 0; i < LINES_NUM; i++)
		lines[i] = new FieldLine(*(worldModelCopy.lines[i]));

	ball = new Ball(*(worldModelCopy.ball));
	body = new Body(*(worldModelCopy.body));

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_TEAMMATE][i] =
			new Player(*(worldModelCopy.fullPlayers[TID_TEAMMATE][i]));
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_OPPONENT][i] =
			new Player(*(worldModelCopy.fullPlayers[TID_OPPONENT][i]));

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_TEAMMATE][i] =
			new Player(*(worldModelCopy.halfPlayers[TID_TEAMMATE][i]));
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_OPPONENT][i] =
			new Player(*(worldModelCopy.halfPlayers[TID_OPPONENT][i]));

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		quarterPlayers[i] =
			new Player(*(worldModelCopy.quarterPlayers[i]));

	if (worldModelCopy.bodyReplacedFlag)
		replaceBodyInPlayersList(worldModelCopy.body->getUniNum());
}

void WorldModel::updateAfterSenseBody()
{
	curKickCalculate.setValid(false);

	resetHearedVarsAfterSenseBody();
}

void WorldModel::initServerParam(const SExpression &exp)
{
	int i;

	SExpAtomic *at = dynamic_cast<SExpAtomic *>(exp[0]);
	assert(at);
	assert(at->asString() == "server_param");

	for (i = 1; i < exp.size(); i++)
	{
		SExpression *se = dynamic_cast<SExpression *>(exp[i]);
		assert(se);
		SExpAtomic *at1 = dynamic_cast<SExpAtomic *>((*se)[0]);
		assert(at1);
		SExpAtomic *at2 = dynamic_cast<SExpAtomic *>((*se)[1]);
		assert(at2);
		serverParam.add(at1->asString(), at2->asString());
	}

	// now I can set default variables of objects.
	ball->setServerParamVars(serverParam);

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_TEAMMATE][i]->setServerParamVars(serverParam);
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_OPPONENT][i]->setServerParamVars(serverParam);

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_TEAMMATE][i]->setServerParamVars(serverParam);
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_OPPONENT][i]->setServerParamVars(serverParam);

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		quarterPlayers[i]->setServerParamVars(serverParam);
}

void WorldModel::initPlayerParam(const SExpression &exp)
{
	int i;

	SExpAtomic *at = dynamic_cast<SExpAtomic *>(exp[0]);
	assert(at);
	assert(at->asString() == "player_param");

	for (i = 1; i < exp.size(); i++)
	{
		SExpression *se = dynamic_cast<SExpression *>(exp[i]);
		assert(se);
		SExpAtomic *at1 = dynamic_cast<SExpAtomic *>((*se)[0]);
		assert(at1);
		SExpAtomic *at2 = dynamic_cast<SExpAtomic *>((*se)[1]);
		assert(at2);
		playerParam.add(at1->asString(), at2->asString());
	}

	playerTypes = vector<Param>(playerParam["player_types"].asInt());
}

void WorldModel::initPlayerType(const SExpression &exp)
{
	int i;

	SExpAtomic *at = dynamic_cast<SExpAtomic *>(exp[0]);
	assert(at);
	assert(at->asString() == "player_type");

	const SExpression &seid = exp["id"];
	at = dynamic_cast<SExpAtomic *>(seid[1]);
	assert(at);
	int id = at->asInt();
	
	for (i = 2; i < exp.size(); i++)
	{
		SExpression *se = dynamic_cast<SExpression *>(exp[i]);
		assert(se);
		SExpAtomic *at1 = dynamic_cast<SExpAtomic *>((*se)[0]);
		assert(at1);
		SExpAtomic *at2 = dynamic_cast<SExpAtomic *>((*se)[1]);
		assert(at2);
		playerTypes[id].add(at1->asString(), at2->asString());
	}

	if (id == 0) // now I can set default players type.
	{
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			fullPlayers[TID_TEAMMATE][i]->setType(0, playerTypes[0]);
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			fullPlayers[TID_OPPONENT][i]->setType(0, playerTypes[0]);

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			halfPlayers[TID_TEAMMATE][i]->setType(0, playerTypes[0]);
		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			halfPlayers[TID_OPPONENT][i]->setType(0, playerTypes[0]);

		for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
			quarterPlayers[i]->setType(0, playerTypes[0]);
	}
}

void WorldModel::parseSee(const SExpression &exp)
{
	SExpAtomic *at;
	SExpression *se1,*se2;
	int i;

	at = dynamic_cast<SExpAtomic *>(exp[0]);
	assert(at);
	assert(at->asString() == "see");

	// See time
	at = dynamic_cast<SExpAtomic *>(exp[1]);
	assert(at);
	seeCycle = at->asInt();
	seeTime = getCurTime();

	for (i = 2; i < exp.size(); i++)
	{
		se1 = dynamic_cast<SExpression *>(exp[i]);
		assert(se1);
		se2 = dynamic_cast<SExpression *>((*se1)[0]);
		assert(se2);
		at = dynamic_cast<SExpAtomic *>((*se2)[0]);
		assert(at);

		switch (at->asString()[0])
		{
		case 'f':
			parseFlag(*se1);
			break;

		case 'g':
			parseGoal(*se1);
			break;

		case 'l':
			parseLine(*se1);
			break;

		case 'b':
		case 'B':
			parseBall(*se1);
			break;

		case 'p':
		case 'P':
			parsePlayer(*se1);
			break;

		case 'F':
		case 'L':
		case 'G':
			break; // Going to trash.

		default:
			assert(0);
		}
	}

}

void WorldModel::parseFlag(const SExpression &exp)
{
	SExpression *se;
	int flagId;

	se = dynamic_cast<SExpression *>(exp[0]);
	assert(se);
	flagId = fieldData.getFlagId(*se);

	flags[flagsNum]->parse(flagId, exp, getCurTime());
	flagsNum++;
}

void WorldModel::parseGoal(const SExpression &exp)
{
	SExpression *se;
	int goalId;

	se = dynamic_cast<SExpression *>(exp[0]);
	assert(se);
	goalId = fieldData.getGoalId(*se);

	goals[goalsNum]->parse(goalId, exp, getCurTime());
	goalsNum++;
}

void WorldModel::parseLine(const SExpression &exp)
{
	SExpression *se;
	int lineId;

	se = dynamic_cast<SExpression *>(exp[0]);
	assert(se);
	lineId = fieldData.getLineId(*se);

	lines[linesNum]->parse(lineId, exp, getCurTime());
	linesNum++;
}

void WorldModel::parseBall(const SExpression &exp)
{
	ball->parse(exp, getCurTime());
}

void WorldModel::parsePlayer(const SExpression &exp)
{
	SExpression *se;
	SExpAtomic *at;
	string ourTeamName;
	TeamId teamId = TID_UNKNOWN;
	unsigned uniNum = 0;
	bool isGoalie = false;
	unsigned i;

	se = dynamic_cast<SExpression *>(exp[0]);
	assert(se);
	
	if (se->size() > 1)
	{
		at = dynamic_cast<SExpAtomic *>((*se)[1]);
		assert(at);

		ourTeamName = (string)"\"" +
			config["Agent"]["Public"]["TeamName"].asString() + "\"";

		if (at->toString() == ourTeamName)
			teamId = TID_TEAMMATE;
		else
			teamId = TID_OPPONENT;
	}

	if (se->size() > 2)
	{
		at = dynamic_cast<SExpAtomic *>((*se)[2]);
		assert(at);
		uniNum = at->asInt();
	}

	if (se->size() > 3)
	{
		at = dynamic_cast<SExpAtomic *>((*se)[3]);
		assert(at);
		assert(at->toString() == "goalie");
		isGoalie = true;
	}

	if (teamId == TID_UNKNOWN) // Quarter
	{
		for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
			if (!quarterPlayers[i]->isValid() &&
				!quarterPlayers[i]->isReadyToUpdate())
				break;

		if (i < QUARTER_PLAYERS_NUM)
			quarterPlayers[i]->parse(teamId, uniNum, isGoalie, exp, getCurTime());
	}

	else if (uniNum == 0) // Half
	{
		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (!halfPlayers[teamId][i]->isValid() &&
				!halfPlayers[teamId][i]->isReadyToUpdate())
				break;

		if (i < HALF_PLAYERS_NUM)
			halfPlayers[teamId][i]->parse(teamId, uniNum, isGoalie, exp, getCurTime());
	}

	else // Full
	{
		assert(uniNum <= FULL_PLAYERS_NUM);
		fullPlayers[teamId][uniNum - 1]->parse(teamId, uniNum, isGoalie, exp,
											   getCurTime());
	}

}

void WorldModel::resetObjects()
{
	unsigned i;

	flagsNum = 0;
	goalsNum = 0;
	linesNum = 0;

	// Reading to parse objects
	ball->setReadyToUpdate(false);

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_TEAMMATE][i]->setReadyToUpdate(false);
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		fullPlayers[TID_OPPONENT][i]->setReadyToUpdate(false);

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_TEAMMATE][i]->setReadyToUpdate(false);
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		halfPlayers[TID_OPPONENT][i]->setReadyToUpdate(false);

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		quarterPlayers[i]->setReadyToUpdate(false);

	// Killing a few players
	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		quarterPlayers[i]->kill();

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
	{
		if (halfPlayers[TID_TEAMMATE][i]->isAlive() &&
			halfPlayers[TID_TEAMMATE][i]->canBodySee(body->getPos(),
													 body->getViewAngle(), body->getVisibleDistance()))
			halfPlayers[TID_TEAMMATE][i]->kill();

		if (halfPlayers[TID_OPPONENT][i]->isAlive() &&
			halfPlayers[TID_OPPONENT][i]->canBodySee(body->getPos(),
													 body->getViewAngle(), body->getVisibleDistance()))
			halfPlayers[TID_OPPONENT][i]->kill();
	}
}

void WorldModel::replaceBodyInPlayersList(unsigned bodyUniNum)
{
	delete fullPlayers[TID_TEAMMATE][bodyUniNum - 1];
	fullPlayers[TID_TEAMMATE][bodyUniNum - 1] = body;

	setBodyReplaced(true);
}

void WorldModel::updateObjects()
{
	unsigned i;

	for (i = 0; i < flagsNum; i++)
		flags[i]->update(fieldData);

	for (i = 0; i < goalsNum; i++)
		goals[i]->update(fieldData);

	for (i = 0; i < linesNum; i++)
		lines[i]->update(fieldData);

	body->update(flags, goals, lines, flagsNum, goalsNum, linesNum);

	if (ball->isReadyToUpdate())
		ball->update(*body);
	else if (ball->isValid())
		ball->virtualUpdate(*body);

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_TEAMMATE][i]->isReadyToUpdate())
			fullPlayers[TID_TEAMMATE][i]->update(*body, playMode);
		else if (i != body->getUniNum() - 1 && // this function do not virtual update body.
				 fullPlayers[TID_TEAMMATE][i]->isAlive())
			fullPlayers[TID_TEAMMATE][i]->virtualUpdate(*body);

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isReadyToUpdate())
			fullPlayers[TID_OPPONENT][i]->update(*body, playMode);
		else if (fullPlayers[TID_OPPONENT][i]->isAlive())
			fullPlayers[TID_OPPONENT][i]->virtualUpdate(*body);

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_TEAMMATE][i]->isReadyToUpdate())
			halfPlayers[TID_TEAMMATE][i]->update(*body, playMode);
		else if (halfPlayers[TID_TEAMMATE][i]->isValid())
			halfPlayers[TID_TEAMMATE][i]->virtualUpdate(*body);

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isReadyToUpdate())
			halfPlayers[TID_OPPONENT][i]->update(*body, playMode);
		else if (halfPlayers[TID_OPPONENT][i]->isValid())
			halfPlayers[TID_OPPONENT][i]->virtualUpdate(*body);

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		if (quarterPlayers[i]->isReadyToUpdate())
			quarterPlayers[i]->update(*body, playMode);
		else if (quarterPlayers[i]->isValid())
			quarterPlayers[i]->virtualUpdate(*body);

	completeUpdateObjects();
}

void WorldModel::simulateObjects(const Body &lastBody, const Command *bodyCycleCommand)
{
	unsigned i;

	if (getPlayMode() == PM_FREE_KICK &&
		getVirtualPlayMode() ==	VPM_GOALIE_CATCH_BALL && body->isGoalie())
		ball->simulateByAction(lastBody, bodyCycleCommand, true);
	else
		ball->simulateByAction(lastBody, bodyCycleCommand);

	ball->simulateByDynamics(*body, 2, ballStatus, curTime);

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (i != body->getUniNum() - 1 && // this function do not simulate body.
			fullPlayers[TID_TEAMMATE][i]->isAlive())
			fullPlayers[TID_TEAMMATE][i]->simulateByDynamics(*body);
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isAlive())
			fullPlayers[TID_OPPONENT][i]->simulateByDynamics(*body);

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_TEAMMATE][i]->isValid())
			halfPlayers[TID_TEAMMATE][i]->simulateByDynamics(*body);
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid())
			halfPlayers[TID_OPPONENT][i]->simulateByDynamics(*body);

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		if (quarterPlayers[i]->isValid())
			quarterPlayers[i]->simulateByDynamics(*body);
}

void WorldModel::completeUpdateObjects()
{
	int i, j;
	Player *playerPtr;

	// Removing same Halves
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
	{
		// Teammates
		if (halfPlayers[TID_TEAMMATE][i]->isValid() &&
			halfPlayers[TID_TEAMMATE][i]->getSeeTime() != getCurTime())
		{

			playerPtr = NULL;
			for (j = 0; j < HALF_PLAYERS_NUM; j++)
				if (halfPlayers[TID_TEAMMATE][j]->isValid() &&
					halfPlayers[TID_TEAMMATE][j]->getSeeTime() == getCurTime() &&
					halfPlayers[TID_TEAMMATE][j]->getDistance(*halfPlayers[TID_TEAMMATE][i]) <=
					halfPlayers[TID_TEAMMATE][j]->getPosDeviation() + halfPlayers[TID_TEAMMATE][i]->getPosDeviation() + 1) // One meter for sure
				{
					if (!playerPtr)
						playerPtr = halfPlayers[TID_TEAMMATE][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
	
			if (playerPtr)
			{
				for (j = 0; j < HALF_PLAYERS_NUM; j++)
					if (i != j && halfPlayers[TID_TEAMMATE][j]->isValid() &&
						halfPlayers[TID_TEAMMATE][j]->getSeeTime() != getCurTime() &&
						halfPlayers[TID_TEAMMATE][j]->getDistance(*playerPtr) <=
						halfPlayers[TID_TEAMMATE][j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
				{
					j = -1;
					break;
				}

				if (j != -1)
					halfPlayers[TID_TEAMMATE][i]->kill();
			}
		}

		// Opponents
		if (halfPlayers[TID_OPPONENT][i]->isValid() &&
			halfPlayers[TID_OPPONENT][i]->getSeeTime() != getCurTime())
		{

			playerPtr = NULL;
			for (j = 0; j < HALF_PLAYERS_NUM; j++)
				if (halfPlayers[TID_OPPONENT][j]->isValid() &&
					halfPlayers[TID_OPPONENT][j]->getSeeTime() == getCurTime() &&
					halfPlayers[TID_OPPONENT][j]->getDistance(*halfPlayers[TID_OPPONENT][i]) <=
					halfPlayers[TID_OPPONENT][j]->getPosDeviation() + halfPlayers[TID_OPPONENT][i]->getPosDeviation() + 1) // One meter for sure
				{
					if (!playerPtr)
						playerPtr = halfPlayers[TID_OPPONENT][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
	
			if (playerPtr)
			{
				for (j = 0; j < HALF_PLAYERS_NUM; j++)
					if (i != j && halfPlayers[TID_OPPONENT][j]->isValid() &&
						halfPlayers[TID_OPPONENT][j]->getSeeTime() != getCurTime() &&
						halfPlayers[TID_OPPONENT][j]->getDistance(*playerPtr) <=
						halfPlayers[TID_OPPONENT][j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
				{
					j = -1;
					break;
				}

				if (j != -1)
					halfPlayers[TID_OPPONENT][i]->kill();
			}
		}

	}

	// Changing Quarters to Halves
	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		if (quarterPlayers[i]->isValid())
		{
			playerPtr = NULL;
			for (j = 0; j < HALF_PLAYERS_NUM; j++)
			{
				// Teammates
				if (halfPlayers[TID_TEAMMATE][j]->isValid() &&
					halfPlayers[TID_TEAMMATE][j]->getSeeTime() != getCurTime() &&
					halfPlayers[TID_TEAMMATE][j]->getDistance(*quarterPlayers[i]) <=
					halfPlayers[TID_TEAMMATE][j]->getPosDeviation() + quarterPlayers[i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = halfPlayers[TID_TEAMMATE][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}

				// Opponents
				if (halfPlayers[TID_OPPONENT][j]->isValid() &&
					halfPlayers[TID_OPPONENT][j]->getSeeTime() != getCurTime() &&
					halfPlayers[TID_OPPONENT][j]->getDistance(*quarterPlayers[i]) <=
					halfPlayers[TID_OPPONENT][j]->getPosDeviation() + quarterPlayers[i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = halfPlayers[TID_OPPONENT][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
			}
	
			if (playerPtr != NULL)
			{
				for (j = 0; j < QUARTER_PLAYERS_NUM; j++)
					if (i != j && quarterPlayers[j]->isValid() &&
						quarterPlayers[j]->getDistance(*playerPtr) <=
						quarterPlayers[j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
					{
						j = -1;
						break;
					}

				if (j != -1)
				{
					playerPtr->setByAnotherPlayer(*quarterPlayers[i], *body);
					quarterPlayers[i]->kill();
				}
			}
		}

	// Changing Quarters to Fulls
	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		if (quarterPlayers[i]->isValid())
		{
			playerPtr = NULL;
			for (j = 0; j < FULL_PLAYERS_NUM; j++)
			{
				// Teammates
				if (fullPlayers[TID_TEAMMATE][j]->isAlive() &&
					!fullPlayers[TID_TEAMMATE][j]->isBody() &&
					fullPlayers[TID_TEAMMATE][j]->getSeeTime() != getCurTime() &&
					fullPlayers[TID_TEAMMATE][j]->getDistance(*quarterPlayers[i]) <=
					fullPlayers[TID_TEAMMATE][j]->getPosDeviation() + quarterPlayers[i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = fullPlayers[TID_TEAMMATE][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}

				// Opponents
				if (fullPlayers[TID_OPPONENT][j]->isAlive() &&
					fullPlayers[TID_OPPONENT][j]->getSeeTime() != getCurTime() &&
					fullPlayers[TID_OPPONENT][j]->getDistance(*quarterPlayers[i]) <=
					fullPlayers[TID_OPPONENT][j]->getPosDeviation() + quarterPlayers[i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = fullPlayers[TID_OPPONENT][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
			}
	
			if (playerPtr != NULL)
			{
				for (j = 0; j < QUARTER_PLAYERS_NUM; j++)
					if (i != j && quarterPlayers[j]->isValid() &&
						quarterPlayers[j]->getDistance(*playerPtr) <=
						quarterPlayers[j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
					{
						j = -1;
						break;
					}

				if (j != -1)
				{
					playerPtr->setByAnotherPlayer(*quarterPlayers[i], *body);
					quarterPlayers[i]->kill();

//LOG << "Tran qua to " << playerPtr->getTeamId() << " " << playerPtr->getUniNum() << endl;
//LOG << "qu ti: " << quarterPlayers[i]->getSeeTime() << " fu ti: " << playerPtr->getSeeTime() << " cur ti: " << curTime << endl;

				}
			}
		}

	// Changing Halves to Fulls
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
	{
		// Teammates
		if (halfPlayers[TID_TEAMMATE][i]->isValid())
		{
			playerPtr = NULL;
			for (j = 0; j < FULL_PLAYERS_NUM; j++)
				if (fullPlayers[TID_TEAMMATE][j]->isAlive() &&
					!fullPlayers[TID_TEAMMATE][j]->isBody() &&
					fullPlayers[TID_TEAMMATE][j]->getSeeTime() != getCurTime() &&
					fullPlayers[TID_TEAMMATE][j]->getDistance(*halfPlayers[TID_TEAMMATE][i]) <=
					fullPlayers[TID_TEAMMATE][j]->getPosDeviation() + halfPlayers[TID_TEAMMATE][i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = fullPlayers[TID_TEAMMATE][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
	
			if (playerPtr != NULL)
			{
				for (j = 0; j < HALF_PLAYERS_NUM; j++)
					if (i != j && halfPlayers[TID_TEAMMATE][j]->isValid() &&
							halfPlayers[TID_TEAMMATE][j]->getDistance(*playerPtr) <=
							halfPlayers[TID_TEAMMATE][j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
					{
						j = -1;
						break;
					}

				if (j != -1)
				{
					playerPtr->setByAnotherPlayer(*halfPlayers[TID_TEAMMATE][i], *body);
					halfPlayers[TID_TEAMMATE][i]->kill();
				}
			}
		}

		// Opponents
		if (halfPlayers[TID_OPPONENT][i]->isValid())
		{
			playerPtr = NULL;
			for (j = 0; j < FULL_PLAYERS_NUM; j++)
				if (fullPlayers[TID_OPPONENT][j]->isAlive() &&
					fullPlayers[TID_OPPONENT][j]->getSeeTime() != getCurTime() &&
					fullPlayers[TID_OPPONENT][j]->getDistance(*halfPlayers[TID_OPPONENT][i]) <=
					fullPlayers[TID_OPPONENT][j]->getPosDeviation() + halfPlayers[TID_OPPONENT][i]->getPosDeviation() + 1) // One meter for sure
				{
					if (playerPtr == NULL)
						playerPtr = fullPlayers[TID_OPPONENT][j];
					else
					{
						playerPtr = NULL;
						break;
					}
				}
	
			if (playerPtr != NULL)
			{
				for (j = 0; j < HALF_PLAYERS_NUM; j++)
					if (i != j && halfPlayers[TID_OPPONENT][j]->isValid() &&
						halfPlayers[TID_OPPONENT][j]->getDistance(*playerPtr) <=
						halfPlayers[TID_OPPONENT][j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
					{
						j = -1;
						break;
					}

				if (j != -1)
				{
					playerPtr->setByAnotherPlayer(*halfPlayers[TID_OPPONENT][i], *body);
					halfPlayers[TID_OPPONENT][i]->kill();
				}
			}
		}

	}

	// Players change place
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
	{
		if (fullPlayers[TID_TEAMMATE][i]->isAlive() &&
			!fullPlayers[TID_TEAMMATE][i]->isBody() &&
			fullPlayers[TID_TEAMMATE][i]->getSeeTime() != getCurTime())
		{
			putOutOfVisibleDistance(*fullPlayers[TID_TEAMMATE][i]);

			if (fabs(fullPlayers[TID_TEAMMATE][i]->getHeadVec().getDirection()) <
				body->getViewAngle() - CHANGE_PLACE_SURE_ANGLE &&
				fullPlayers[TID_TEAMMATE][i]->getAbsVec().getMagnitude() <
				UNUM_FAR_LENGTH - CHANGE_PLACE_SURE_DIST)
				putOutOfVisibleAngle(*fullPlayers[TID_TEAMMATE][i]);

			changePosDevBySee(*fullPlayers[TID_TEAMMATE][i]);
		}

		if (fullPlayers[TID_OPPONENT][i]->isAlive() &&
			fullPlayers[TID_OPPONENT][i]->getSeeTime() != getCurTime())
		{
			putOutOfVisibleDistance(*fullPlayers[TID_OPPONENT][i]);

			if (fabs(fullPlayers[TID_OPPONENT][i]->getHeadVec().getDirection()) <
				body->getViewAngle() - CHANGE_PLACE_SURE_ANGLE &&
				fullPlayers[TID_OPPONENT][i]->getAbsVec().getMagnitude() <
				UNUM_FAR_LENGTH - CHANGE_PLACE_SURE_DIST)
				putOutOfVisibleAngle(*fullPlayers[TID_OPPONENT][i]);

			changePosDevBySee(*fullPlayers[TID_OPPONENT][i]);
		}
	}

	for (i = 0; i <= HALF_PLAYERS_NUM; i++)
	{
		if (halfPlayers[TID_TEAMMATE][i]->isValid() &&
			halfPlayers[TID_TEAMMATE][i]->getSeeTime() != getCurTime())
		{
			putOutOfVisibleDistance(*halfPlayers[TID_TEAMMATE][i]);

			if (fabs(halfPlayers[TID_TEAMMATE][i]->getHeadVec().getDirection()) <
				body->getViewAngle() - CHANGE_PLACE_SURE_ANGLE &&
				halfPlayers[TID_TEAMMATE][i]->getAbsVec().getMagnitude() <
				TEAM_FAR_LENGTH - CHANGE_PLACE_SURE_DIST)
				putOutOfVisibleAngle(*halfPlayers[TID_TEAMMATE][i]);

			changePosDevBySee(*halfPlayers[TID_TEAMMATE][i]);
		}

		if (halfPlayers[TID_OPPONENT][i]->isValid() &&
			halfPlayers[TID_OPPONENT][i]->getSeeTime() != getCurTime())
		{
			putOutOfVisibleDistance(*halfPlayers[TID_OPPONENT][i]);

			if (fabs(halfPlayers[TID_OPPONENT][i]->getHeadVec().getDirection()) <
				body->getViewAngle() - CHANGE_PLACE_SURE_ANGLE &&
				halfPlayers[TID_OPPONENT][i]->getAbsVec().getMagnitude() <
				TEAM_FAR_LENGTH - CHANGE_PLACE_SURE_DIST)
				putOutOfVisibleAngle(*halfPlayers[TID_OPPONENT][i]);

			changePosDevBySee(*halfPlayers[TID_OPPONENT][i]);
		}
	}

	// Ball change place
	if (ball->isValid() && ball->getSeeTime() != getCurTime())
	{
		putOutOfVisibleDistance(*ball);
		changePosDevBySee(*ball);
	}

}

void WorldModel::updateMarkingPlayers() 
{
    markers.clear();
    
    for (int i = 0; i < FULL_PLAYERS_NUM; ++i) {
        if (getFullPlayer(TID_OPPONENT, i).isValid()) {
            if (getBody().getDistance(getFullPlayer(TID_OPPONENT, i)) <= 7) {
                // saving  information related to OTHER PLAYERS
                unsigned unum = getFullPlayer(TID_OPPONENT,i).getUniNum();
                correlationMap[unum].second = getCurCycle();
                if (correlationMap[unum].first.size() >= 10) {
                    for (unsigned j = 1 ; j < correlationMap[unum].first.size(); ++j)
                        correlationMap[unum].first[j-1] = correlationMap[unum].first[j];
                    correlationMap[unum].first[correlationMap[unum].first.size()-1]=
                        getFullPlayer(TID_OPPONENT,i).getPos();
                } else {
                    correlationMap[unum].first.push_back(getFullPlayer(TID_OPPONENT,i).getPos());
                }
            }
        }
    }

    // advanced STL (use IQ...)
    for (map<unsigned, pair<vector<Vector>,unsigned> >::iterator
             i = correlationMap.begin();
             i != correlationMap.end();) {
        if ( i->second.second != getCurCycle()) {
            map<unsigned, pair<vector<Vector>,unsigned> >::iterator j = i;
            ++i;
            correlationMap.erase(j);
        } else {
            ++i;
        }
    }

    if (ourPosition.size() >= 10) {
        for (unsigned j = 1 ; j < ourPosition.size(); ++j)
            ourPosition[j-1] = ourPosition[j];
        ourPosition[9]= getBody().getPos();
    } else {
        ourPosition.push_back(getBody().getPos());
    }

    for (map<unsigned, pair<vector<Vector>,unsigned> >::iterator
             i = correlationMap.begin();
             i != correlationMap.end(); ++i) {
        Vector ourSigma, oppSigma;
        float markingWeight = 0;
        if (i->second.first.size() > 1) {
            // ready for descision
            float decay = 1;
            for (int c = i->second.first.size() - 1; c >= 0; --c) {
                if (ourPosition[c].asPoint().getDistance (
                        i->second.first[c].asPoint()) <= (MARKING_AREA * (getBody().getKickableMargin() + getBody().getSize())))
                {
                    markingWeight += decay;
                }
                decay *= MARKING_DECAY_FACTOR;
            }
        } else {
            LOG << "Not enough data to decide about statistics..." << endl;
        }
        if (markingWeight > MARKING_THRESHOLD) {
            markers.insert(i->first);
            LOG << "Adding " << i->first << " to markers..." << endl;
        }
        
    }
}

void WorldModel::putOutOfVisibleAngle(Object &object)
{
	float rotateAngle;

	LOG << "obj.angle: " << object.getHeadVec().getDirection() << endl;

	if (object.getHeadVec().getDirection() > 0)
		rotateAngle = body->getViewAngle() - 
			object.getHeadVec().getDirection() +
			CHANGE_PLACE_SURE_ANGLE;
	else
		rotateAngle = -body->getViewAngle() -
			object.getHeadVec().getDirection() -
			CHANGE_PLACE_SURE_ANGLE;

	object.setAbsVec().rotate(rotateAngle);
	object.setBodyVec().rotate(rotateAngle);
	object.setHeadVec().rotate(rotateAngle);
	object.setPos() = body->getPos() + object.getAbsVec();

	if (dynamic_cast<Ball *>(&object))
		LOG << "Complete updates: ball putted out of visible angle." << endl;
	if (Player *player = dynamic_cast<Player *>(&object))
		LOG << "Complete updates: player " << player->getTeamId() << " "
			<< player->getUniNum() << " putted out of visible angle." << endl;
}

void WorldModel::putOutOfVisibleDistance(Object &object)
{
	if (object.getAbsVec().getMagnitude() < body->getVisibleDistance())
	{
		object.setAbsVec().setAsPolar(body->getVisibleDistance() +
									  PUT_OUT_VIS_SURE_DIST, object.getAbsVec().getDirection());
		object.setBodyVec().setAsPolar(body->getVisibleDistance() +
									   PUT_OUT_VIS_SURE_DIST, object.getBodyVec().getDirection());
		object.setHeadVec().setAsPolar(body->getVisibleDistance() +
									   PUT_OUT_VIS_SURE_DIST, object.getHeadVec().getDirection());

		object.setPos() = body->getPos() + object.getAbsVec();

		if (dynamic_cast<Ball *>(&object))
			LOG << "Complete updates: ball putted out of sense area." << endl;
		if (Player *player = dynamic_cast<Player *>(&object))
			LOG << "Complete updates: player " << player->getTeamId() << " "
				<< player->getUniNum() << " putted out of sense area." << endl;
	}
}

void WorldModel::changePosDevBySee(Object &object)
{
	const Player *player = dynamic_cast<const Player *>(&object);
	if (player && player->isGoalie() &&
		player->getTeamId() == TID_OPPONENT)
	{
		if (object.getAbsVec().getMagnitude() < object.getPosDeviation() +
			1.5 + PUT_OUT_VIS_SURE_DIST)
			object.setPosDeviation(object.getAbsVec().getMagnitude() -
								   1.5 - PUT_OUT_VIS_SURE_DIST);
	}

	else if (object.getAbsVec().getMagnitude() < object.getPosDeviation() +
			 body->getVisibleDistance() + PUT_OUT_VIS_SURE_DIST)
		object.setPosDeviation(object.getAbsVec().getMagnitude() -
							   body->getVisibleDistance() - PUT_OUT_VIS_SURE_DIST);
}

void WorldModel::logObjects()
{
	unsigned i;

	LOG << "Body"
		<< " Pos" << body->getPos()
		<< " Vel" << body->getVel()
		<< " Ang" << Point(body->getBodyDir(), body->getHeadDir())
		<< " Poi" << body->getPointingDir() << endl;

	LOG << "Ball"
		<< " Pos" << ball->getPos()
		<< " Vel" << ball->getVel()
		<< " Ang" << Point(ball->getBodyVec().getDirection(),
						   ball->getHeadVec().getDirection())
		<< " Len" << ball->getAbsVec().getMagnitude()
		<< " PoD" << ball->getPosDeviation()
		<< " VPoD" << ball->getValidPosDeviation() << endl;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_TEAMMATE][i]->isValid() &&
			!fullPlayers[TID_TEAMMATE][i]->isBody())
			LOG << "OurF" << i+1
				<< " Pos" << fullPlayers[TID_TEAMMATE][i]->getPos()
				<< " Vel" << fullPlayers[TID_TEAMMATE][i]->getVel()
				<< " Ang" << Point(fullPlayers[TID_TEAMMATE][i]->getBodyDir(),
								   fullPlayers[TID_TEAMMATE][i]->getHeadDir())
				<< " Poi" << fullPlayers[TID_TEAMMATE][i]->getPointingDir()
				<< " Sta" << fullPlayers[TID_TEAMMATE][i]->getStamina()
				<< " PoD" << fullPlayers[TID_TEAMMATE][i]->getPosDeviation()
				<< " VPoD" << fullPlayers[TID_TEAMMATE][i]->getValidPosDeviation() << endl;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid())
			LOG << "OppF" << i+1
				<< " Pos" << fullPlayers[TID_OPPONENT][i]->getPos()
				<< " Vel" << fullPlayers[TID_OPPONENT][i]->getVel()
				<< " Ang" << Point(fullPlayers[TID_OPPONENT][i]->getBodyDir(),
								   fullPlayers[TID_OPPONENT][i]->getHeadDir())
				<< " Poi" << fullPlayers[TID_OPPONENT][i]->getPointingDir()
				<< " Sta" << fullPlayers[TID_OPPONENT][i]->getStamina()
				<< " PoD" << fullPlayers[TID_OPPONENT][i]->getPosDeviation()
				<< " VPoD" << fullPlayers[TID_OPPONENT][i]->getValidPosDeviation() << endl;

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_TEAMMATE][i]->isValid())
			LOG << "OurH" << i+1
				<< " Pos" << halfPlayers[TID_TEAMMATE][i]->getPos()
				<< " Vel" << halfPlayers[TID_TEAMMATE][i]->getVel()
				<< " Ang" << Point(halfPlayers[TID_TEAMMATE][i]->getBodyDir(),
								   halfPlayers[TID_TEAMMATE][i]->getHeadDir())
				<< " Poi" << halfPlayers[TID_TEAMMATE][i]->getPointingDir()
				<< " Sta" << halfPlayers[TID_TEAMMATE][i]->getStamina()
				<< " PoD" << halfPlayers[TID_TEAMMATE][i]->getPosDeviation()
				<< " VPoD" << halfPlayers[TID_TEAMMATE][i]->getValidPosDeviation() << endl;

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid())
			LOG << "OppH" << i+1
				<< " Pos" << halfPlayers[TID_OPPONENT][i]->getPos()
				<< " Vel" << halfPlayers[TID_OPPONENT][i]->getVel()
				<< " Ang" << Point(halfPlayers[TID_OPPONENT][i]->getBodyDir(),
								   halfPlayers[TID_OPPONENT][i]->getHeadDir())
				<< " Poi" << halfPlayers[TID_OPPONENT][i]->getPointingDir()
				<< " Sta" << halfPlayers[TID_OPPONENT][i]->getStamina()
				<< " PoD" << halfPlayers[TID_OPPONENT][i]->getPosDeviation()
				<< " VPoD" << halfPlayers[TID_OPPONENT][i]->getValidPosDeviation() << endl;

	for (i = 0; i < QUARTER_PLAYERS_NUM; i++)
		if (quarterPlayers[i]->isValid())
			LOG << "Quar" << i+1
				<< " Pos" << quarterPlayers[i]->getPos()
				<< " Vel" << quarterPlayers[i]->getVel()
				<< " Ang" << Point(quarterPlayers[i]->getBodyDir(),
								   quarterPlayers[i]->getHeadDir())
				<< " Poi" << quarterPlayers[i]->getPointingDir()
				<< " Sta" << quarterPlayers[i]->getStamina()
				<< " PoD" << quarterPlayers[i]->getPosDeviation()
				<< " VPoD" << quarterPlayers[i]->getValidPosDeviation() << endl;

}

void WorldModel::updateCurInterCalculate()
{
	curInterCalculate.startSession(0);
//	curInterCalculate.setLogging(true);
	curInterCalculate.updateByWorldModel(this);
	curInterCalculate.run();
	curInterCalculate.logCheckedPlayers();
}

void WorldModel::updateCurKickCalculate()
{
	curKickCalculate.update(this);
}

// Some Utilities
bool WorldModel::isBallKickable() const
{
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();

	if (getBall().getAbsVec().getMagnitude() <= kickableArea)
		return true;
	return false;
}

bool WorldModel::isBallKickable(const Ball &theBall) const
{
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();

	if (theBall.getAbsVec().getMagnitude() <= kickableArea)
		return true;
	return false;
}

bool WorldModel::isBallCatchable() const
{
	float catchableArea;
	catchableArea = hypot(getBody().getCatchableAreaL() , 
						  (getBody().getCatchableAreaW() / 2));

	if (getBall().getAbsVec().getMagnitude() < catchableArea)
		return true;
	return false;
}

bool WorldModel::isBallCatchable(const Ball &theBall) const
{
	float catchableArea;
	catchableArea = hypot(getBody().getCatchableAreaL() , 
						  (getBody().getCatchableAreaW() / 2));

	if (theBall.getAbsVec().getMagnitude() < catchableArea)
		return true;
	return false;
}

bool WorldModel::isBallInTmmKickable() const
{
	register int i;
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();
	
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE, i).isValid())
			if(!getFullPlayer(TID_TEAMMATE, i).isBody())
				if(getFullPlayer(TID_TEAMMATE, i).getDistance(getBall()) 
				   <=	kickableArea)
					return true;
					
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_TEAMMATE, i).isValid())
			if(!getHalfPlayer(TID_TEAMMATE, i).isBody())
				if(getHalfPlayer(TID_TEAMMATE, i).getDistance(getBall()) 
				   <=	kickableArea)
					return true;

	return false;
}

bool WorldModel::isBallInTmmKickable(const Ball &theBall) const
{
	register int i;
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();
	
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE, i).isValid())
			if(!getFullPlayer(TID_TEAMMATE, i).isBody())
				if(getFullPlayer(TID_TEAMMATE, i).getDistance(theBall) 
				   <=	kickableArea)
					return true;
					
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_TEAMMATE, i).isValid())
			if(!getHalfPlayer(TID_TEAMMATE, i).isBody())
				if(getHalfPlayer(TID_TEAMMATE, i).getDistance(theBall) 
				   <=	kickableArea)
					return true;

	return false;
}

bool WorldModel::isBallInOppKickable() const
{
	register int i;
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();
	
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT, i).isValid())
			if(getFullPlayer(TID_OPPONENT, i).getDistance(getBall()) 
			   <=	kickableArea)
				return true;
					
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT,i).isValid())
			if(getHalfPlayer(TID_OPPONENT,i).getDistance(getBall()) 
			   <=	kickableArea)
				return true;

	return false;
}

bool WorldModel::isBallInOppKickable(const Ball &theBall) const
{
	register int i;
	float kickableArea;
	kickableArea = getBall().getSize() + getBody().getKickableMargin() + 
		getBody().getSize();
	
	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT, i).isValid())
			if(getFullPlayer(TID_OPPONENT, i).getDistance(theBall) 
			   <=	kickableArea)
				return true;
					
	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT,i).isValid())
			if(getHalfPlayer(TID_OPPONENT,i).getDistance(theBall) 
			   <=	kickableArea)
				return true;

	return false;
}

bool WorldModel::isBallInField() const
{
	if (getBall().getPos().getX() < 52.5 + 0.5 && 
		getBall().getPos().getX() > -52.5 - 0.5 &&
		getBall().getPos().getY() < 34 + 0.5 && 
		getBall().getPos().getY() > -34 - 0.5)
		return true;
	return false;
}

bool WorldModel::isBallInField(const Ball &theBall)
{
	if (theBall.getPos().getX() < 52.5 + 0.5 && 
		theBall.getPos().getX() > -52.5 - 0.5 &&
		theBall.getPos().getY() < 34 + 0.5 && 
		theBall.getPos().getY() > -34 - 0.5)
		return true;
	return false;
}

float WorldModel::getBallOutDistance() const
{
	float xDistance = 0;
	float yDistance = 0;
	float distance;

	if (getBall().getPos().getX() < -52.5)
		xDistance = -getBall().getPos().getX() - 52.5;
	else if (getBall().getPos().getX() > 52.5)
		xDistance = +getBall().getPos().getX() - 52.5;

	if (getBall().getPos().getY() < -34)
		yDistance = -getBall().getPos().getY() - 34;
	else if (getBall().getPos().getY() > 34)
		yDistance = +getBall().getPos().getY() - 34;

	distance = hypot(xDistance, yDistance);
	return distance;
}

float WorldModel::getOutDistance(const Point &point)
{
	float xDistance = 0;
	float yDistance = 0;

	if (point.x < -52.5)
		xDistance = -point.x - 52.5;
	else if (point.x > 52.5)
		xDistance = +point.x - 52.5;

	if (point.y < -34)
		yDistance = -point.y - 34;
	else if (point.y > 34)
		yDistance = +point.y - 34;

	return hypot(xDistance, yDistance);
}

bool WorldModel::isOppBallShooted(const Ball &theBall) const
{
	Point interceptGoalLine, interceptGoal;

	if (theBall.getVel().getX() == 0)
		return 0;

	float delatX = -52.5 - theBall.getPos().getX();
	float delatY = (theBall.getVel().getY() * delatX) / theBall.getVel().getX();

	float delatX2 = -52.5 - theBall.getPos().getX();
	float delatY2 = (theBall.getVel().getY() * delatX) / theBall.getVel().getX();

	interceptGoalLine.x = theBall.getPos().getX() + delatX;
	interceptGoalLine.y = theBall.getPos().getY() + delatY;

	interceptGoal.x = theBall.getPos().getX() + delatX2;
	interceptGoal.y = theBall.getPos().getY() + delatY2;

	if (fabs(interceptGoal.y) > 8.5)
		return 0;

	Ball simBall;
	simBall = theBall;

	while (simBall.getVel().getMagnitude() > .4)
		simBall.simulateByDynamics(getBody());

///////////////////////	float standardX = GoalieXHome;		// This part must retype (GoalieXHome ?)!!!
	float standardX = -51;

	if (theBall.getPos().getX() < standardX)
		standardX = -52.5;

	if (simBall.getPos().getX() < standardX)
		return true;
	return false;
}

bool WorldModel::isOppBallShootedToTir(const Ball &theBall, float upYTirPoint, 
									   float downYTirPoint, Point &shootIntercept) const
{
	Point interceptGoalLine, interceptGoal;
	
	if (theBall.getVel().getX() == 0)
		return 0;

	float deltaX = -52.5 - theBall.getPos().getX();
	float deltaY = (theBall.getVel().getY() * deltaX) / theBall.getVel().getX();

	float deltaX2 = -52.5 - theBall.getPos().getX();
	float deltaY2 = (theBall.getVel().getY() * deltaX) / theBall.getVel().getX();

	interceptGoalLine.x = theBall.getPos().getX() + deltaX;
	interceptGoalLine.y = theBall.getPos().getY() + deltaY;
	
	interceptGoal.x = theBall.getPos().getX() + deltaX2;
	interceptGoal.y = theBall.getPos().getY() + deltaY2;

	
	if (interceptGoal.y < (downYTirPoint - 1) || interceptGoal.y > (upYTirPoint + 1))
		return 0;

	Ball simBall;
	simBall = theBall;

	while(simBall.getVel().getMagnitude() > .4)
		simBall.simulateByDynamics(getBody());

	float standardX = -51.5;

	if (theBall.getPos().getX() < standardX)
		standardX = -52.5;
	
	if (simBall.getPos().getX() < standardX)
	{
		shootIntercept = interceptGoal;
		return 1; 
	}
	return 0;
}

float WorldModel::getTmmDefenseLine() const
{
	float min = 0xFFFF;

	for (unsigned i = 1; i <= 5; i++)
		if (fullPlayers[TID_TEAMMATE][i - 1]->isValid())
			min = fmin(min, fullPlayers[TID_TEAMMATE][i - 1]->getPos().getX());

	if (fullPlayers[TID_TEAMMATE][8 - 1]->isValid())
		min = fmin(min, fullPlayers[TID_TEAMMATE][8 - 1]->getPos().getX());

	if (min == 0xFFFF)
		return -52.5;
	return min;
}

const Player &WorldModel::getNearestTmmToPoint(const Point & point) const
{
	register int i;
	int nearestPlayer = -1;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE, i).isValid())
		{
			if (nearestPlayer == -1)
				nearestPlayer = i;
			else if (point.getDistance(getFullPlayer(TID_TEAMMATE, i).getPos().asPoint()) <
                                 point.getDistance(getFullPlayer(TID_TEAMMATE, nearestPlayer).getPos().asPoint()))
				nearestPlayer = i;
		}

	if (nearestPlayer == -1)
		return getBody();
	else
		return getFullPlayer(TID_TEAMMATE, nearestPlayer);
}

const Player* WorldModel::getNearestTmmToPointFromList(const Point & point,
		std::string players, std::string blackPlayers) const
{
	register int i;
	int nearestPlayer = -1;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE, i).isValid() &&
				isInPlayers(getFullPlayer(TID_TEAMMATE, i).getUniNum(),
				players.c_str()) &&
				!isInPlayers(getFullPlayer(TID_TEAMMATE, i).getUniNum(),
				blackPlayers.c_str()))
		{
			if (nearestPlayer == -1)
				nearestPlayer = i;
			else if (point.getDistance(getFullPlayer(TID_TEAMMATE, i).getPos().asPoint()) <
                                 point.getDistance(getFullPlayer(TID_TEAMMATE, nearestPlayer).getPos().asPoint()))
				nearestPlayer = i;
		}

	if (nearestPlayer == -1)
		return NULL;
	else
		return fullPlayers[TID_TEAMMATE][nearestPlayer];
}

const Player &WorldModel::getNearestTmmToBall() const
{
	return getNearestTmmToPoint(getBall().getPos().asPoint());
}

const Player *WorldModel::getNearestOppToBall() const
{
	register int i;
	int nearestPlayer = -1;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT, i).isValid())
		{
			if (nearestPlayer == -1)
				nearestPlayer = i;
			else if (getFullPlayer(TID_OPPONENT, i).getDistance(getBall()) <
					 getFullPlayer(TID_OPPONENT, nearestPlayer).getDistance(getBall()))
				nearestPlayer = i;
		}

	if (nearestPlayer == -1)
		return NULL;
	else
		return &getFullPlayer(TID_TEAMMATE, nearestPlayer);
}


bool WorldModel::isPlayerInOurField(const Player &player)
{
	if (player.getPos().getX() <= 0)
		return true;
	return false;
}

bool WorldModel::isPlayerInOppField(const Player &player)
{
	if (player.getPos().getX() > 0)
		return true;
	return false;
}

bool WorldModel::isPlayerInOurShootArea(const Player &player)
{
	if (player.getPos().getX() < -20 &&
		player.getPos().getY() < 25 &&
		player.getPos().getY() > -25 )
		return true;
	return false;
}

bool WorldModel::isPlayerInOppShootArea(const Player &player)
{
	if (player.getPos().getX() > 30 &&
		player.getPos().getY() < 25 &&
		player.getPos().getY() > -25 )
		return true;
	return false;
}

float WorldModel::getOppOffsideLine(bool isBallFlag) const
{
	float offsideLine;
	unsigned i;
	bool isGoalieInList = false;
	
	for (int i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT,i).isValid() &&
			getFullPlayer(TID_OPPONENT,i).isGoalie())
			isGoalieInList = true;

	for (int i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
			getHalfPlayer(TID_OPPONENT,i).isGoalie())
			isGoalieInList = true;
			
			
	
	if (isGoalieInList)
	{
		offsideLine = -0xFFFF;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_OPPONENT,i).isValid() &&
				getFullPlayer(TID_OPPONENT,i).getPos().getX() > offsideLine &&
				!getFullPlayer(TID_OPPONENT,i).isGoalie()) 
				offsideLine = getFullPlayer(TID_OPPONENT,i).getPos().getX();

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
				getHalfPlayer(TID_OPPONENT,i).getPos().getX() > offsideLine &&
				!getHalfPlayer(TID_OPPONENT,i).isGoalie())
				offsideLine = getHalfPlayer(TID_OPPONENT,i).getPos().getX();
	}
	else
	{
		float firstMaxX = -0xFFFF, secMaxX = -0xFFFF;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_OPPONENT,i).isValid() &&
				getFullPlayer(TID_OPPONENT,i).getPos().getX() > secMaxX)
			{
				secMaxX = getFullPlayer(TID_OPPONENT,i).getPos().getX();
				if (secMaxX > firstMaxX)
					swap(firstMaxX, secMaxX);
			}

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
				getHalfPlayer(TID_OPPONENT,i).getPos().getX() > secMaxX)
			{
				secMaxX = getHalfPlayer(TID_OPPONENT,i).getPos().getX();
				if (secMaxX > firstMaxX)
					swap(firstMaxX, secMaxX);
			}
		offsideLine = secMaxX;		
	
	}

	if (isBallFlag)
		if (offsideLine < getBall().getPos().getX())
			offsideLine = getBall().getPos().getX();

	if (offsideLine < 0)
		offsideLine = 0;
		
	return offsideLine;
}

float WorldModel::getOppLocalOffLine() const
{
	// BE EXTREMELY CAREFULL.
	// This function DO NOT pay attention to the goalie and do not pay
	// attention to ball X and nehative offside lines (!) in calculating
	// local offside line.
	// NOT TO USE FOR POSITIONING.
	// It is specialized for breaking offside line.
	// This maybe return -0xFFFF.

//	bool isGoalieInList = false;
	
/*	for (int i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT,i).isValid() &&
			getFullPlayer(TID_OPPONENT,i).isGoalie())
			isGoalieInList = true;

	for (int i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
			getHalfPlayer(TID_OPPONENT,i).isGoalie())
			isGoalieInList = true;*/

//	if (isGoalieInList)
//	{
		float offsideLine = -0xFFFF;
		for (int i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_OPPONENT,i).isValid() &&
				getFullPlayer(TID_OPPONENT,i).getPos().getX() > offsideLine &&
				!getFullPlayer(TID_OPPONENT,i).isGoalie() &&
				(fabs(getFullPlayer(TID_OPPONENT,i).getPos().getY() - getBall().getPos().getY()) < OPP_OFFLINE_LOCAL_Y &&
				 fabs(getFullPlayer(TID_OPPONENT,i).getPos().getX() - getBall().getPos().getX()) < OPP_OFFLINE_LOCAL_X))
				offsideLine = getFullPlayer(TID_OPPONENT,i).getPos().getX();

		for (int i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
				getHalfPlayer(TID_OPPONENT,i).getPos().getX() > offsideLine &&
				!getHalfPlayer(TID_OPPONENT,i).isGoalie() &&
				(fabs(getHalfPlayer(TID_OPPONENT,i).getPos().getY() - getBall().getPos().getY()) < OPP_OFFLINE_LOCAL_Y &&
				 fabs(getHalfPlayer(TID_OPPONENT,i).getPos().getX() - getBall().getPos().getX()) < OPP_OFFLINE_LOCAL_X))
				offsideLine = getHalfPlayer(TID_OPPONENT,i).getPos().getX();
/*	}
	else
	{
		float firstMaxX = -0xFFFF, secMaxX = -0xFFFF;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_OPPONENT,i).isValid() &&
				getFullPlayer(TID_OPPONENT,i).getPos().getX() > secMaxX)
			{
				secMaxX = getFullPlayer(TID_OPPONENT,i).getPos().getX();
				if (secMaxX > firstMaxX)
					swap(firstMaxX, secMaxX);
			}

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
				getHalfPlayer(TID_OPPONENT,i).getPos().getX() > secMaxX)
			{
				secMaxX = getHalfPlayer(TID_OPPONENT,i).getPos().getX();
				if (secMaxX > firstMaxX)
					swap(firstMaxX, secMaxX);
			}
		offsideLine = secMaxX;		
	
	}

	if (isBallFlag)
		if (offsideLine < getBall().getPos().getX())
			offsideLine = getBall().getPos().getX();

	if (offsideLine < 0)
		offsideLine = 0;*/
		
	return offsideLine;
}

float WorldModel::getTmmOffsideLine(bool isBallFlag) const
{
	float offsideLine;
	unsigned i;
	bool isGoalieInList = false;
	
	for (int i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE,i).isValid() &&
			getFullPlayer(TID_TEAMMATE,i).isGoalie())
			isGoalieInList = true;

	for (int i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_TEAMMATE,i).isValid() &&
			getHalfPlayer(TID_TEAMMATE,i).isGoalie())
			isGoalieInList = true;
			
			
	
	if (isGoalieInList)
	{
		offsideLine = 0xFFFF;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_TEAMMATE,i).isValid() &&
				getFullPlayer(TID_TEAMMATE,i).getPos().getX() < offsideLine &&
				!getFullPlayer(TID_TEAMMATE,i).isGoalie()) 
				offsideLine = getFullPlayer(TID_TEAMMATE,i).getPos().getX();

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_TEAMMATE,i).isValid() &&
				getHalfPlayer(TID_TEAMMATE,i).getPos().getX() < offsideLine &&
				!getHalfPlayer(TID_TEAMMATE,i).isGoalie())
				offsideLine = getHalfPlayer(TID_TEAMMATE,i).getPos().getX();
	}
	else
	{
		float firstMaxX = 0xFFFF, secMaxX = 0xFFFF;
		for (i = 0; i < FULL_PLAYERS_NUM; i++)
			if (getFullPlayer(TID_TEAMMATE,i).isValid() &&
				getFullPlayer(TID_TEAMMATE,i).getPos().getX() < secMaxX)
			{
				secMaxX = getFullPlayer(TID_TEAMMATE,i).getPos().getX();
				if (secMaxX < firstMaxX)
					swap(firstMaxX, secMaxX);
			}

		for (i = 0; i < HALF_PLAYERS_NUM; i++)
			if (getHalfPlayer(TID_TEAMMATE,i).isValid() &&
				getHalfPlayer(TID_TEAMMATE,i).getPos().getX() < secMaxX)
			{
				secMaxX = getHalfPlayer(TID_TEAMMATE,i).getPos().getX();
				if (secMaxX < firstMaxX)
					swap(firstMaxX, secMaxX);
			}
		offsideLine = secMaxX;		
	
	}

	if (isBallFlag)
		if (offsideLine > getBall().getPos().getX())
			offsideLine = getBall().getPos().getX();

	if (offsideLine > 0)
		offsideLine = 0;
		
	return offsideLine;
}


int WorldModel::getOppPlayersBeforeLine(float lineX) const
{
	int counter = 0;
	register int i;

	for (i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT,i).isValid() &&
			getFullPlayer(TID_OPPONENT,i).getPos().getX() <= lineX)
		{
			if (!getFullPlayer(TID_OPPONENT,i).isGoalie())
				counter++;
			else
				counter += 3;
		}

	for (i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT,i).isValid() &&
			getHalfPlayer(TID_OPPONENT,i).getPos().getX() <= lineX)
		{
			if (!getHalfPlayer(TID_OPPONENT,i).isGoalie())
				counter++;
			else
				counter += 3;
		}
	
	return counter;
}

int WorldModel::getOppPlayersBetweenLines(float line1X, float line2X) const
{
	return getOppPlayersBeforeLine(line2X) - getOppPlayersBeforeLine(line1X);
}

const Player *WorldModel::getTmmGoalie() const
{
	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_TEAMMATE, i).isValid() &&
			getFullPlayer(TID_TEAMMATE, i).isGoalie())
			return fullPlayers[TID_TEAMMATE][i];

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_TEAMMATE, i).isValid() &&
			getHalfPlayer(TID_TEAMMATE, i).isGoalie())
			return halfPlayers[TID_TEAMMATE][i];

	return NULL;
}

const Player *WorldModel::getOppGoalie() const
{
	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (getFullPlayer(TID_OPPONENT, i).isValid() &&
			getFullPlayer(TID_OPPONENT, i).isGoalie())
			return fullPlayers[TID_OPPONENT][i];

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (getHalfPlayer(TID_OPPONENT, i).isValid() &&
			getHalfPlayer(TID_OPPONENT, i).isGoalie())
			return halfPlayers[TID_OPPONENT][i];

	return NULL;
}

Point WorldModel::getOppDefenseMedium()  const
{
	Point mid(0, 0);
	int counter = 0;

	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid() &&
				!fullPlayers[TID_OPPONENT][i]->isGoalie())
		{
			mid.y += fullPlayers[TID_OPPONENT][i]->getPos().getY();
			mid.x += fullPlayers[TID_OPPONENT][i]->getPos().getX();
			counter++;
		}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid() &&
				!halfPlayers[TID_OPPONENT][i]->isGoalie())
		{
			mid.y += halfPlayers[TID_OPPONENT][i]->getPos().getY();
			mid.x += halfPlayers[TID_OPPONENT][i]->getPos().getX();
			counter++;
		}
	
	counter = (counter)?counter:1;
	mid.y /= counter;
	mid.x /= counter;
	return mid;
}

Point WorldModel::getOppDefenseYMin()  const
{
	Point yMin(0, 100);

	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid() &&
				!fullPlayers[TID_OPPONENT][i]->isGoalie())
			if (fullPlayers[TID_OPPONENT][i]->getPos().getY() < yMin.y)
			{
				yMin.y = fullPlayers[TID_OPPONENT][i]->getPos().getY();
				yMin.x = fullPlayers[TID_OPPONENT][i]->getPos().getX();
			}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid() &&
				!halfPlayers[TID_OPPONENT][i]->isGoalie())
			if (halfPlayers[TID_OPPONENT][i]->getPos().getY() < yMin.y)
			{
				yMin.y = halfPlayers[TID_OPPONENT][i]->getPos().getY();
				yMin.x = halfPlayers[TID_OPPONENT][i]->getPos().getX();
			}
	
	return yMin;
}

Point WorldModel::getOppDefenseYMax()  const
{
	Point yMax(0, -100);

	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid() &&
				!fullPlayers[TID_OPPONENT][i]->isGoalie())
			if (fullPlayers[TID_OPPONENT][i]->getPos().getY() > yMax.y)
			{
				yMax.y = fullPlayers[TID_OPPONENT][i]->getPos().getY();
				yMax.x = fullPlayers[TID_OPPONENT][i]->getPos().getX();
			}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid() &&
				!halfPlayers[TID_OPPONENT][i]->isGoalie())
			if (halfPlayers[TID_OPPONENT][i]->getPos().getY() > yMax.y)
			{
				yMax.y = halfPlayers[TID_OPPONENT][i]->getPos().getY();
				yMax.x = halfPlayers[TID_OPPONENT][i]->getPos().getX();
			}
	
	return yMax;
}

float WorldModel::getSecurityStatus(const Player &player, float oval_a,
		float oval_b) const
{
	float securityStatus = 100000;
	float tmpSecurityStatus = 0;
	if (!player.isValid()) return 0;
	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid())
		{
			float oppPlayerX = fullPlayers[TID_OPPONENT][i]->getPos().getX() -
					player.getPos().getX();
			float oppPlayerY = fullPlayers[TID_OPPONENT][i]->getPos().getY() -
					player.getPos().getY();
			tmpSecurityStatus = abs(2 * oval_b *
					oppPlayerX * oppPlayerY *	sqrt(abs(
					1 / (oval_a * oppPlayerY) + 1 / (oval_b * oppPlayerX))));
			if (abs(oppPlayerX) < .05)
				tmpSecurityStatus = abs(2 * (oval_b / oval_a) * oppPlayerY);
			if (abs(oppPlayerY) < .05)
				tmpSecurityStatus = abs(2 * oppPlayerX);
			if (tmpSecurityStatus < securityStatus)
				securityStatus = tmpSecurityStatus;
		}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid())
		{
			float oppPlayerX = halfPlayers[TID_OPPONENT][i]->getPos().getX() -
					player.getPos().getX();
			float oppPlayerY = halfPlayers[TID_OPPONENT][i]->getPos().getY() -
					player.getPos().getY();
			tmpSecurityStatus = abs(2 * oval_b * oppPlayerX * oppPlayerY *
					sqrt(1 / (oval_a * oppPlayerY) +
					1 / (oval_a * oppPlayerX)));
			if (abs(oppPlayerX) < .05)
				tmpSecurityStatus = abs(2 * (oval_b / oval_a) * oppPlayerY);
			if (abs(oppPlayerY) < .05)
				tmpSecurityStatus = abs(2 * oppPlayerX);
			if (tmpSecurityStatus < securityStatus)
				securityStatus = tmpSecurityStatus;
		}
	return securityStatus;
}

float WorldModel::getPathSecurityStatus(const Player &srcPlayer,
		const Point trgPoint, float forwardRate) const
{
	float minAngleSecurity = 100000;
	Vector meToTRGPoint;
	meToTRGPoint.setByPoints(srcPlayer.getPos(), trgPoint);
	float trgDirection = absoluteAngle(meToTRGPoint.getDirection());
	for (unsigned i = 0; i < FULL_PLAYERS_NUM; i++)
		if (fullPlayers[TID_OPPONENT][i]->isValid())
		{
			Vector meToTmpOpp;
			meToTmpOpp.setByPoints(srcPlayer.getPos(),
					fullPlayers[TID_OPPONENT][i]->getPos());
			if (meToTmpOpp.getMagnitude() <= meToTRGPoint.getMagnitude() + 1)
			{
				float tmpAngle = absoluteAngle(abs(trgDirection -
						absoluteAngle(meToTmpOpp.getDirection())));
				float tmpAngleSecurity = DEG2RAD * tmpAngle;
				if ((trgDirection > meToTmpOpp.getDirection() &&
						normalizeAngle(trgDirection) >= 0) ||
						(trgDirection < meToTmpOpp.getDirection() &&
						normalizeAngle(trgDirection) <= 0))
					tmpAngleSecurity = DEG2RAD * tmpAngle + (forwardRate - 1) *
							DEG2RAD * tmpAngle * (sin(DEG2RAD * tmpAngle) + 1);
				if (tmpAngleSecurity < minAngleSecurity)
					minAngleSecurity = tmpAngleSecurity;
			}
		}

	for (unsigned i = 0; i < HALF_PLAYERS_NUM; i++)
		if (halfPlayers[TID_OPPONENT][i]->isValid())
		{
			Vector meToTmpOpp;
			meToTmpOpp.setByPoints(srcPlayer.getPos(),
					halfPlayers[TID_OPPONENT][i]->getPos());
			if (meToTmpOpp.getMagnitude() >= meToTRGPoint.getMagnitude() + 4)
			{
				float tmpAngle = absoluteAngle(trgDirection - meToTmpOpp.getDirection());
				float tmpAngleSecurity = DEG2RAD * tmpAngle;
				if ((trgDirection > meToTmpOpp.getDirection() &&
						normalizeAngle(trgDirection) >= 0) ||
						(trgDirection < meToTmpOpp.getDirection() &&
						normalizeAngle(trgDirection) <= 0))
					tmpAngleSecurity = DEG2RAD * tmpAngle + (forwardRate - 1) *
							DEG2RAD * tmpAngle * (sin(DEG2RAD * tmpAngle) + 1);
				if (tmpAngleSecurity < minAngleSecurity)
					minAngleSecurity = tmpAngleSecurity;
			}
		}
	return minAngleSecurity;
}

void WorldModel::parseHear(const SExpression &exp)
{
	float hearDir;
	unsigned hearCycle;
	unsigned senderNum;
	SExpAtomic *at;
	SExpression *exp2;

	resetHearedVarsBeforeHear();

	at = dynamic_cast<SExpAtomic *>(exp[1]);
	assert(at);
	hearCycle = at->asInt();
	hearedPlanInfo = "";

	at = dynamic_cast<SExpAtomic *>(exp[2]);
	assert(at);

	if (at->toString() == "referee")
	{
		parseRefereeHear(exp);
		refereeHearTime = curTime;
	}
	else if (at->toString().substr(0, 12) == "online_coach")
	{
		if ((fieldSide == SI_LEFT && at->toString().substr(13, 4) == "left") ||
			(fieldSide == SI_RIGHT && at->toString().substr(13, 5) == "right"))
		{ // teammate coach.
			exp2 = dynamic_cast<SExpression *>(exp[3]);
			assert(exp2);
			parseTeammateCoachHear(*exp2);
		}
		else
		{ // opponent coach.
		}
	}
	else if (at->toString() == "self")
	{
		at = dynamic_cast<SExpAtomic *>(exp[3]);
		assert(at);
		selfHear(at->toString());
	}
	else // a player has said that.
	{
		at = dynamic_cast<SExpAtomic *>(exp[2]);
		assert(at);
		hearDir = at->asInt();

		at = dynamic_cast<SExpAtomic *>(exp[3]);
		assert(at);
		if (at->toString() == "our")
		{
			at = dynamic_cast<SExpAtomic *>(exp[4]);
			assert(at);
			senderNum = at->asInt();

			at = dynamic_cast<SExpAtomic *>(exp[5]);
			assert(at);
			teammateHear(hearDir, senderNum, at->toString());
		}
		else
		{
			assert(at->toString() == "opp");

			at = dynamic_cast<SExpAtomic *>(exp[4]);
			assert(at);
			opponentHear(hearDir, at->toString());
		}
	}

}

void WorldModel::parseRefereeHear(const SExpression &exp)
{
	SExpAtomic *at;

	at = dynamic_cast<SExpAtomic *>(exp[3]);
	assert(at);

	playModeRemainCycle = 0;

	// normal PlayModes
	if (at->toString() == "before_kick_off")
		playMode = PM_BEFORE_KICK_OFF;

	else if (at->toString() == "play_on")
	{
		playMode = PM_PLAY_ON;
		virtualPlayMode = VPM_NONE;
		playOnStartCycle = curTime;
	}

	else if (at->toString() == "kick_off_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_KICK_OFF;
		else
			playMode = PM_KICK_OFF_OPP;

		playModeRemainCycle = 200;

		ball->updateByHear(0, 0, 0, 0, *body);
	}
	else if (at->toString() == "kick_off_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_KICK_OFF;
		else
			playMode = PM_KICK_OFF_OPP;

		playModeRemainCycle = 200;
		ball->updateByHear(0, 0, 0, 0, *body);
	}

	else if (at->toString() == "free_kick_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_FREE_KICK;
		else
			playMode = PM_FREE_KICK_OPP;

		playModeRemainCycle = 200;
	}
	else if (at->toString() == "free_kick_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_FREE_KICK;
		else
			playMode = PM_FREE_KICK_OPP;

		playModeRemainCycle = 200;
	}

	else if (at->toString() == "indirect_free_kick_l")
	{
		if (fieldSide == SI_LEFT)
		{
			playMode = PM_FREE_KICK;
			virtualPlayMode = VPM_INDIRECT_FREE_KICK;
		}
		else
		{
			playMode = PM_FREE_KICK_OPP;
			virtualPlayMode = VPM_INDIRECT_FREE_KICK_OPP;
		}

		playModeRemainCycle = 200;
	}
	else if (at->toString() == "indirect_free_kick_r")
	{
		if (fieldSide == SI_RIGHT)
		{
			playMode = PM_FREE_KICK;
			virtualPlayMode = VPM_INDIRECT_FREE_KICK;
		}
		else
		{
			playMode = PM_FREE_KICK_OPP;
			virtualPlayMode = VPM_INDIRECT_FREE_KICK_OPP;
		}

		playModeRemainCycle = 200;
	}

	else if (at->toString() == "kick_in_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_KICK_IN;
		else
			playMode = PM_KICK_IN_OPP;

		playModeRemainCycle = 200;
	}
	else if (at->toString() == "kick_in_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_KICK_IN;
		else
			playMode = PM_KICK_IN_OPP;

		playModeRemainCycle = 200;
	}

	else if (at->toString() == "offside_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_OFFSIDE;
		else
			playMode = PM_OFFSIDE_OPP;

		playModeRemainCycle = 30;
	}
	else if (at->toString() == "offside_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_OFFSIDE;
		else
			playMode = PM_OFFSIDE_OPP;

		playModeRemainCycle = 30;
	}

	else if (at->toString() == "goal_kick_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_GOAL_KICK;
		else
			playMode = PM_GOAL_KICK_OPP;

		virtualPlayMode = VPM_NONE;
		playModeRemainCycle = 200;
	}
	else if (at->toString() == "goal_kick_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_GOAL_KICK;
		else
			playMode = PM_GOAL_KICK_OPP;

		virtualPlayMode = VPM_NONE;
		playModeRemainCycle = 200;
	}

	else if (at->toString() == "corner_kick_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_CORNER_KICK;
		else
			playMode = PM_CORNER_KICK_OPP;

		playModeRemainCycle = 200;
	}
	else if (at->toString() == "corner_kick_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_CORNER_KICK;
		else
			playMode = PM_CORNER_KICK_OPP;

		playModeRemainCycle = 200;
	}

	else if (at->toString() == "free_kick_fault_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_FREE_KICK_FAULT;
		else
			playMode = PM_FREE_KICK_FAULT_OPP;

		playModeRemainCycle = 30;
	}
	else if (at->toString() == "free_kick_fault_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_FREE_KICK_FAULT;
		else
			playMode = PM_FREE_KICK_FAULT_OPP;

		playModeRemainCycle = 30;
	}

	else if (at->toString() == "catch_fault_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_CATCH_FAULT;
		else
			playMode = PM_CATCH_FAULT_OPP;

		playModeRemainCycle = 30;
	}
	else if (at->toString() == "catch_fault_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_CATCH_FAULT;
		else
			playMode = PM_CATCH_FAULT_OPP;

		playModeRemainCycle = 30;
	}

	else if (at->toString() == "penalty_setup_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_PENALTY_SETUP;
		else
			playMode = PM_PENALTY_SETUP_OPP;

		playModeRemainCycle = 100;
		virtualPlayMode = VPM_NONE;
	}
	else if (at->toString() == "penalty_setup_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_PENALTY_SETUP;
		else
			playMode = PM_PENALTY_SETUP_OPP;

		playModeRemainCycle = 100;
		virtualPlayMode = VPM_NONE;
	}

	else if (at->toString() == "penalty_ready_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_PENALTY_READY;
		else
			playMode = PM_PENALTY_READY_OPP;

		playModeRemainCycle = 50;
		virtualPlayMode = VPM_NONE;
	}
	else if (at->toString() == "penalty_ready_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_PENALTY_READY;
		else
			playMode = PM_PENALTY_READY_OPP;

		playModeRemainCycle = 50;
		virtualPlayMode = VPM_NONE;
	}

	else if (at->toString().substr(0, 6) == "goal_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_GOAL;
		else
			playMode = PM_GOAL_OPP;

		playModeRemainCycle = 50;
	}
	else if (at->toString().substr(0, 6) == "goal_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_GOAL;
		else
			playMode = PM_GOAL_OPP;

		playModeRemainCycle = 50;
	}

	// virtual PlayModes
	else if (at->toString() == "drop_ball")
	{
		virtualPlayMode = VPM_DROP_BALL;
		ball->setPosDeviation(MAX_POS_DEV);
		ball->setSeePosDeviation(MAX_POS_DEV);
		ball->setValidPosDeviation(MAX_POS_DEV);
	}

	else if (at->toString() == "time_up_without_a_team")
		playMode = PM_TIME_UP_WITHOUT_A_TEAM;

	else if (at->toString() == "time_up")
		playMode = PM_TIME_UP;

	else if (at->toString() == "time_over")
		playMode = PM_TIME_UP;
	
	else if (at->toString() == "half_time")
		virtualPlayMode = VPM_HALF_TIME;

	else if (at->toString() == "time_extended")
		virtualPlayMode = VPM_TIME_EXTENDED;

	else if (at->toString() == "penalty_draw")
		virtualPlayMode = VPM_PENALTY_DRAW;

	else if (at->toString() == "goalie_catch_ball_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_GOALIE_CATCH_BALL;
		else
			virtualPlayMode = VPM_GOALIE_CATCH_BALL_OPP;
	}
	else if (at->toString() == "goalie_catch_ball_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_GOALIE_CATCH_BALL;
		else
			virtualPlayMode = VPM_GOALIE_CATCH_BALL_OPP;
	}

	else if (at->toString() == "foul_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_FOUL;
		else
			virtualPlayMode = VPM_FOUL_OPP;
	}
	else if (at->toString() == "foul_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_FOUL;
		else
			virtualPlayMode = VPM_FOUL_OPP;
	}

	else if (at->toString() == "back_pass_l")
	{
		if (fieldSide == SI_LEFT)
			playMode = PM_BACK_PASS;
		else
			playMode = PM_BACK_PASS_OPP;
	}
	else if (at->toString() == "back_pass_r")
	{
		if (fieldSide == SI_RIGHT)
			playMode = PM_BACK_PASS;
		else
			playMode = PM_BACK_PASS_OPP;
	}

	else if (at->toString() == "penalty_onfield_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_ON_FIELD;
		else
			virtualPlayMode = VPM_PENALTY_ON_FIELD_OPP;
	}
	else if (at->toString() == "penalty_onfield_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_ON_FIELD;
		else
			virtualPlayMode = VPM_PENALTY_ON_FIELD_OPP;
	}

	else if (at->toString() == "penalty_miss_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_MISS;
		else
			virtualPlayMode = VPM_PENALTY_MISS_OPP;
	}
	else if (at->toString() == "penalty_miss_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_MISS;
		else
			virtualPlayMode = VPM_PENALTY_MISS_OPP;
	}

	else if (at->toString() == "penalty_score_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_SCORE;
		else
			virtualPlayMode = VPM_PENALTY_SCORE_OPP;
	}
	else if (at->toString() == "penalty_score_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_SCORE;
		else
			virtualPlayMode = VPM_PENALTY_SCORE_OPP;
	}

	else if (at->toString() == "penalty_winner_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_WINNER;
		else
			virtualPlayMode = VPM_PENALTY_WINNER_OPP;
	}
	else if (at->toString() == "penalty_winner_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_WINNER;
		else
			virtualPlayMode = VPM_PENALTY_WINNER_OPP;
	}

	else if (at->toString() == "penalty_foul_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_FOUL;
		else
			virtualPlayMode = VPM_PENALTY_FOUL_OPP;
	}
	else if (at->toString() == "penalty_foul_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_FOUL;
		else
			virtualPlayMode = VPM_PENALTY_FOUL_OPP;
	}

	else if (at->toString() == "penalty_taken_l")
	{
		if (fieldSide == SI_LEFT)
			virtualPlayMode = VPM_PENALTY_TAKEN;
		else
			virtualPlayMode = VPM_PENALTY_TAKEN_OPP;
	}
	else if (at->toString() == "penalty_taken_r")
	{
		if (fieldSide == SI_RIGHT)
			virtualPlayMode = VPM_PENALTY_TAKEN;
		else
			virtualPlayMode = VPM_PENALTY_TAKEN_OPP;
	}

	else
		assert(0);
}

void WorldModel::selfHear(const std::string &message)
{
	if ((message[1] == 'A' || message[1] == 'B') &&
		curTime != refereeHearTime + 1)
	{
		hearedHeader = message[1];
		hearedSender = body->getUniNum();
		hearedInterceptNum = SayEncoding::getCharNum(message[8]);
// 		hearedRequestPassNum = SayEncoding::getCharNum(message[9]); AHMAD
	}
}

void WorldModel::teammateHear(float hearDir, unsigned senderNum,
							  const string &message)
{
	hearedHeader = message[1];
	
	switch (message[1])
	{
	case 'X':
			planTmmHear(message, senderNum);
			break;

	case 'W':
			planTmmHear(message, senderNum);
			break;

	case 'c':
			planTmmHear(message, senderNum);
			break;

	case 'A':
			beforePassTmmHear(message, senderNum);
			break;

	case 'B':
			beforePassRelatedTmmHear(message, senderNum);
			break;

	case 'C':
			suggestPassTmmHear(message, senderNum);

/*	case 'Y':
			if (curTime != refereeHearTime + 1)
				routinTmmHearWithoutBall(message, senderNum);
			break;*/

	case 'Z':
			if (curTime != refereeHearTime + 1)
				routinTmmHear(message, senderNum);
			break;

	case 'F':
			if (curTime != refereeHearTime + 1)
				freeKickKickerHear(message, senderNum);
			break;

	case 'R':
		if (curTime != refereeHearTime + 1)
			radarHear(message, senderNum);
		break;

	default:
			assert(0);
	}

	logObjects();
}

void WorldModel::opponentHear(float hearDir, const string &message)
{
}

void WorldModel::routinTmmHear(const std::string &message,
		unsigned senderNum)
{
	float x, y, velMag, velDir;

	if (SayEncoding::getCharNum(message[8]) >= SAY_UPPER_CASE_NUM &&
		!isBallKickable()) // It means sender now has the ball.
	{
		// Updating the ball
		SayEncoding::decodeObjectPosition(message, 2, x, y);
		SayEncoding::decodeObjectVelocity(message, 5, velMag, velDir);

		ball->updateByHear(x, y, velMag, velDir, *body);

		// Updating the sender
		fullPlayers[TID_TEAMMATE][senderNum - 1]->
				updateByHear(x, y, velMag, velDir, *body);

		hearedInterceptNum = SayEncoding::getCharNum(message[8]) -
				SAY_UPPER_CASE_NUM;
	}

	else if (ball->getSeePosDeviation() > 2 && !isBallKickable() &&
		ball->isUpdateByHearPermitted())
	{
		// Updating the ball
		SayEncoding::decodeObjectPosition(message, 2, x, y);
		SayEncoding::decodeObjectVelocity(message, 5, velMag, velDir);

		ball->updateByHear(x, y, velMag, velDir, *body);

		hearedInterceptNum = SayEncoding::getCharNum(message[8]);
	}

	hearedSender = senderNum;
	planStatusSender = hearedSender;
	hearedPlanStatus = SayEncoding::getCharNum(message[9]);

//	mustAttention = SayEncoding::getCharNum(message[10]);
}

/*void WorldModel::routinTmmHearWithoutBall(const std::string &message,
		unsigned senderNum)
{
	hearedSender = senderNum;
	hearedInterceptNum = SayEncoding::getCharNum(message[2]);
	hearedRequestPassNum = SayEncoding::getCharNum(message[3]);

	mustAttention = SayEncoding::getCharNum(message[4]);
}*/

void WorldModel::beforePassTmmHear(const std::string &message,
								   unsigned senderNum)
{
	float x, y, velMag, velDir;

	SayEncoding::decodeObjectPosition(message, 2, x, y);
	SayEncoding::decodeObjectVelocity(message, 5, velMag, velDir);

	hearedSender = senderNum;
	hearedInterceptNum = SayEncoding::getCharNum(message[8]);
	ball->updateByHear(x, y, velMag, velDir, *body);
}

void WorldModel::beforePassRelatedTmmHear(const std::string &message,
										  unsigned senderNum)
{
	float x, y, velMag, velDir;

	SayEncoding::decodeObjectPosition(message, 2, x, y);
	SayEncoding::decodeObjectVelocity(message, 5, velMag, velDir);

	hearedSender = senderNum;
	hearedInterceptNum = SayEncoding::getCharNum(message[8]);

	unsigned num = hearedInterceptNum;
	if (num > 11)
		num -= 11;

	if (fullPlayers[TID_TEAMMATE][num - 1]->isValid())
		ball->updateByHear(
			fullPlayers[TID_TEAMMATE][num - 1]->getPos().getX() + x,
			fullPlayers[TID_TEAMMATE][num - 1]->getPos().getY() + y,
			velMag, velDir, *body);
}

void WorldModel::suggestPassTmmHear(const std::string &message,
									unsigned senderNum)
{
	float magnitude, direction, weight;

	SayEncoding::decodeObjectVelocity(message, 2, magnitude, direction);
	SayEncoding::decodeWeight(message, 5, weight, -50, 150);

	hearedSender = senderNum;
	hearedSuggestPassNum = SayEncoding::getCharNum(message[8]);
	hearedSuggestPassDir = direction;
	hearedSuggestPassSpeed = magnitude;
	hearedSuggestPassWeight = weight;

	hearedSuggestPassTime = curTime;
}

void WorldModel::parseTeammateCoachHear(const SExpression &exp)
{
	string messageString;
	string expString;
	unsigned dummyPtr = 0;

	int i, j;
	int typeNum;

	SExpression *exp2;
	SExpression *exp3;
	SExpAtomic *at;
	SExpAtomic *at2;

	LOG << "WorldModel::parseTeammateCoachHear" << endl;

	if (dynamic_cast<const SExpAtomic *>(exp[0])->toString() == "freeform")
	{
		expString = dynamic_cast<const SExpAtomic *>(exp[1])->toString();
		expString = "(" + expString.substr(1, expString.length() - 2) + ")";
		exp2 = new SExpression(expString, dummyPtr);
		for (i = 0; i < exp2->size(); i++)
		{
			exp3 = dynamic_cast<SExpression *>((*exp2)[i]);
			at = dynamic_cast<SExpAtomic *>((*exp3)[0]);
			assert(at);

			// Opponent Types (T)
			if (at->toString() == "T")
			{
				at2 = dynamic_cast<SExpAtomic *>((*exp3)[1]);
				assert(at2);
				messageString = at2->toString();
				assert(messageString.length() == 11);
				for (j = 0; j < 11; j++)
				{
					typeNum = messageString[j] - '0';
					setFullPlayer(TID_OPPONENT, j).setType(typeNum,	playerTypes[typeNum]);
					LOG << "\tAOTR: Opp Num " << j + 1 << " setted to type " << typeNum << endl;
				}
			}
		}
	}
	
}

void WorldModel::freeKickKickerHear(const std::string &message,
									unsigned senderNum)
{
	unsigned adviceNumber;
	
	adviceNumber = SayEncoding::getCharNum(message[2]);

	LOG << "WorldModel::freeKickAdvicerHear sender=" << senderNum <<
		" adviceNumber=" << adviceNumber << endl;

	freeKickAdvicerSayCycle = curTime;
	freeKickAdvicerSayNumber = adviceNumber;
}

void WorldModel::planTmmHear(const string &message, unsigned senderNum) 
{
	hearedPlanInfo = message.substr(1, message.size() - 2);
	hearedPlanTime = curTime;
	hearedPlanSender = senderNum;

	vector<Vector> recivedPositions;
	if (hearedPlanInfo[0] == 'W')
	{
		SayEncoding::decodeWMSend(hearedPlanInfo.substr(1,9), recivedPositions);
		for (vector <Vector>::iterator i = recivedPositions.begin() + 3;
				i != recivedPositions.end(); i++)
			i->setAsPolar(i->getMagnitude(), -i->getDirection());

		for (vector <Vector>::iterator i = recivedPositions.begin() + 1;
				i != recivedPositions.end(); i++)
			if (i->getMagnitude() != 19.75)
				i->rotate(recivedPositions[0].getDirection());

		for (vector <Vector>::iterator i = recivedPositions.begin();
				i != recivedPositions.end(); i++)
			if (i->getMagnitude() != 19.75)
			{
				*i += getNearestTmmToBall().getPos();
				LOG << "Heared Player:" << *i << endl;
			}

		unsigned player = 0;
		if (isInPlayers(body->getUniNum(), "27"))
			player = 10;
		else if (isInPlayers(body->getUniNum(), "48"))
			player = 11;
		else if (body->getUniNum() == 3)
			player = (body->getPos().getY() > 0)?10:11;

		if (player)
		{
			fullPlayers[TID_TEAMMATE][player - 1]->updateByHear(recivedPositions[0].asPoint().x,
					recivedPositions[0].asPoint().y, 0, 0, *body);
			fullPlayers[TID_TEAMMATE][player - 1]->setModel(PLM_FULL);
			fullPlayers[TID_TEAMMATE][player - 1]->setPosDeviation(.5);
			fullPlayers[TID_TEAMMATE][player - 1]->setSeeTime(curTime);
		}

		for (vector <Vector>::iterator i = recivedPositions.begin() + 1;
				i != recivedPositions.end(); i++)
			if (i->getMagnitude() != 19.75)
				for (unsigned j = 0; j < HALF_PLAYERS_NUM; j++)
					if (!halfPlayers[TID_OPPONENT][j]->isValid() &&
						!halfPlayers[TID_OPPONENT][j]->isReadyToUpdate())
					{
						halfPlayers[TID_OPPONENT][j]->updateByHear(i->asPoint().x,
								i->asPoint().y, 0, 0, *body);
						halfPlayers[TID_OPPONENT][j]->setModel(PLM_HALF);
						halfPlayers[TID_OPPONENT][j]->setPosDeviation(.5);
						halfPlayers[TID_OPPONENT][j]->setSeeTime(curTime);
						break;
					}

		int i, j;
		Player *playerPtr;
		for (i = 0; i < HALF_PLAYERS_NUM; i++)
		{
			// Opponents
			if (halfPlayers[TID_OPPONENT][i]->isValid())
			{
				playerPtr = NULL;
				for (j = 0; j < FULL_PLAYERS_NUM; j++)
					if (fullPlayers[TID_OPPONENT][j]->isAlive() &&
					  fullPlayers[TID_OPPONENT][j]->isValid() &&
						fullPlayers[TID_OPPONENT][j]->getSeeTime() != getCurTime() &&
						fullPlayers[TID_OPPONENT][j]->getDistance(*halfPlayers[TID_OPPONENT][i]) <=
						fullPlayers[TID_OPPONENT][j]->getPosDeviation() + halfPlayers[TID_OPPONENT][i]->getPosDeviation() + 1) // One meter for sure
					{
						if (playerPtr == NULL)
							playerPtr = fullPlayers[TID_OPPONENT][j];
						else
						{	
							playerPtr = NULL;
							break;
						}
					}

				if (playerPtr != NULL)
				{
					for (j = 0; j < HALF_PLAYERS_NUM; j++)
						if (i != j && halfPlayers[TID_OPPONENT][j]->isValid() &&
							halfPlayers[TID_OPPONENT][j]->getDistance(*playerPtr) <=
							halfPlayers[TID_OPPONENT][j]->getPosDeviation() + playerPtr->getPosDeviation() + 1) // One meter for sure
						{
							if (playerPtr->getDistance(*halfPlayers[TID_TEAMMATE][i]) >
									halfPlayers[TID_TEAMMATE][j]->getDistance(*playerPtr))
							{
								j = -1;
								break;
							}
						}

					if (j != -1)
					{
						playerPtr->setByAnotherPlayer(*halfPlayers[TID_OPPONENT][i], *body);
						halfPlayers[TID_OPPONENT][i]->kill();
					}
				}
			}
		}
	}
}

void WorldModel::radarHear(const std::string &message, unsigned senderNum)
{
	LOG << "WorldModel::radarHear" << endl;

	if (getBall().getPosDeviation() > 3 ||
		getBall().getAbsVec().getMagnitude() > 7)
		return;

	unsigned id;
	float magnitude, direction;

	for (int i = 0; message[3 * i + 2] != '"'; i++)
	{
		SayEncoding::decodeRadarObject(message, (unsigned)(3 * i + 2),
				id, magnitude, direction);

		LOG << "\tPlayer " << i << " ID: " << id
			<< " Mag: " << magnitude	<< " Dir: " << direction << endl;

		TeamId teamId = (TeamId)(int)(id / 11);
		int uniNum = (id % 11) + 1;

		setFullPlayer(teamId, uniNum - 1).
				updateByRadar(magnitude, direction, getBody(), getBall());

		if (i == 0) // updating view parts
		{
			for (int j = 0; j < VIEW_PARTS_NUM; j++)
			{
				float deltaAngle = abs(getDeltaAngle(
					getFullPlayer(teamId, uniNum - 1).getAbsVec().getDirection(),
					getViewPartDir(j)));

				if (deltaAngle <= (VIEW_PART_SIZE / 2) - 2.5)
				{
					viewPartCycles[j] = 1;
					break;
				}
			}
		}
	}

	hearedSender = senderNum;
}

void WorldModel::initViewPartDirs()
{
	for (unsigned i = 0; i < VIEW_PARTS_NUM; i++)
		viewPartDirs[i] = i * (360 / VIEW_PARTS_NUM);
}

void WorldModel::resetSeeDeltaCycles()
{
	for (unsigned i = 0; i < VIEW_PARTS_NUM; i++)
		viewPartCycles[i] = MAX_SEE_DELTA_CYCLE;
}

void WorldModel::updateSeeDeltaCyclesAfterSee()
{
	for (unsigned i = 0; i < VIEW_PARTS_NUM; i++)
		if (getSeeViewPartValue(i, body->getHeadDir(),
				body->getViewAngle()) >= 0.75)
			viewPartCycles[i] = 0;
}

void WorldModel::updateSeeDeltaCyclesAfterSenseBody()
{
	for (unsigned i = 0; i < VIEW_PARTS_NUM; i++)
		if (viewPartCycles[i] < MAX_SEE_DELTA_CYCLE)
			viewPartCycles[i]++;
}

float WorldModel::getSeeDeltaCycle(float dir, float minRate) const
{
	int pairNum;
	float deltaAngle;

	for (unsigned i = 0; i < VIEW_PARTS_NUM; i++)
	{
		deltaAngle = getDeltaAngle(dir, viewPartDirs[i]);

		if (fabs(deltaAngle) < EXTRA_VIEW_SIZE)
			return viewPartCycles[i];

		if (fabs(deltaAngle) < VIEW_PART_SIZE / 2)
		{
			if (deltaAngle > 0)
				pairNum = i + 1;
			else
				pairNum = i - 1;

			if (pairNum == VIEW_PARTS_NUM)
				pairNum = 0;
			else if (pairNum == -1)
				pairNum = VIEW_PARTS_NUM - 1;

			if (fabs(deltaAngle) > VIEW_PART_SIZE - EXTRA_VIEW_SIZE / 2.00)
				return viewPartCycles[i] * minRate +
					viewPartCycles[pairNum] * (1.00 - minRate);

			return viewPartCycles[i] * (1.00 - minRate) +
				viewPartCycles[pairNum] * minRate;
		}
	}

	assert(0);
	return MAX_SEE_DELTA_CYCLE;
}

float WorldModel::getSeeViewPartValue(unsigned viewPartNum,
									  float headDir, float viewAngle) const
{
	float deltaAngle;

	deltaAngle = fabs(getDeltaAngle(viewPartDirs[viewPartNum], headDir));
	deltaAngle += VIEW_PART_SIZE / 2.0 - viewAngle;

	if (deltaAngle <= EXTRA_VIEW_SIZE + 5) // 5 Degrees for sure
		return reRate(deltaAngle, 0, EXTRA_VIEW_SIZE + 5, 0.75, 1, RRM_REVERSE);

	return reRate(deltaAngle, EXTRA_VIEW_SIZE + 5, 45, 0, 0.75, RRM_REVERSE);
}

void WorldModel::updatePlayModeRemainCycleAfterSenseBody()
{
	if (playModeRemainCycle > 0)
		playModeRemainCycle--;
}

void WorldModel::updateBallStatus()
{
	float checkedTmmNum = curInterCalculate.getTmmCheckedPlayersNum();
	float checkedOppNum = curInterCalculate.getOppCheckedPlayersNum();
	bool kickableTmm = isBallInTmmKickable();
	bool kickableOpp = isBallInOppKickable();
	bool kickableBody = isBallKickable();

	if (!curInterCalculate.existFastestPlayer())
		ballStatus = BS_FREE_BALL;
	else if (kickableBody)
	{
		if (kickableOpp)
			ballStatus = BS_KICKABLE_BODY_OPP;
		else if (kickableTmm)
			ballStatus = BS_KICKABLE_BODY_TMM;
		else
			ballStatus = BS_KICKABLE_BODY;
	}
	else if (kickableTmm || kickableOpp)
	{
		if (kickableTmm && kickableOpp)
			ballStatus = BS_KICKABLE_TMM_OPP;
		else if (kickableTmm)
			ballStatus = BS_KICKABLE_TMM;
		else
			ballStatus = BS_KICKABLE_OPP;
	}
	else if (curInterCalculate.getCheckedListNum(*body) != NOVALUE) // body is a checkPlayer.
	{
		if (checkedOppNum > 0)
			ballStatus = BS_FREE_BALL_BODY_OPP;
		else if (checkedTmmNum > 1)
			ballStatus = BS_FREE_BALL_BODY_TMM;
		else
			ballStatus = BS_FREE_BALL_BODY;
	}
	else
	{
		if (checkedTmmNum > 0 && checkedOppNum > 0)
			ballStatus = BS_FREE_BALL_TMM_OPP;
		else if (checkedTmmNum > 0)
			ballStatus = BS_FREE_BALL_TMM;
		else
			ballStatus = BS_FREE_BALL_OPP;
	}

	LOG << "ballStatus: " << ballStatus << endl;
}

void WorldModel::resetHearedVarsBeforeHear()
{
	hearedHeader = '\0';
	hearedSender = 0;
}

void WorldModel::resetHearedVarsAfterSenseBody()
{
	hearedPlanStatus = 0;
	hearedInterceptNum = 0;
	hearedSuggestPassTime = 0;
}

void WorldModel::parseChangePlayerType(const SExpression &exp)
{
	SExpAtomic *at;
	unsigned playerNum;
	unsigned playerType;

	if (exp.size() == 3) // teammate change player type
	{
		at = dynamic_cast<SExpAtomic *>(exp[1]);
		assert(at);
		playerNum = at->asInt();

		at = dynamic_cast<SExpAtomic *>(exp[2]);
		assert(at);
		playerType = at->asInt();

		assert(playerNum > 0 && playerNum <= FULL_PLAYERS_NUM);
		assert((int)playerType < playerParam["player_types"].asInt());
		fullPlayers[TID_TEAMMATE][playerNum - 1]->setType(playerType,
														  playerTypes[playerType]);
	}
	else // opponent change player type
	{
		assert(exp.size() == 2);
	}
}

int WorldModel::checkPlayModeForPenalty() const
{
	if (virtualPlayMode == VPM_PENALTY_ON_FIELD)
		return 1;// Rotate Field Flags
	if (virtualPlayMode == VPM_PENALTY_ON_FIELD_OPP)
		return 2;// Rotate Field Flags ONLY for goalie

	return 0;
}

// Getting functions
bool WorldModel::isBodyReplaced() const
{
	return bodyReplacedFlag;
}

unsigned WorldModel::getFlagsNum() const
{
	return flagsNum;
}

unsigned WorldModel::getGoalsNum() const
{
	return goalsNum;
}

unsigned WorldModel::getLinesNum() const
{
	return linesNum;
}

unsigned WorldModel::getSeeCycle() const
{
	return seeCycle;
}

unsigned WorldModel::getCurCycle() const
{
	return curCycle;
}

unsigned WorldModel::getCurTime() const
{
	return curTime;
}
unsigned WorldModel::getRefereeHearTime() const
{
	return refereeHearTime;
}

const set<unsigned> & WorldModel::getMarkers() const
{
    return markers;
}

Side WorldModel::getFieldSide() const
{
	return fieldSide;
}

PlayMode WorldModel::getPlayMode() const
{
	return playMode;
}

VirtualPlayMode WorldModel::getVirtualPlayMode() const
{
	return virtualPlayMode;
}

float WorldModel::getViewPartDir(unsigned num) const
{
	return viewPartDirs[num];
}

unsigned WorldModel::getViewPartCycle(unsigned num) const
{
	return viewPartCycles[num];
}

unsigned WorldModel::getFreeKickAdvicerSayNumber() const
{
	return freeKickAdvicerSayNumber;
}

unsigned WorldModel::getFreeKickAdvicerSayCycle() const
{
	return freeKickAdvicerSayCycle;
}

const Timer &WorldModel::getTimer() const
{
	return (const Timer &)*timer;
}
unsigned WorldModel::getPlayModeRemainCycle() const
{
	return playModeRemainCycle;
}

const Param &WorldModel::getServerParam() const
{
	return (const Param &)serverParam;
}

const Param &WorldModel::getPlayerParam() const
{
	return (const Param &)playerParam;
}

const Param &WorldModel::getPlayerType(unsigned num) const
{
	return (const Param &)playerTypes[num];
}

const InterceptCalculate &WorldModel::getCurInterCalculate() const
{
	return (const InterceptCalculate &)curInterCalculate;
}

const KickCalculate &WorldModel::getCurKickCalculate() const
{
	return (const KickCalculate &)curKickCalculate;
}

const Ball &WorldModel::getBall() const
{
	return (const Ball &)*ball;
}

const Body &WorldModel::getBody() const
{
	return (const Body &)*body;
}

const Player &WorldModel::getFullPlayer(TeamId teamId, unsigned num) const
{
	return (const Player &)*fullPlayers[teamId][num];
}

const Player &WorldModel::getHalfPlayer(TeamId teamId, unsigned num) const
{
	return (const Player &)*halfPlayers[teamId][num];
}

const Player &WorldModel::getQuarterPlayer(unsigned num) const
{
	return (const Player &)*quarterPlayers[num];
}

BallStatus WorldModel::getBallStatus() const
{
	return ballStatus;
}

char WorldModel::getHearedHeader() const
{
	return hearedHeader;
}

unsigned WorldModel::getHearedSender() const
{
	return hearedSender;
}

unsigned WorldModel::getHearedInterceptNum() const
{
	return hearedInterceptNum;
}

unsigned WorldModel::getHearedPlanStatus() const
{
	return hearedPlanStatus;
}

int WorldModel::getPlanStatusSender() const
{
	return planStatusSender;
}

unsigned WorldModel::getPlayOnPastCycles() const
{
	return curTime - playOnStartCycle;
}

unsigned WorldModel::getHearedSuggestPassNum() const
{
	return hearedSuggestPassNum;
}

unsigned WorldModel::getHearedSuggestPassTime() const
{
	return hearedSuggestPassTime;
}

float WorldModel::getHearedSuggestPassDir() const
{
	return hearedSuggestPassDir;
}

float WorldModel::getHearedSuggestPassSpeed() const
{
	return hearedSuggestPassSpeed;
}

float WorldModel::getHearedSuggestPassWeight() const
{
	return hearedSuggestPassWeight;
}

const string &WorldModel::getHearedPlanInfo() const
{
	return hearedPlanInfo;
}

unsigned WorldModel::getHearedPlanTime() const 
{
	return hearedPlanTime;
}

unsigned WorldModel::getHearedPlanSender() const 
{
	return hearedPlanSender;
}

// Setting functions
void WorldModel::setBodyReplaced(bool bodyReplacedFlagArg)
{
	bodyReplacedFlag = bodyReplacedFlagArg;
}

void WorldModel::setFlagsNum(unsigned flagsNumArg)
{
	flagsNum = flagsNumArg;
}

void WorldModel::setGoalsNum(unsigned goalsNumArg)
{
	goalsNum = goalsNumArg;
}

void WorldModel::setLinesNum(unsigned linesNumArg)
{
	linesNum = linesNumArg;
}

void WorldModel::setSeeCycle(unsigned seeCycleArg)
{
	seeCycle = seeCycleArg;
}

void WorldModel::setCurCycle(unsigned curCycleArg)
{
	curCycle = curCycleArg;
}

void WorldModel::setCurTime(unsigned curTimeArg)
{
	curTime = curTimeArg;
}

void WorldModel::setFieldSide(Side fieldSideArg)
{
	fieldSide = fieldSideArg;

	if (fieldSide == SI_LEFT)
		fieldData.setRightSideRate(1);
	else
		fieldData.setRightSideRate(-1);
}

FieldData &WorldModel::setFieldData()
{
	return fieldData;
}

void WorldModel::setVirtualPlayMode(VirtualPlayMode virtualPlayModeArg)
{
	virtualPlayMode = virtualPlayModeArg;
}

Timer &WorldModel::setTimer()
{
	return (Timer &)*timer;
}

Ball &WorldModel::setBall()
{
	return (Ball &)*ball;
}

Body &WorldModel::setBody()
{
	return *body;
}

Player &WorldModel::setFullPlayer(TeamId teamId, unsigned num)
{
	return (Player &)*fullPlayers[teamId][num];
}

Player &WorldModel::setHalfPlayer(TeamId teamId, unsigned num)
{
	return (Player &)*halfPlayers[teamId][num];
}

Player &WorldModel::setQuarterPlayer(unsigned num)
{
	return (Player &)*quarterPlayers[num];
}

